/*----------------------------------------------
   Img2Level MSTE Tool - Roberto Prieto
   Copyright (C) 2007-2012 MegaStorm Systems
  ----------------------------------------------
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
    
    Roberto Prieto
    contact@megastormsystems.com
    http://www.megastormsystems.com

    Note that all library licenses (including the one for this library)
    are located at /licenses directory. Please, include them on your applications.

 --------------------------------------------------------------------------

  Image <-> Level converter with optional support to custom map
  -------------------------------------------------------------
  
  - This tool performs all the possible conversions between image<->level or level<->level
  - It supports BMP/PNG images and native MSM or TMX (Tiled Qt) levels
 */

// Includes
#include "CRM32Pro.h"
#include "../source/MSTE.h"
#include "../source/MSTE_MapData.h"
#include "../source/MSTE_Tools.h"

// Defines
#define TOOL_VERSION "Image2Level v1.0"
#define MAX_LAYERS 5

// Prototypes
void PrintHelp(int);
void RemoveLayers(cMSTE *, int);

// Optional Custom Map 
// This is just an example
class cMyCustomMap : public cMSTE_MapBase
{
public:	
    // Mandatory methods
    // 1.1.Constructor
    cMyCustomMap(int height,int width);
    // 1.2.Method to be used by our maps factory to create an instance of this class
    static cMSTE_MapBase *createInstance(int y,int x);

    // Optional methods
    // 2.1.Destructor
    ~cMyCustomMap();
    // 2.2.My custom methods
    int PrintInfo(char,int);
    int SaveMapdataNode(int,int);
    int LoadMapdataNode(int,int);
    int CheckMap();
    int CellRender(int map_x, int map_y, int surf_x, int surf_y, cMSTE *TE,int iOpacity);

public:
    // Dynamic 2D array for my custom map
    unsigned int **iCustomMap;
    int iW, iH;
};

// -------------MAIN FUNCTION----------------
int main(int argc,char *argv[])
{
    int iLayer = -1, iOffX = 32, iOffY = 32, iFlag;
    char szFirst[255], szSecond[255], szTmp[4];
    cMSTE *TE;

    // 1.Log system initialize and initial checks
    ILogSystem.Init("I2L.log",LOG_FILE,LOG_NORMAL,TOOL_VERSION);
    if(argc < 3)
    {
        PrintHelp(1);
        return 0;
    }
    else PrintHelp(0);

    // 2.CRM32Pro and MSTE initialization
    if(CRM32Pro.Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_TIMER) < 0)
    {
        ILogSystem.Msg(LOG_NORMAL," [LOG] Error: couldn't initialize CRM32Pro: %s\n",SDL_GetError());
        return 1;
    }
    CRM32Pro.Config.Title = TOOL_VERSION;
    CRM32Pro.Config.VideoWidth = 320;
    CRM32Pro.Config.VideoHeight = 200;
    CRM32Pro.Config.bMTFriendly = 1;
    if(!CRM32Pro.SetVideoMode())
    {
        CRM32Pro.Quit();
        return 1;
    }
    TE = new cMSTE;
    TE->Init(MAX_LAYERS);
    TE->SetRenderTarget(CRM32Pro.screen);
    if(TE->RegisterMapDataHandler(&cMyCustomMap::createInstance)) ILogSystem.Msg(LOG_NORMAL," [I2L] Info: custom map successfully registered\n");

    // 3.Read 1st and 2nd parameters
    szCopy(szFirst,argv[1],255);
    szCopy(szSecond,argv[2],255);

    // 4.1.Image to Level section
    if(CheckFileExtension(szFirst,".bmp") || CheckFileExtension(szFirst,".png"))
    {
        // Get optional cell sizes
        if(argc > 3)
        {
            szCopy(szTmp,argv[3],4);
            sscanf(szTmp, "%d", &iOffX);
        }
        if(argc > 4)
        {
            szCopy(szTmp,argv[4],4);
            sscanf(szTmp, "%d", &iOffY);
        }       
        // Tiled Qt format
        if(CheckFileExtension(szSecond,".tmx"))
        {
            printf(" - Converting image '%s' with cell size (%d,%d) to Tiled Qt format...\n",szFirst,iOffX,iOffY);
            TE->ImportImage(0,szFirst,iOffX,iOffY);
            iFlag = TE->ExportTMX(szSecond,"Img2Level level");        
        }
        // Native MSM format
        else if(CheckFileExtension(szSecond,".msm"))
        {
            printf(" - Converting image '%s' with cell size (%d,%d) to MSTE native format...\n",szFirst,iOffX,iOffY);
            TE->ImportImage(0,szFirst,iOffX,iOffY);
            iFlag = TE->SaveLevel(szSecond,NULL,"Img2Level level");    
        }
        // Error
        else { printf(" - Unsupported destination format '%s'\n",szSecond); iFlag = 0; }
        if(iFlag == 1) printf(" - Conversion successfully performed\n");
        else printf(" - Error performing the conversion! Check the log file for further details\n");
    }

    // 4.2.Level to Image or Level section
    else if(CheckFileExtension(szFirst,".msm") || CheckFileExtension(szFirst,".tmx"))
    {
        // Get optional layer
        if(argc > 3)
        {
            szCopy(szTmp,argv[3],4);
            sscanf(szTmp, "%d", &iLayer);
        }
        // Load input
        if(CheckFileExtension(szFirst,".msm")) iFlag = TE->LoadLevel(szFirst);
        else iFlag = TE->ImportTMX(szFirst);

        // BMP and PNG
        if(CheckFileExtension(szSecond,".bmp") || CheckFileExtension(szSecond,".png"))
        {
            printf(" - Converting '%s' (layer: %d) to '%s' image...\n",szFirst,iLayer,szSecond);
            iFlag = TE->ExportImage(iLayer,szSecond);
        }
        // Tiled Qt format
        else if(CheckFileExtension(szSecond,".tmx"))
        {
            printf(" - Converting '%s' (layer: %d) to '%s' Tiled Qt...\n",szFirst,iLayer,szSecond);
            if(iLayer != -1) RemoveLayers(TE,iLayer);
            iFlag = TE->ExportTMX(szSecond);
        }
        // Native MSM format
        else if(CheckFileExtension(szSecond,".msm"))
        {
            printf(" - Converting '%s' (layer: %d) to '%s' MSM...\n",szFirst,iLayer,szSecond);
            if(iLayer != -1) RemoveLayers(TE,iLayer);
            iFlag = TE->SaveLevel(szSecond);
        }
        // Error
        else { printf(" - Unsupported destination format '%s'\n",szSecond); iFlag = 0; }
        if(iFlag == 1) printf(" - Conversion successfully performed\n");
        else printf(" - Error performing the conversion! Check the log file for further details\n");
    }

    // 4.3.Print help...
    else PrintHelp(1);

    // Free resources and return
    TE->Quit();
    delete TE;   
    CRM32Pro.Quit();
    getchar();
    return 0;
}

// Remove all layers but iLayer
void RemoveLayers(cMSTE *TE, int iLayer)
{
    for(int i = 0; i < TE->AllocateLayers(-1); ++i) if(i != iLayer) TE->FreeLayer(i);
}

// PrintHelp()
void PrintHelp(int m)
{
    printf("-----------------------------------------------------------\n");
    printf("        %s MSTE Tool - Roberto Prieto  \n",TOOL_VERSION);
    printf("         Copyright (C) 2007-2012 MegaStorm Systems\n");
    printf("-----------------------------------------------------------\n\n");
    if(m > 0)
    {
        printf(" Usage:\n");
        printf("   -  'Img2Level BMP|PNG MSM|TMX [Tile X size] [Tile Y size]\n");
        printf("   -  'Img2Level MSM|TMX BMP|PNG [Layer]\n");
        printf("   -  'Img2Level MSM|TMX MSM|TMX [Layer]\n\n");
        printf(" Tile X and Y sizes are 32 by default\n");
        printf(" Layer is -1 for converting all layers\n\n");
        getchar();
    }
}

// Optional Custom Map 
// This is just an example

// Constructor: we have to call to base class constructor (MANDATORY)
cMyCustomMap::cMyCustomMap(int height,int width) : cMSTE_MapBase(height,width)
{
    iW = width; iH = height;
    if(iW > 0 && iH > 0)	
    {
        iCustomMap = Create2DArray<unsigned int>(height,width);
        // Loop filling our custom map with a default value
        for(int y = 0; y < iH; y++)
            for(int x = 0; x < iW; x++) iCustomMap[y][x] = 5;
    }
    else iCustomMap = NULL;
    // Assign a new map type name
    strcpy(sMapType,"MyMap"); 
}

// Method to be used by our maps factory to create an instance of this class (MANDATORY)
cMSTE_MapBase *cMyCustomMap::createInstance(int y,int x) 
{ 
    // Create an instance of this class.
    return new cMyCustomMap(y,x); 
} 

// Destructor (OPTIONAL)
cMyCustomMap::~cMyCustomMap()
{
    if(iCustomMap != NULL) Delete2DArray(iCustomMap);
    iW = iH = 0;
    iCustomMap = NULL;
}

// Update the given cell (OPTIONAL)
int cMyCustomMap::CellRender(int map_x, int map_y, int surf_x, int surf_y, cMSTE *TE, int iOpacity)
{
    // Check our custom map first, if there is a value != 10, we proceed to update the cell of mapbase (render the tile)
    if(iCustomMap[map_y][map_x] != 10) 
        cMSTE_MapBase::CellRender(map_x,map_y,surf_x,surf_y,TE,iOpacity);
    return 1;
}

// Map validation (OPTIONAL) 
int cMyCustomMap::CheckMap()
{
    int iRet = 1;
    unsigned int iMax;

    // Check the base map
    //if(cMSTE_MapBase::CheckMap() == 0) return 0;

    // Previous check
    if(iCustomMap == NULL) return 0;
    if(iW == 0) return 0;
    if(iH == 0) return 0;
    iMax = 20;

    // Loop through iCustomMap
    for(int i = 0; i < iH; i++)
        for(int j = 0; j < iW; j++)
        {
            if(iCustomMap[i][j] < 5) 
            { 
                ILogSystem.Msg(LOG_NORMAL,"  [MyCustomMap] Warning: map[%d][%d] has a wrong value of %d.\n",i,j,iCustomMap[i][j]); 				
                iRet = 0; 
            }
            else if(iCustomMap[i][j] > iMax)
            {
                ILogSystem.Msg(LOG_NORMAL,"  [MyCustomMap] Warning: map[%d][%d] has a wrong value of %d.\n",i,j,iCustomMap[i][j]); 				
                iRet = 0; 
            }
        }	
        // Return
        return iRet;	
}

// My custom PrintInfo
int cMyCustomMap::PrintInfo(char bTilesetReady, int iLogLevel)
{
    // 1.First of all, call to mapbase info
    cMSTE_MapBase::PrintInfo(bTilesetReady,iLogLevel);

    // 2.Now print our custom info
    ILogSystem.Msg(iLogLevel,"         . Map type '%s' - Custom map size %dx%d\n",sMapType,iW,iH);
    return 1;
}


// My custom SaveMapdataNode
int cMyCustomMap::SaveMapdataNode(int idXML, int iOutput)
{
    char	*szTmp = NULL;
    int		retvalue = 0;

    // 1.Get our CSV string
    retvalue = A2DtoCSV(iCustomMap,iW,iH,szTmp);

    // 2.Select our type
    switch(iOutput)
    {
        // TMX output
    case 1:
        // A1.Save first mapdata attributes as a data tag does not support properties tag (as of Tiled 0.81)
        retvalue = CRM32Pro.XMLNodeChild(idXML); // Go back to "properties" block
        retvalue = WriteTMXProperty(idXML,"custom-type",sMapType);
        retvalue = WriteTMXProperty(idXML,"custom-width",iW);
        retvalue = WriteTMXProperty(idXML,"custom-height",iH);	
        retvalue = WriteTMXProperty(idXML,"custom-data",szTmp);
        CRM32Pro.XMLNodeParent(idXML); // Exit from "properties" block
        // A2.Now, call to export base mapdata
        cMSTE_MapBase::SaveMapdataNode(idXML,1);
        break;

        // MSM output
    default:
    case 0:
        // B1.First of all, save the mapbase node
        cMSTE_MapBase::SaveMapdataNode(idXML,0);
        // B2.Now save our custom mapdata. These two fields are mandatory.
        retvalue = CRM32Pro.XMLNodeCreate(idXML,"mapdata"); // Create a new node and point to it.
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"type",sMapType);				
        // B3.And finally, save our custom attributes
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"width",iW);
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"height",iH);	
        // B4.Save our CSV string
        CRM32Pro.XMLTextSet(idXML,szTmp);
        break;
    }	

    // 3.Point back and return
    if(szTmp != NULL) { delete []szTmp; szTmp = NULL; }
    CRM32Pro.XMLNodeParent(idXML); // To our layer
    return retvalue;
}

// My custom LoadMapdataNode
// idXML is pointing inside a mapdata/data subnode
int cMyCustomMap::LoadMapdataNode(int idXML, int iInput)
{
    char *szTmp = NULL;

    // 1.Select our type
    switch(iInput)
    {
        // TMX input
    case 1:		
        // First of all, load the mapbase node as we are on the data subnode
        cMSTE_MapBase::LoadMapdataNode(idXML,1);	
        // Save current position and look for <properties>
        CRM32Pro.XMLNodeStore(idXML);	// save node pointer
        CRM32Pro.XMLNodeParent(idXML);	// go back to layer node
        CRM32Pro.XMLNodeChild(idXML);	// go to the first child
        do
        {
            // Is it a properties node?
            if(!strcmp(CRM32Pro.XMLNodeGetName(idXML),"properties"))
            {
                // Try to load our special layer properties
                if(!ReadTMXProperty(idXML,"custom-width",&iW)) return 0;
                if(!ReadTMXProperty(idXML,"custom-height",&iH)) return 0;
                if(!ReadTMXProperty(idXML,"custom-data",&szTmp)) return 0;

            }
        } while(CRM32Pro.XMLNodeNext(idXML)); // next layer subnode
        CRM32Pro.XMLNodeRestore(idXML);	// restore node pointer
        break;

        // MSM input
    default:
    case 0:
        // The mapbase is loaded on the LoadLevel() method
        // Get attributes
        if(!CRM32Pro.XMLAttributeGet(idXML,"width",&iW)) return 0;
        if(!CRM32Pro.XMLAttributeGet(idXML,"height",&iH)) return 0;
        // Some validations
        if(iW <=0 || iH <= 0) return 0;
        // B2.Create the iCustomMap and load the data
        szTmp = CRM32Pro.XMLTextGet(idXML);		
        break;
    }

    // 2.Try to process de custom data
    if(szTmp == NULL) return 0;
    if(iCustomMap != NULL) delete []iCustomMap;
    iCustomMap = Create2DArray<unsigned int>(iH,iW);
    if(CSVtoA2D(iCustomMap,iW,iH,szTmp)) return 1;

    // 3.End
    return 0;
}
