/*----------------------------------------------
   MegaStorm Tile-based Engine - Roberto Prieto
   Copyright (C) 2007-2012 MegaStorm Systems
  ----------------------------------------------
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
    
    Roberto Prieto
    contact@megastormsystems.com
    http://www.megastormsystems.com

    Used software licenses, including the one for this engine, are located
    at /licenses directory. Please, include them on your distributions.

 --------------------------------------------------------------------------
             MSTE Main engine class
    
    Changelog: read MSTE.h to see all changes
   
 -------------------------------------------------------------------------- */
#include <CRM32Pro.h>
#include <math.h>
#include "MSTE.h"
#include "MSTE_Tools.h"
#include "MSTE_MapData.h"

//-------------------------------------------------------------------------------
// General stuff methods
//-------------------------------------------------------------------------------
//#define DEBUG_SMOOTH

#ifdef DEBUG_SMOOTH
FILE *fpCSV = NULL;
#endif

// Constructor
cMSTE::cMSTE()
{
    bTileEngine = 0;
    sRenderTarget = NULL;
    strcpy(szName,"");
    strcpy(szCreationDate,"");
    strcpy(szModificationDate,"");
    // Reset RegisteredMapData
    for(int i = 0; i < TE_REGISTEREDMAPDATA_MAX; i++) RegisteredMapData[i] = NULL;
    // Clear layer and tileset vectors
    vLayers.clear();
    vTilesets.clear();
    // Default log verbosity
    LogVerbosity(2);

#ifdef DEBUG_SMOOTH
    fpCSV = fopen("smooth.csv","wt");
#endif
}

// Destructor
cMSTE::~cMSTE()
{
    if(bTileEngine) Quit();
#ifdef DEBUG_SMOOTH
    fclose(fpCSV); 
#endif
}

// Layer default constructor 
cMSTE::st_layers::st_layers()
{
    reset();
    iLastUpdateTime = SDL_GetTicks();
}

// Layer default destructor 
cMSTE::st_layers::~st_layers()
{
    if(bResControl >= TE_RESCONTROL_AUTO) delete map;
    reset();
}

// Layer reset
void cMSTE::st_layers::reset()
{
    iStatus = 0;
    bResControl = TE_RESCONTROL_USER;
    bTilesetReady = 0;
    fParallaxRatio = fSpeedX = fSpeedY = 0.0f;
    map = NULL;
    strcpy(szName,TE_LAYER_EMPTY);		
    OnUpdateEnd = OnRenderEnd = NULL;	
    iLastUpdateTime = iGridColor = iDisableNextSmooth = 0;
    iAlpha = 255;	
    fPosX = fPosY = fGfxX = fGfxY = 0;
    fLogicX = fLogicY = fLogicX_pre = fLogicY_pre = 0;
    iOffsetX = iOffsetY = 0;  
}

// Tileset default constructor 
cMSTE::st_tilesets::st_tilesets()
{
    iGID = 0;
    szSource = NULL;
    tTileset = NULL;
    iNumTiles = 0;
    bResControl = TE_RESCONTROL_USER;	
}

// Tileset constructor
cMSTE::st_tilesets::st_tilesets(int gID, char *Source, CRM32Pro_CTile *tile, char bRes)
{
    int iX = 0, iY = 0;

    if(gID <= 0) return; // Wrong global id
    iGID = gID;
    if(Source != NULL)
    {
        szSource = new char[strlen(Source) + 1];
        strcpy(szSource,Source);
    }
    tTileset = tile;
    tTileset->GetOffset(&iX,&iY);
    iNumTiles = (tTileset->GetSizeX() / iX) * (tTileset->GetSizeY() / iY);
    bResControl = bRes;
}

// Tileset default destructor 
cMSTE::st_tilesets::~st_tilesets()
{
    iGID = 0;
    if(szSource != NULL) delete []szSource;
    szSource = NULL;
    if(bResControl >= TE_RESCONTROL_AUTO) delete tTileset;
    tTileset = NULL;
    iNumTiles = 0;
    bResControl = TE_RESCONTROL_USER;	
}

//! Initialize the engine.
/*! \param iNumLayers Number of layers to be allocated. By default 3 layers. You can use at any time AllocateLayers() for increasing or decreasing dinamically it
\param str Pointer to a string with the selected name for this level.
\return 0 Failed
\return 1 Succeed */
int cMSTE::Init(int iNumLayers, char *str)
{
    if(bTileEngine) return 0;
    if(str == NULL) return 0;
    if(!CRM32Pro.RequireVersion(5,22)) return 0;

    // Init internal vars			
    RegisteredMapData[0] = &cMSTE_MapBase::createInstance;	
    bTileEngine = 1;
    if(!strcmp(str,""))
    {
        SetName("Default-LevelName");
    }
    else SetName(str);
    szCopy(szCreationDate,GetDateStamp(),sizeof(szCreationDate));
    szCopy(szModificationDate,GetDateStamp(),sizeof(szModificationDate));
    ILogSystem.Msg(iLogN,"\n [MSTE] Info: engine initialized.\n");
    // Allocate layers
    AllocateLayers(iNumLayers);
    return 1;
}

//! Close the engine.
/*! In each layer with TE_RESCONTROL_USER flag, we dont free any tileset or mapdata. 
This have to be done at the developer preference time BUT always after call this method.
<br>And with TE_RESCONTROL_AUTO flag, the engine will remove the tileset and mapdata.
\return 0 Failed
\return 1 Succeed */
int cMSTE::Quit()
{
    FreeLevel();
    vLayers.clear(); // Remove allocated layers
    memset(&rViewport,0,sizeof(SDL_Rect));
    sRenderTarget = NULL;	
    bTileEngine = 0;
    ILogSystem.Msg(iLogN,"\n [MSTE] Info: engine sucessfully closed.\n");
    return 1;
}

//! Set or get the number of layers
/*! \param iN number of layers to be allocated. With -1, returns the current number of allocated layers.<br>
If the number of current layers is lesser than iN, this method will allocate empty new layers. Otherwise it will 
remove needed layers using FreeLayer() method.
\return 0 Failed
\return 1 Succeed */
int cMSTE::AllocateLayers(int iN)
{
    int iL = (int)vLayers.size();

    // Return allocated layers
    if(iN < 0) return iL;

    // Check what we have to do
    iN -= iL;
    // Add new layers
    if(iN > 0) 
    {
        for(int i = 0; i < iN; i++)
        {
            vLayers.push_back(new st_layers());
        }
        ILogSystem.Msg(iLogN," [MSTE] Info: adding %d layer slots. Allocated layers: %d.\n",iN,vLayers.size());
    }
    // Remove layers
    else if(iN < 0) 
    {
        iN *= -1;
        for(int i = 0; i < iN; i++)
        {
            FreeLayer((int)vLayers.size() - 1);
            vLayers.pop_back();
        }
        ILogSystem.Msg(iLogN," [MSTE] Info: removing %d layer slots. Allocated layers: %d.\n",iN,vLayers.size());
    }
    // do nothing
    // else { }

    return 1;
}

//! Free all resources of current level(layer maps and tilesets).
/*! It does a reset of the current level. All layers are reset but the layer slots are kept.
<br> It holds the render target, viewport and the number of allocated layers. 
\return 0 Failed
\return 1 Succeed */
int cMSTE::FreeLevel()
{
    int i;

    // Remove tilesets calling its destructor
    for(i = 0; i < (int)vTilesets.size(); i++) delete vTilesets[i];
    vTilesets.clear();

    // Remove the layers
    for(i = 0; i < (int)vLayers.size(); i++) FreeLayer(i); 

    // Remove registered maps. [0] is always cMSTE_MapBase and we keep it
    for(i = 1; i < TE_REGISTEREDMAPDATA_MAX; i++) RegisteredMapData[i] = NULL;

    // Reset engine attributes
    strcpy(szName,"");
    strcpy(szCreationDate,"");
    strcpy(szModificationDate,"");
    return 1;
}

//! Free a layer
/*! If the layer has TE_RESCONTROL_USER flag, we dont free the mapdata. 
This have to be done at the developer preference time BUT always after call this method.
<br>And with TE_RESCONTROL_AUTO flag, this method will remove the mapdata of the given layer.
<br>It also reset all the layer settings and this layer slot is ready to be used. We dont remove the layer slot, use AllocateLayers() for doing that.
\param layer Selected layer to free(from 0 to number of allocated layers)
\return 0 Failed
\return 1 Succeed */
int cMSTE::FreeLayer(int layer)
{
    if(bTileEngine == 0) return 0;
    if(layer < 0 || layer >= (int)vLayers.size()) return 0;

    // Remove the layer resetting it for future usages (the vector slot is kept)
    vLayers[layer]->st_layers::~st_layers();

    return 1;
}

//! Print to the ILogSystem information about the current engine state.
/*! \param layer Show information about given layer number(from 0 to number of allocated layers) or -1(default value) to show 
information about all layers.*/
void cMSTE::InfoPrint(int layer)
{
    int i, start, end;

    // 1.Needs an init
    if(bTileEngine == 0)
    {
        ILogSystem.Msg(iLogL," [MSTE] InfoPrint: the scroll engine was not initialized.\n");
        return;
    }
    // 2.Needs a render surface
    if(bTileEngine == 1)
    {
        ILogSystem.Msg(iLogL," [MSTE] InfoPrint: the render target was not set.\n");
        return;
    }
    // 3.Loop through all layers or selected one
    start = 0; end = (int)vLayers.size() - 1;
    ILogSystem.Msg(iLogL,"\n [MSTE] InfoPrint - MSTE v%s - Level name '%s'\n",MSTE_VERSION,szName);
    ILogSystem.Msg(iLogL,"        Created on %s\n",szCreationDate);
    ILogSystem.Msg(iLogL,"        Last modification on %s\n\n",szModificationDate);
    if(layer != -1) { start = layer; end = layer; }
    else ILogSystem.Msg(iLogL,"      . Layers allocated: %d\n",(int)vLayers.size());
    if(start < 0 || end >= (int)vLayers.size()) return;
    for(i = start; i <= end; i++)
    {
        if(!strcmp(vLayers[i]->szName,TE_LAYER_EMPTY)) continue;

        ILogSystem.Msg(iLogL,"      -> Layer name '%s' with ID '%d' ",vLayers[i]->szName,i);
        if(vLayers[i]->fParallaxRatio == 0.0f) ILogSystem.Msg(iLogL,"(Independent)\n");
        else ILogSystem.Msg(iLogL,"(Linked - Parallax ratio %.2f)\n",vLayers[i]->fParallaxRatio);

        ILogSystem.Msg(iLogL,"         . Flags: ");
        if(vLayers[i]->iStatus == 0) ILogSystem.Msg(iLogL,"None");
        else
        {
            if(vLayers[i]->iStatus & TE_LAYERFLAG_UPDATECELLS) ILogSystem.Msg(iLogL,"UpdateCells ");
            else if(vLayers[i]->iStatus & TE_LAYERFLAG_UPDATE) ILogSystem.Msg(iLogL,"Update ");
            if(vLayers[i]->iStatus & TE_LAYERFLAG_RENDER) ILogSystem.Msg(iLogL,"Render ");
            if(vLayers[i]->iStatus & TE_LAYERFLAG_SMOOTHSCROLL) ILogSystem.Msg(iLogL,"SmoothScroll ");
            if(vLayers[i]->iStatus & TE_LAYERFLAG_AUTOSCROLL) ILogSystem.Msg(iLogL,"AutoScroll ");
            if(vLayers[i]->iStatus & TE_LAYERFLAG_LOOPX) ILogSystem.Msg(iLogL,"LoopX ");
            if(vLayers[i]->iStatus & TE_LAYERFLAG_LOOPY) ILogSystem.Msg(iLogL,"LoopY ");
            if(vLayers[i]->iStatus & TE_LAYERFLAG_SHOWGRID) ILogSystem.Msg(iLogL,"ShowGrid ");
            if(vLayers[i]->iStatus & TE_LAYERFLAG_PAUSE) ILogSystem.Msg(iLogL,"Pause ");
        }
        ILogSystem.Msg(iLogL," - Alpha value %d\n",vLayers[i]->iAlpha);

        ILogSystem.Msg(iLogL,"         . Position (%.2f,%.2f) - Speed (%.2f,%.2f) pixels/second - Max Offset (%d,%d)\n",
            vLayers[i]->fPosX,vLayers[i]->fPosY,vLayers[i]->fSpeedX,vLayers[i]->fSpeedY,vLayers[i]->iOffsetX, vLayers[i]->iOffsetY);

        if(vLayers[i]->map != NULL) vLayers[i]->map->PrintInfo(vLayers[i]->bTilesetReady,iLogL);
        else ILogSystem.Msg(iLogL,"         . No mapdata is present\n");

        ILogSystem.Msg(iLogL,"         . OnUpdateEnd: %s - OnRenderEnd: %s - Resource control: %s\n\n",
            (vLayers[i]->OnUpdateEnd != NULL)?"enabled":"disabled",
            (vLayers[i]->OnRenderEnd != NULL)?"enabled":"disabled",
            (vLayers[i]->bResControl==TE_RESCONTROL_AUTO)?"auto":((vLayers[i]->bResControl==TE_RESCONTROL_INTERNAL)?"internal":"user"));
    }

    // 4.Loop through all tilesets
    if(layer == -1) ILogSystem.Msg(iLogL,"      . Tilesets loaded: %d\n",vTilesets.size());
    for(i = 0; (i < (int)vTilesets.size() && layer == -1); i++)
    {
        if(vTilesets[i]->tTileset != NULL)
        {
            int iTmp = 0, iTmp2 = 0;
            ILogSystem.Msg(iLogL,"      -> Tileset name '%s' with global ID '%d'\n", vTilesets[i]->tTileset->GetName(),vTilesets[i]->iGID);
            ILogSystem.Msg(iLogL,"         . Source '%s' - Resource control: %s\n",vTilesets[i]->szSource,
                (vTilesets[i]->bResControl==TE_RESCONTROL_AUTO)?"auto":((vTilesets[i]->bResControl==TE_RESCONTROL_INTERNAL)?"internal":"user"));
            vTilesets[i]->tTileset->GetOffset(&iTmp,&iTmp2);
            ILogSystem.Msg(iLogL,"         . Number of tiles %d of %dx%d\n",vTilesets[i]->iNumTiles,iTmp,iTmp2);

            iTmp = vTilesets[i]->tTileset->GetAlpha();
            ILogSystem.Msg(iLogL,"         . Alpha blending: ");
            if(iTmp == 256) ILogSystem.Msg(iLogL,"per-pixel");
            else if(iTmp == 255) ILogSystem.Msg(iLogL,"disabled");
            else ILogSystem.Msg(iLogL,"per-surface (%d)",iTmp);

            iTmp = vTilesets[i]->tTileset->GetColorKey();
            ILogSystem.Msg(iLogL," - Colorkey: ");
            if(iTmp != -1)
            {
                if(vTilesets[i]->tTileset->GetSurface()->format->BitsPerPixel == 8)
                {
                    ILogSystem.Msg(iLogL,"palette color number %d\n",iTmp);
                }
                else
                {
                    Uint8 r = 0, g = 0, b = 0;
                    SDL_GetRGB(iTmp,vTilesets[i]->tTileset->GetSurface()->format,&r,&g,&b);
                    ILogSystem.Msg(iLogL,"RGB(%d,%d,%d)\n",r,g,b);
                }
            }
            else ILogSystem.Msg(iLogL,"disabled\n");
        }		  
    }
    ILogSystem.Msg(iLogL,"\n");
}

//! Log verbosity control
/*! \param iLevel 0 for disabling the log output, 1 for displaying only LOG_LOW messages, 2 for displaying
LOG_LOW and LOG_NORMAL messages and 3 for displaying all levels */
void cMSTE::LogVerbosity(int iLevel)
{
    // Assume Level 0
    iLogL = iLogN = iLogH = LOG_HIGH + 1;
   
    // Level 1
    if(iLevel == 1) iLogL = LOG_LOW; 
    // Level 2
    else if(iLevel == 2)
    {
        iLogL = LOG_LOW;
        iLogN = LOG_NORMAL;
    }
    // Level 3
    else if(iLevel >= 3)
    {
        iLogL = LOG_LOW;
        iLogN = LOG_NORMAL;
        iLogH = LOG_HIGH;
    }
}

//-------------------------------------------------------------------------------
// Setup engine methods
//-------------------------------------------------------------------------------

//! Set a new level name.
/*! \param levelname Pointer to the new name. Maximum of 32 characters.*/
void cMSTE::SetName(char *levelname)
{
    // We only need an initialized engine
    if(!bTileEngine) return;
    if(levelname == NULL) return;
    szCopy(szName,levelname,sizeof(szName));
}

//! Get the level name
/*! \return Pointer to the Level name*/
char *cMSTE::GetName()
{
    // We only need an initialized engine
    if(!bTileEngine) return NULL;
    return szName;
}

//! Set the viewport.
/*! You can change the viewport at any time even while you are rendering your application.<br>
By default, it uses the whole render target surface.
\param rect Pointer to SDL_Rect with the desired viewport. If the rect width or height is 0, it will expand up to the current CRM32Pro.screen
\return 0 Failed
\return 1 Succeed*/
int cMSTE::SetViewport(SDL_Rect *rect)
{
    // We only need an initialized engine
    if(!bTileEngine) return 0;
    // Check limits
    if(sRenderTarget != NULL)
    {
        if(rect->w > sRenderTarget->w || rect->w == 0) rect->w = sRenderTarget->w;
        if(rect->h > sRenderTarget->h || rect->h == 0) rect->h = sRenderTarget->h;
    }
    memcpy(&rViewport,rect,sizeof(SDL_Rect));
    // Info
    ILogSystem.Msg(iLogN," [MSTE] Info: new viewport set at (%d,%d)-(%d,%d).\n",rect->x,rect->y,rect->x+rect->w,rect->y+rect->h);
    return 1;
}

//! Get the current viewport.
/*! \return NULL Failed
\return Any other value, a pointer to a SDL_Rect with the current viewport */
SDL_Rect *cMSTE::GetViewport()
{
    // We only need an initialized engine
    if(!bTileEngine) return NULL;
    return &rViewport;
}

//! Set the render target surface.
/*! Note it will reset the current viewport.
\param surf A pointer to desired SDL_Surface to render all layers. It usually is CRM32Pro.screen.
\return 0 Failed
\return 1 Succeed */
int cMSTE::SetRenderTarget(SDL_Surface *surf)
{
    // We only accept an initialized engine
    if(!bTileEngine) return 0;
    if(surf == NULL) return 0;
    sRenderTarget = surf;
    bTileEngine = 2;
    // Viewport by default
    rViewport.x = 0; rViewport.y = 0;
    rViewport.w = surf->w; rViewport.h = surf->h;
    // Info
    ILogSystem.Msg(iLogN," [MSTE] Info: new render target assigned. Setting a new viewport.\n");
    return 1;
}

//! Get the current render target surface.
/*! \return NULL Failed
\return Any other value, a pointer to current SDL_Surface */
SDL_Surface *cMSTE::GetRenderTarget()
{
    if(bTileEngine != 2) return NULL;
    return sRenderTarget;
}

//! Get the level modification date (last save level date)
char *cMSTE::GetModificationDate()
{
    // 1.Initial checks
    if(!bTileEngine) return NULL;
    // 2.Return
    return szModificationDate;
}

//! Get the level creation date (first save level date)
char *cMSTE::GetCreationDate()
{
    // 1.Initial checks
    if(!bTileEngine) return NULL;
    // 2.Return
    return szCreationDate;
}

//-------------------------------------------------------------------------------
// Core methods
//-------------------------------------------------------------------------------

//! Validate the mapbase against available tilesets. Internal use only.
/*! It will use the CheckMap() method for each layer map checking it against errors.<br>
It also check that all tilesets needed are loaded tagging .bTilesetReady = 0 or 1. It also can disable a layer if the tileset is not present.
\param layer Number of layer(from 0 to number of allocated layers) to be validated 
\return 0 Failed (layer is not ready and can not be enabled)
\return 1 Succeed (layer is ready to be enabled) */
int cMSTE::ValidateMapBase(int layer)
{
    int iRet = 1;

    // 1.Check we have a map
    if(vLayers[layer]->map == NULL) return 0;

    // 2.Check layer map against available and used tilesets
    // - go through all layer map tiles checking we have it on a tileset
    // - also checks the size
    for(int iY = 0; (iY < vLayers[layer]->map->GetMapHeight() && iRet == 1); iY++)
    {
        for(int iX = 0; iX < vLayers[layer]->map->GetMapWidth(); iX++)
        {
            // 2.1.Is there any tileset?
            if(vTilesets.empty())
            {
                iRet = 0;
                break;
            }			

            // 2.2.Is the cell empty?			
            unsigned int iCellValue = vLayers[layer]->map->iTilemap[iY][iX];
            if(iCellValue == 0) continue;

            // 2.3.Is the tile in our loaded tilesets?
            int iValue = CheckTile(iCellValue);
            if(iValue == -1) // No!
            {
                ILogSystem.Msg(iLogN," [MSTE] Error: tile '%ul' in layer '%s' with ID '%d' at [%d,%d] was not found in the loaded tilesets.\n",vLayers[layer]->map->iTilemap[iY][iX],vLayers[layer]->szName,layer,iX,iY);
                iRet = 0;
                break;
            }
            else // Yes!
            {
                int tx = 0, ty = 0;
                vTilesets[iValue]->tTileset->GetOffset(&tx,&ty);

                // Detect tiles greater than cell size
                if(vLayers[layer]->map->GetCellWidth() < tx)
                {
                    int iOffset = (int)ceil((float)tx / (float)vLayers[layer]->map->GetCellWidth()) - 1;
                    iOffset *= vLayers[layer]->map->GetCellWidth();
                    if(vLayers[layer]->iOffsetX < iOffset) vLayers[layer]->iOffsetX = iOffset;
                }
                if(vLayers[layer]->map->GetCellHeight() < ty)
                {
                    int iOffset = (int)ceil((float)ty / (float)vLayers[layer]->map->GetCellHeight()) - 1;
                    iOffset *= vLayers[layer]->map->GetCellHeight();
                    if(vLayers[layer]->iOffsetY < iOffset) vLayers[layer]->iOffsetY = iOffset;
                }
            }
        }
    }

    // 3.Have we failed?
    if(iRet == 0) // Yes
    {
        // Check if the layer was enabled, if so, we disabled it as the tileset is not present
        if(vLayers[layer]->iStatus != TE_LAYERFLAG_DISABLE)
        {
            vLayers[layer]->iStatus = TE_LAYERFLAG_DISABLE;					
        }
        // This layer has not all the needed tilesets
        if(vLayers[layer]->bTilesetReady == 1)
        {
            vLayers[layer]->bTilesetReady = 0;
        }
        return 0;
    }
    else // No
    {
        // This layer has all the needed tilesets
        vLayers[layer]->bTilesetReady = 1;
        return 1;
    }
}

//! Check if the tile exists on any loaded tileset
/*! The tilesets are not sorted so is a bit more effecient to keep the most used tiles in the low Global ID tilesets.
\param iTN Tile ID as contained on the layer map. A value of 0 means the cell is empty and return -1. It supports some special effects.
\return -1 Not found in loaded tilesets
\return >=0 Found and return in which it was found.*/
int	cMSTE::CheckTile(unsigned int iTN)
{
    unsigned int iTmp;
    int ret = -1;

    // Is an empty cell?
    if(iTN == 0) return -1;

    // Check for unsupported Tiled Qt special effects
    // Is better to let them pass, CTile can not render them but if exported they can be readed again by Tiled Qt
    //if(GetTileEffect(iTN) == 0xE0000000) return -1;
    //else if(GetTileEffect(iTN) == 0x20000000) return -1;
    // Remove supported effects for this check
    iTN = ResetTileEffect(iTN);

    // Look for the tile
    for(int i = 0; (i < (int)vTilesets.size() && ret == -1); i++) 
    {
        iTmp = vTilesets[i]->iGID;
        if(iTN >= iTmp && iTN < (iTmp + vTilesets[i]->iNumTiles)) ret = i;
    }
    return ret;
}

//! Try to move to new X position. Internal use only.
/*! \param layer Number of layer(from 0 to number of allocated layers)
\param newpos Selected new X position
\return 0 Succeed but no movement was done
\return 1 Succeed*/
int cMSTE::TryMoveX(int layer, float newpos)
{
    // 1.Check for new position or return
    if(vLayers[layer]->fPosX == newpos) return 0;

    // 2.Loop property in action! Basically, we allow any position (even negative).
    //   At render time it will be converted to a map cell
    if(vLayers[layer]->iStatus & TE_LAYERFLAG_LOOPX) 
    {
        vLayers[layer]->fPosX = newpos;
        return 1;
    }

    // 3.1.Negative bounds check
    if(newpos < 0) newpos = 0;
    // 3.2.Over end bound check (depends on the viewport)
    else if((newpos + rViewport.w) > (float)(vLayers[layer]->map->GetMapWidth() * vLayers[layer]->map->GetCellWidth())) 
        newpos = (float)(vLayers[layer]->map->GetMapWidth() * vLayers[layer]->map->GetCellWidth()) - rViewport.w;
    // 3.3.Check again for changes on the position
    if(vLayers[layer]->fPosX == newpos) return 0;

    // 4.Assigment new position
    vLayers[layer]->fPosX = newpos;

    // 5.Return
    return 1;
}

//! Try to move to new Y position. Internal use only.
/*! \param layer Number of layer(from 0 to number of allocated layers)
\param newpos Selected new Y position
\return 0 Succeed but no movement was done
\return 1 Succeed*/
int cMSTE::TryMoveY(int layer, float newpos)
{
    // 1.Check for new position or return
    if(vLayers[layer]->fPosY == newpos) return 0;

    // 2.Loop property in action! Basically, we allow any position (even negative).
    //   At render time it will be converted to a map cell
    if(vLayers[layer]->iStatus & TE_LAYERFLAG_LOOPY) 
    {
        vLayers[layer]->fPosY = newpos;
        return 1;
    }

    // 3.1.Negative bounds check
    if(newpos < 0) newpos = 0;
    // 3.2.Over end bound check (depends on the viewport)
    else if((newpos + rViewport.h) > (float)(vLayers[layer]->map->GetMapHeight() * vLayers[layer]->map->GetCellHeight())) 
        newpos = (float)(vLayers[layer]->map->GetMapHeight() * vLayers[layer]->map->GetCellHeight()) - rViewport.h;
    // 3.3.Check again for changes on the position
    if(vLayers[layer]->fPosY == newpos) return 0;

    // 4.Assigment new position
    vLayers[layer]->fPosY = newpos;

    // 5.Return
    return 1;
}

//! Create a mapdata instance of the given type. 
/*! It is our mapdata types factory.
\param type string with the mapdata type to be created
\return NULL Selected mapdata type is not registered on RegisteredMapData[]
\return Any other value, a pointer to the created map*/
cMSTE_MapBase *cMSTE::CreateMapDataInstance(char *type, int y, int x)
{
    cMSTE_MapBase *myMap = NULL;

    // Look for the desired map type
    for(int i = 0; i < TE_REGISTEREDMAPDATA_MAX; i++)
    {
        if(RegisteredMapData[i] != NULL)
        {
            myMap = RegisteredMapData[i](y,x);
            if(!strcmp(type,myMap->GetMapType())) return myMap;
            delete myMap;
            myMap = NULL;
        }
    }
    return myMap;
}

//! Register a new mapdata type in our factory
/* \param A pointer to the static method createInstance(int,int) of a new mapdata type
\return 0 Failed
\return 1 Succeed*/
int cMSTE::RegisterMapDataHandler(cMSTE_MapBase *(*createInstance)(int y, int x))
{
    // Look for an empty slot. Avoid the [0] as it busy with cMSTE_MapBase::createInstance
    for(int i = 1; i < TE_REGISTEREDMAPDATA_MAX; i++)
    {
        if(RegisteredMapData[i] == NULL)
        {
            RegisteredMapData[i] = createInstance;
            return 1;
        }
    }
    return 0;
}   

//-------------------------------------------------------------------------------
// Layer properties methods
//-------------------------------------------------------------------------------

//! Set the layer flags
/*! \param layer Number of layer(from 0 to number of allocated layers)
\param iFlags See TE_LAYERFLAG_xxx defines for further information. They can be OR'ed
\param iState <=0 for disabling the given bFlags or >1 for enabling it. With TE_LAYERFLAG_DISABLE it does not have sense.
\return 0 Failed
\return 1 Succeed*/
int cMSTE::SetLayerFlags(int layer,Uint32 iFlags, char iState)
{
    st_layers *lP = NULL;

    // 1.Initial checks
    if(bTileEngine != 2) return 0;
    if(layer < 0 || layer >= (int)vLayers.size()) return 0;
    if(iState < 0) iState = 0; if(iState > 1) iState = 1;
    if(iFlags < 0) iFlags = 0;
    lP = vLayers[layer];

    // 2.Deactivating a layer
    if((lP->iStatus != TE_LAYERFLAG_DISABLE) && (iFlags == TE_LAYERFLAG_DISABLE)) 
    {		
        ILogSystem.Msg(iLogN," [MSTE] Info: layer '%s' with ID '%d' disabled.\n",lP->szName,layer);
        lP->iStatus = TE_LAYERFLAG_DISABLE;
        return 1;
    }

    // 3.1.Check for TE_LAYERFLAG_LOOPX
    if(iFlags & TE_LAYERFLAG_LOOPX)
    {
        // Unsetting the flag
        if((iState == 0) && (lP->iStatus & TE_LAYERFLAG_LOOPX))
        {
            lP->iStatus = lP->iStatus ^ TE_LAYERFLAG_LOOPX;
            ILogSystem.Msg(iLogN," [MSTE] Info: unset LoopX flag for layer '%s' with ID '%d'.\n",lP->szName,layer);
        }
        // Setting the flag
        else if((iState == 1) && !(lP->iStatus & TE_LAYERFLAG_LOOPX))
        {
            lP->iStatus = lP->iStatus | TE_LAYERFLAG_LOOPX;
            ILogSystem.Msg(iLogN," [MSTE] Info: set LoopX flag for layer '%s' with ID '%d'.\n",lP->szName,layer);
        }
    }

    // 3.2.Check for TE_LAYERFLAG_LOOPY
    if(iFlags & TE_LAYERFLAG_LOOPY)
    {
        // Unsetting the flag
        if((iState == 0) && (lP->iStatus & TE_LAYERFLAG_LOOPY))
        {
            lP->iStatus = lP->iStatus ^ TE_LAYERFLAG_LOOPY;
            ILogSystem.Msg(iLogN," [MSTE] Info: unset LoopY flag for layer '%s' with ID '%d'.\n",lP->szName,layer);
        }
        // Setting the flag
        else if((iState == 1) && !(lP->iStatus & TE_LAYERFLAG_LOOPY))
        {
            lP->iStatus = lP->iStatus | TE_LAYERFLAG_LOOPY;
            ILogSystem.Msg(iLogN," [MSTE] Info: set LoopY flag for layer '%s' with ID '%d'.\n",lP->szName,layer);
        }
    }

    // 3.3.Check for TE_LAYERFLAG_AUTOSCROLL
    if(iFlags & TE_LAYERFLAG_AUTOSCROLL)
    {
        // Unsetting the flag
        if((iState == 0) && (lP->iStatus & TE_LAYERFLAG_AUTOSCROLL))
        {
            lP->iStatus = lP->iStatus ^ TE_LAYERFLAG_AUTOSCROLL;
            ILogSystem.Msg(iLogN," [MSTE] Info: unset AutoScroll flag for layer '%s' with ID '%d'.\n",lP->szName,layer);
        }
        // Setting the flag
        else if((iState == 1) && !(lP->iStatus & TE_LAYERFLAG_AUTOSCROLL))
        {
            lP->iStatus = lP->iStatus | TE_LAYERFLAG_AUTOSCROLL;
            ILogSystem.Msg(iLogN," [MSTE] Info: set AutoScroll flag for layer '%s' with ID '%d'.\n",lP->szName,layer);
        }
    }

    // 3.4.Check for TE_LAYERFLAG_UPDATE
    if(iFlags & TE_LAYERFLAG_UPDATE)
    {
        // Unsetting the flag
        if((iState == 0) && (lP->iStatus & TE_LAYERFLAG_UPDATE))
        {
            lP->iStatus = lP->iStatus ^ TE_LAYERFLAG_UPDATE;
            ILogSystem.Msg(iLogN," [MSTE] Info: unset Update flag for layer '%s' with ID '%d'.\n",lP->szName,layer);
        }
        // Setting the flag (or trying to...)
        else if((iState == 1) && !(lP->iStatus & TE_LAYERFLAG_UPDATE))
        {
            if(lP->map == NULL)
            {
                ILogSystem.Msg(iLogN," [MSTE] Warning: attempt to set Update flag for layer '%s' with ID '%d' failed (Mapdata was not set)\n",vLayers[layer]->szName,layer);
            }			
            else if(lP->bTilesetReady == 0)
            {
                ILogSystem.Msg(iLogN," [MSTE] Warning: attempt to set Update flag for layer '%s' with ID '%d' failed (Needed tilesets are not loaded)\n",vLayers[layer]->szName,layer);
            }
            else
            {
                lP->iStatus = lP->iStatus | TE_LAYERFLAG_UPDATE;
                ILogSystem.Msg(iLogN," [MSTE] Info: set Update flag for layer '%s' with ID '%d'.\n",lP->szName,layer);
            }
        }
    }

    // 3.5.Check for TE_LAYERFLAG_UPDATECELLS
    if(iFlags & TE_LAYERFLAG_UPDATECELLS)
    {
        // Unsetting the flag
        if((iState == 0) && (lP->iStatus & TE_LAYERFLAG_UPDATECELLS))
        {
            lP->iStatus = lP->iStatus ^ TE_LAYERFLAG_UPDATECELLS;
            ILogSystem.Msg(iLogN," [MSTE] Info: unset UpdateCells flag for layer '%s' with ID '%d'.\n",lP->szName,layer);
        }
        // Setting the flag (or trying to...)
        else if((iState == 1) && !(lP->iStatus & TE_LAYERFLAG_UPDATECELLS))
        {
            if(lP->map == NULL)
            {
                ILogSystem.Msg(iLogN," [MSTE] Warning: attempt to set UpdateCells flag for layer '%s' with ID '%d' failed (Mapdata was not set)\n",vLayers[layer]->szName,layer);
            }			
            else if(lP->bTilesetReady == 0)
            {
                ILogSystem.Msg(iLogN," [MSTE] Warning: attempt to set UpdateCells flag for layer '%s' with ID '%d' failed (Needed tilesets are not loaded)\n",vLayers[layer]->szName,layer);
            }
            else
            {
                lP->iStatus = lP->iStatus | TE_LAYERFLAG_UPDATECELLS;
                ILogSystem.Msg(iLogN," [MSTE] Info: set UpdateCells flag for layer '%s' with ID '%d'.\n",lP->szName,layer);
            }
        }
    }

    // 3.6.Check for TE_LAYERFLAG_RENDER
    if(iFlags & TE_LAYERFLAG_RENDER)
    {
        // Unsetting the flag
        if((iState == 0) && (lP->iStatus & TE_LAYERFLAG_RENDER))
        {
            lP->iStatus = lP->iStatus ^ TE_LAYERFLAG_RENDER;
            ILogSystem.Msg(iLogN," [MSTE] Info: unset Render flag for layer '%s' with ID '%d'.\n",lP->szName,layer);
        }
        // Setting the flag (or trying to...)
        else if((iState == 1) && !(lP->iStatus & TE_LAYERFLAG_RENDER))
        {
            if(lP->map == NULL)
            {
                ILogSystem.Msg(iLogN," [MSTE] Warning: attempt to set Render flag for layer '%s' with ID '%d' failed (Mapdata was not set)\n",vLayers[layer]->szName,layer);
            }
            else if(lP->bTilesetReady == 0)
            {
                ILogSystem.Msg(iLogN," [MSTE] Warning: attempt to set Render flag for layer '%s' with ID '%d' failed (Needed tilesets are not loaded)\n",vLayers[layer]->szName,layer);
            }
            else
            {
                lP->iStatus = lP->iStatus | TE_LAYERFLAG_RENDER;
                ILogSystem.Msg(iLogN," [MSTE] Info: set Render flag for layer '%s' with ID '%d'.\n",lP->szName,layer);
            }
        }
    }

    // 3.7.Check for TE_LAYERFLAG_SMOOTHSCROLL
    if(iFlags & TE_LAYERFLAG_SMOOTHSCROLL)
    {
        // Unsetting the flag
        if((iState == 0) && (lP->iStatus & TE_LAYERFLAG_SMOOTHSCROLL))
        {
            lP->iStatus = lP->iStatus ^ TE_LAYERFLAG_SMOOTHSCROLL;
            ILogSystem.Msg(iLogN," [MSTE] Info: unset SmoothScroll flag for layer '%s' with ID '%d'.\n",lP->szName,layer);
        }
        // Setting the flag (or trying to...)
        else if((iState == 1) && !(lP->iStatus & TE_LAYERFLAG_SMOOTHSCROLL))
        {
            if(lP->map == NULL)
            {
                ILogSystem.Msg(iLogN," [MSTE] Warning: attempt to set SmoothScroll flag for layer '%s' with ID '%d' failed (Mapdata was not set)\n",vLayers[layer]->szName,layer);
            }
            else if(lP->bTilesetReady == 0)
            {
                ILogSystem.Msg(iLogN," [MSTE] Warning: attempt to set SmoothScroll flag for layer '%s' with ID '%d' failed (Needed tilesets are not loaded)\n",vLayers[layer]->szName,layer);
            }
            else
            {
                lP->iStatus = lP->iStatus | TE_LAYERFLAG_SMOOTHSCROLL;
                ILogSystem.Msg(iLogN," [MSTE] Info: set SmoothScroll flag for layer '%s' with ID '%d'.\n",lP->szName,layer);
            }
        }
    }

    // 3.8.Check for TE_LAYERFLAG_SHOWGRID
    if(iFlags & TE_LAYERFLAG_SHOWGRID)
    {
        // Unsetting the flag
        if((iState == 0) && (lP->iStatus & TE_LAYERFLAG_SHOWGRID))
        {
            lP->iStatus = lP->iStatus ^ TE_LAYERFLAG_SHOWGRID;
            ILogSystem.Msg(iLogN," [MSTE] Info: unset ShowGrid flag for layer '%s' with ID '%d'.\n",lP->szName,layer);
        }
        // Setting the flag (or trying to...)
        else if((iState == 1) && !(lP->iStatus & TE_LAYERFLAG_SHOWGRID))
        {
            lP->iStatus = lP->iStatus | TE_LAYERFLAG_SHOWGRID;
            ILogSystem.Msg(iLogN," [MSTE] Info: set ShowGrid flag for layer '%s' with ID '%d'.\n",lP->szName,layer);
        }
    }

    // 3.9.Check for TE_LAYERFLAG_PAUSE
    if(iFlags & TE_LAYERFLAG_PAUSE)
    {
        // Unsetting the flag
        if((iState == 0) && (lP->iStatus & TE_LAYERFLAG_PAUSE))
        {
            lP->iStatus = lP->iStatus ^ TE_LAYERFLAG_PAUSE;
            ILogSystem.Msg(iLogN," [MSTE] Info: unset Pause flag for layer '%s' with ID '%d'.\n",lP->szName,layer);
        }
        // Setting the flag (or trying to...)
        else if((iState == 1) && !(lP->iStatus & TE_LAYERFLAG_PAUSE))
        {
            lP->iStatus = lP->iStatus | TE_LAYERFLAG_PAUSE;
            ILogSystem.Msg(iLogN," [MSTE] Info: set Pause flag for layer '%s' with ID '%d'.\n",lP->szName,layer);
        }
    }

    // Return
    return 1;
}


//! Get the layer flags
/*! \param layer Number of layer(from 0 to number of allocated layers)
\return -1 Failed
\return Any other number, the status flag. See TE_LAYERFLAG_xxx defines for further information */
Uint32 cMSTE::GetLayerFlags(int layer)
{
    // We only need an initialized engine
    if(!bTileEngine) return -1;
    if(layer < 0 || layer >= (int)vLayers.size()) return -1;
    return vLayers[layer]->iStatus;
}


//! Set layer resource control flag
/*! \param layer Selected layer to set the resource control flag (from 0 to number of allocated layers).
\param rescontrol See TE_RESCONTROL_xxx defines.
\return 0 Failed
\return 1 Succeed*/
int cMSTE::SetLayerResControl(int layer, char rescontrol)
{
    // We only need an initialized engine
    if(!bTileEngine) return 0;
    if(layer < 0 || layer >= (int)vLayers.size()) return 0;
    if(rescontrol < 0) rescontrol = 0;
    if(rescontrol > 1) rescontrol = 1;
    if(vLayers[layer]->bResControl == TE_RESCONTROL_INTERNAL) return 0;
    if(vLayers[layer]->bResControl != rescontrol)
    {
        ILogSystem.Msg(iLogN," [MSTE] Info: layer '%s' with ID '%d' changed to ",vLayers[layer]->szName,layer);
        if(TE_RESCONTROL_AUTO == 0) ILogSystem.Msg(iLogN,"auto resource control mode.\n");
        else ILogSystem.Msg(iLogN,"user resource control mode.\n");
        vLayers[layer]->bResControl = rescontrol;	
    }	
    return 1;	
}

//! Get layer resource control flag
/*! \param layer Selected layer to get the current resource control flag (from 0 to number of allocated layers)
\return -1 Failed
\return Any other value, the current flag. See TE_RESCONTROL_xxx defines. */
char cMSTE::GetLayerResControl(int layer)
{
    // We only need an initialized engine
    if(!bTileEngine) return -1;
    if(layer < 0 || layer >= (int)vLayers.size()) return -1;
    return vLayers[layer]->bResControl;
}

//! Set layer alpha value
/*! \param layer Selected layer to set the alpha value (from 0 to number of allocated layers).
\param alpha Alpha value from 0(transparent) to 255(opaque)
\return 0 Failed
\return 1 Succeed*/
int cMSTE::SetLayerAlpha(int layer, int alpha)
{
    // We only need an initialized engine
    if(!bTileEngine) return 0;
    if(layer < 0 || layer >= (int)vLayers.size()) return 0;
    if(alpha < 0) alpha = 0;
    if(alpha > 255) alpha = 255;
    ILogSystem.Msg(iLogN," [MSTE] Info: layer '%s' with ID '%d' changed alpha value to '%d'\n",vLayers[layer]->szName,layer,alpha);
    vLayers[layer]->iAlpha = alpha;
    return 1;
}

//! Get layer alpha value
/*! \param layer Selected layer to get the alpha value (from 0 to number of allocated layers)
\return -1 Failed
\return Any other value, the current alpha value from 0(transparent) to 255(opaque) */
int cMSTE::GetLayerAlpha(int layer)
{
    // We only need an initialized engine
    if(!bTileEngine) return -1;
    if(layer < 0 || layer >= (int)vLayers.size()) return -1;
    return vLayers[layer]->iAlpha;
}

//! Set the layer parallax ratio
/*! This is the way to change the real speed of any layer without to affect the rest ones.<br>
\param layer Selected layer to set the parallax ratio (from 0 to number of allocated layers). 
\param pratio Parallax ratio value. With 0, this layer is independent of the others. Any other value and the layer is linked. 
If any of the linked layers move, it will affect to the others using this parallax ratio as base.
\return 0 Failed
\return 1 Succeed */
int cMSTE::SetLayerParallaxRatio(int layer, float pratio)
{
    // We only need an initialized engine
    if(!bTileEngine) return 0;
    if(layer < 0 || layer >= (int)vLayers.size()) return 0;
    if(vLayers[layer]->fParallaxRatio != pratio)
        ILogSystem.Msg(iLogN," [MSTE] Info: layer '%s' with ID '%d' changed parallax ratio to '%.2f' value.\n",vLayers[layer]->szName,layer,pratio);
    vLayers[layer]->fParallaxRatio = pratio;
    return 1;
}

//! Get current layer parallax ratio
/*! \param layer Selected layer to get the parallax ratio (from 0 to number of allocated layers).
\return Current parallax ratio */
float cMSTE::GetLayerParallaxRatio(int layer)
{
    // We only need an initialized engine
    if(!bTileEngine) return 0.0f;
    if(layer < 0 || layer >= (int)vLayers.size()) return 0.0f;
    return vLayers[layer]->fParallaxRatio;
}


//! Set the grid color for the base mapdata
/*! It is used for debugging purposes
\param layer Selected layer to set the grid color (from 0 to number of allocated layers). 
\param iColor packed unsigned int with the color value. 
\return 0 Failed
\return 1 Succeed */
int cMSTE::SetLayerGridColor(int layer, Uint32 iColor)
{
    // We only need an initialized engine
    if(!bTileEngine) return 0;
    if(layer < 0 || layer >= (int)vLayers.size()) return 0;
    vLayers[layer]->iGridColor = iColor;
    return 1;
}

//! Set the layer absolute position (in pixels).
/*! Changing the position of a layer will modify the position for the rest of linked ones according
to each parallax ratio value and holding the "global" position.<br>
Remember that a layer with a parallax ratio of 0.0 is independent from the others.<br>
Requires a valid base mapdata.<br>
There are some special flags:<ul>
<li>TE_KEEP_VALUE: do not modify this axis value </li>
<li>TE_LAYERPOSITION_BEGINNING: move this layer to the beginning. In this case, the smoothing system is not used if presents.</li>
<li>TE_LAYERPOSITION_MIDDLE will: move this layer to the middle. In this case, the smoothing system is not used if presents.</li>
<li>TE_LAYERPOSITION_END: move this layer to the end. In this case, the smoothing system is not used if presents.<li>
<li>TE_LAYERPOSITION_CENTER: added to a position for centering it depending on current viewport</li></ul>
\param layer Selected layer to try to set a new position (from 0 to number of allocated layers). 
\param xpos Any supported flag or other value to try to move the current position on X axis. 
\param ypos Any supported flag or other value to try to move the current position on Y axis.
\return -1 Failed
\return 0 Succeed but no change on the position done
\return 1 Succeed and the position changed*/
int cMSTE::SetLayerPosition(int layer, float xpos, float ypos)
{
    int i, tx, ty, iFlagSpecial = 0, iRet = -1;
    float fTmp;

    // 1.We need an initialized engine
    if(!bTileEngine) return 0;
    if(layer < 0 || layer >= (int)vLayers.size()) return iRet;
    // 1.1.And..are the map and tileset enabled?
    if(vLayers[layer]->map == NULL) return iRet;
    if(vLayers[layer]->bTilesetReady == 0) return iRet;
    // 1.2.Is the layer paused?
    iRet = 0;
    if(vLayers[layer]->iStatus & TE_LAYERFLAG_PAUSE) return iRet;
    // 1.3.Get cell sizes
    tx = vLayers[layer]->map->GetCellWidth();
    ty = vLayers[layer]->map->GetCellHeight();
    
    // 2.X axis position
    if(xpos != TE_KEEP_VALUE)
    {
        // 2.1.Check for special flags
        if(xpos > TE_LAYERPOSITION_CENTER)
        {
            xpos -= TE_LAYERPOSITION_CENTER; // Remove the flag
            xpos -= ((float)rViewport.w / 2);
        }   
        // 2.2.Special cases. Disable next smooth update (if applies)
        if(xpos == TE_LAYERPOSITION_BEGIN) 
        { 
            iFlagSpecial = 1; 
            xpos = 0; 
        }
        else if(xpos == TE_LAYERPOSITION_MIDDLE) 
        {
            iFlagSpecial = 1;
            xpos = ((float)(vLayers[layer]->map->GetMapWidth() * tx) / 2) - (rViewport.w / 2);
        }
        else if(xpos == TE_LAYERPOSITION_END)
        {
            iFlagSpecial = 1;
            xpos = (float)(vLayers[layer]->map->GetMapWidth() * tx) - rViewport.w;
        }

        // 2.3.If we could move it, we try to move the rest layers. With ratio=0, independent layer.
        if(TryMoveX(layer, xpos))
        {
            iRet = 1;
            ILogSystem.Msg(iLogH," [MSTE] Info: layer '%s' with ID '%d' changed X position to '%.2f' value(Request)\n",vLayers[layer]->szName,layer,xpos);
            if(iFlagSpecial) vLayers[layer]->iDisableNextSmooth = 1;
            for(i = 0; ((i < (int)vLayers.size()) && (vLayers[layer]->fParallaxRatio != 0)); i++)
            {
                // Loop though rest of layers. If any has ratio=0, is an independent layer and we dont modify it.
                if((i != layer) && (vLayers[i]->fParallaxRatio != 0))
                {				
                    fTmp = (vLayers[layer]->fPosX / vLayers[layer]->fParallaxRatio) * vLayers[i]->fParallaxRatio;
                    if(TryMoveX(i,fTmp))
                    {
                        ILogSystem.Msg(iLogH," [MSTE] Info: layer '%s' with ID '%d' changed X position to '%.2f' value(Cascade)\n",vLayers[i]->szName,i,fTmp);	
                        if(iFlagSpecial) vLayers[i]->iDisableNextSmooth = 1;
                    }
                }
            }
        }
    }		

    // 3.Y position
    iFlagSpecial = 0;
    if(ypos != TE_KEEP_VALUE)
    {
        // 3.1.Check for special OR'ed flags
        if(ypos > TE_LAYERPOSITION_CENTER)
        {
            ypos -= TE_LAYERPOSITION_CENTER; // Remove the flag
            ypos -= ((float)rViewport.h / 2);
        }   
        // 3.1.Special cases. Disable next smooth update (if applies)
        if(ypos == TE_LAYERPOSITION_BEGIN) 
        {
            iFlagSpecial = 1;
            ypos = 0;
        }
        else if(ypos == TE_LAYERPOSITION_MIDDLE) 
        {
            iFlagSpecial = 1;
            ypos = ((float)(vLayers[layer]->map->GetMapHeight() * ty) / 2) - (rViewport.h / 2);
        }
        else if(ypos == TE_LAYERPOSITION_END)
        {
            iFlagSpecial = 1;
            ypos = (float)(vLayers[layer]->map->GetMapHeight() * ty) - rViewport.h;
        }

        // 3.2.If we could move it, we try to move the rest layers. With ratio=0, independent layer.
        if(TryMoveY(layer, ypos))
        {
            iRet = 1;
            ILogSystem.Msg(iLogH," [MSTE] Info: layer '%s' with ID '%d' changed Y position to '%.2f' value(Request)\n",vLayers[layer]->szName,layer,ypos);
            if(iFlagSpecial) vLayers[layer]->iDisableNextSmooth = 1;
            for(i = 0; ((i < (int)vLayers.size()) && (vLayers[layer]->fParallaxRatio != 0)); i++)
            {
                // Loop though rest of layers. If any has ratio=0, is an independent layer and we dont modify it.
                if((i != layer) && (vLayers[i]->fParallaxRatio != 0))
                {				
                    fTmp = (vLayers[layer]->fPosY / vLayers[layer]->fParallaxRatio) * vLayers[i]->fParallaxRatio;
                    if(TryMoveY(i,fTmp))
                    {
                        ILogSystem.Msg(iLogH," [MSTE] Info: layer '%s' with ID '%d' changed Y position to '%.2f' value(Cascade)\n",vLayers[i]->szName,i,fTmp);			
                        if(iFlagSpecial) vLayers[i]->iDisableNextSmooth = 1;
                    }
                }
            }
        }
    }		

    return iRet;
}

//! Get the current layer position (in pixels).
/*! \param layer Selected layer to get its current position (from 0 to number of allocated layers).
\param xpos Pointer to float var to store the position on X axis. NULL if you dont want to retrieve it.
\param ypos Pointer to float var to store the position on Y axis. NULL if you dont want to retrieve it.
\return 0 Failed
\return 1 Succeed */
int cMSTE::GetLayerPosition(int layer, float & xpos, float & ypos)
{
    // We only need an initialized engine
    if(!bTileEngine) return -1;
    if(layer < 0 || layer >= (int)vLayers.size()) return -1;
    if(xpos != NULL) xpos = vLayers[layer]->fPosX;
    if(ypos != NULL) ypos = vLayers[layer]->fPosY;
    return 1;
}

//! Set the layer AutoScroll speed (in pixels per second).
/*! Changing the AutoScroll speed of a layer will modify the same value for the rest of linked ones according
to each parallax ratio value and holding the "global" position.<br> 
\param layer Selected layer to try to set the autoscroll speed (from 0 to number of allocated layers).
\param xspeed TE_KEEP_VALUE will keep the current X speed. Any other value, will try to set it as current X speed. 
\param yspeed TE_KEEP_VALUE will keep the current Y speed. Any other value, will try to set it as current Y speed. 
\return 0 Failed
\return 1 Succeed */
int cMSTE::SetLayerAutoScroll(int layer, float xspeed, float yspeed)
{
    int i;
    float fTmp;

    // We only need an initialized engine
    if(!bTileEngine) return 0;
    if(layer < 0 || layer >= (int)vLayers.size()) return 0;
    // X speed
    if(xspeed != TE_KEEP_VALUE) 
    {
        if(vLayers[layer]->fSpeedX != xspeed)
            ILogSystem.Msg(iLogN," [MSTE] Info: layer '%s' with ID '%d' changed X speed to '%.2f' value(Requested)\n",vLayers[layer]->szName,layer,xspeed);
        vLayers[layer]->fSpeedX = xspeed;
        // Changing speed of others layers. With ratio=0, independent layer.
        for(i = 0; ((i < (int)vLayers.size()) && (vLayers[layer]->fParallaxRatio != 0)); i++)
        {
            // Loop though rest of layers. If any has ratio=0, is an independent layer and we dont modify it.
            if((i != layer) && (vLayers[i]->fParallaxRatio != 0))
            {		
                fTmp = (xspeed / vLayers[layer]->fParallaxRatio) * vLayers[i]->fParallaxRatio;
                if(vLayers[i]->fSpeedX != fTmp)
                {					
                    ILogSystem.Msg(iLogN," [MSTE] Info: layer '%s' with ID '%d' changed X speed to '%.2f' value(Cascade)\n",vLayers[i]->szName,i,fTmp);
                    vLayers[i]->fSpeedX = fTmp;
                }
            }
        }
    }
    // Y speed
    if(yspeed != TE_KEEP_VALUE) 
    {
        if(vLayers[layer]->fSpeedY != yspeed)
            ILogSystem.Msg(iLogN," [MSTE] Info: layer '%s' with ID '%d' changed Y speed to '%.2f' value(Requested)\n",vLayers[layer]->szName,layer,yspeed);
        vLayers[layer]->fSpeedY = yspeed;
        /// Changing speed of others layers. With ratio=0, independent layer.
        for(i = 0; ((i < (int)vLayers.size()) && (vLayers[layer]->fParallaxRatio != 0)); i++)
        {
            // Loop though rest of layers. If any has ratio=0, is an independent layer and we dont modify it.
            if((i != layer) && (vLayers[i]->fParallaxRatio != 0))
            {	
                fTmp = (yspeed / vLayers[layer]->fParallaxRatio) * vLayers[i]->fParallaxRatio;
                if(vLayers[i]->fSpeedY != fTmp)
                {					
                    ILogSystem.Msg(iLogN," [MSTE] Info: layer '%s' with ID '%d' changed Y speed to '%.2f' value(Cascade)\n",vLayers[i]->szName,i,fTmp);
                    vLayers[i]->fSpeedY = fTmp;
                }
            }
        }

    }
    return 1;
}

//! Get the layer AutoScroll speed (in pixels per seconds).
/*! \param layer Selected layer to get its current AutoScroll speed (from 0 to number of allocated layers).
\param xspeed Pointer to float var to store the speed on X axis. NULL if you dont want to retrieve it.
\param yspeed Pointer to float var to store the speed on Y axis. NULL if you dont want to retrieve it.
\return 0 Failed
\return 1 Succeed */
int cMSTE::GetLayerAutoScroll(int layer, float & xspeed, float & yspeed)
{
    // We only need an initialized engine
    if(!bTileEngine) return -1;
    if(layer < 0 || layer >= (int)vLayers.size()) return -1;
    if(xspeed != NULL) xspeed = vLayers[layer]->fSpeedX;
    if(yspeed != NULL) yspeed = vLayers[layer]->fSpeedY;
    return 1;
}

//! Assign a mapdata to the given layer
/*! If it success loading the mapdata, it'll call ValidateMapBase().
<br>If the mapdata is already set, it will remove it following resource control policies (see TE_RESCONTROL_xxx)
\param layer Selected layer to try to set a new mapdata (from 0 to number of allocated layers)
\param map Pointer to a valid mapdata: cMSTE_MapBase or a derivated one. NULL to try to remove the existing mapdata.
\return 0 Failed
\return 1 Succeed*/
int cMSTE::SetLayerMapData(int layer, cMSTE_MapBase *map)
{
    int bLS = 0;

    // 1.Initial checks
    if(!bTileEngine) return 0;
    if(layer < 0 || layer >= (int)vLayers.size()) return 0;

    // 2.Is it a removal attempt?
    if(map == NULL)
    {
        if(vLayers[layer]->map != NULL)
        {
            // Check if the layer was enabled, if so, we disabled it as the mapdata is being removed
            if(vLayers[layer]->iStatus != TE_LAYERFLAG_DISABLE)
            {
                vLayers[layer]->iStatus = TE_LAYERFLAG_DISABLE;					
                bLS = 1; // used for informing about the layer disabled
            }

            // Check if the layers was ready
            if(vLayers[layer]->bTilesetReady == 1)
            {
                vLayers[layer]->bTilesetReady = 0;
            }

            // Print info
            ILogSystem.Msg(iLogN," [MSTE] Info: mapdata for layer '%s' with ID '%d' unloaded:\n",vLayers[layer]->szName,layer);
            vLayers[layer]->map->PrintInfo(-1,iLogN);
            ILogSystem.Msg(iLogN,"         . layer is not ready %s\n",(bLS)?"and was disabled":"");

            // With auto or internal flag, we are the manager of the current tileset
            if(vLayers[layer]->bResControl >= TE_RESCONTROL_AUTO) delete vLayers[layer]->map;

            // Remove mapdata
            vLayers[layer]->map = NULL;

            // Mapdata removed
            return 1;
        }
        else return 0;
    }

    // 3.Check the new mapdata
    if(map->CheckMap() == 0) return 0;

    // 4.New mapdata assigment, was there a previous map assigned?
    if((vLayers[layer]->map != map) && (vLayers[layer]->map != NULL))
    {
        // With auto or internal flag, we are the manager of the current tileset
        if(vLayers[layer]->bResControl >= TE_RESCONTROL_AUTO) delete vLayers[layer]->map;
    }
    bLS = vLayers[layer]->bTilesetReady; // If the new mapdata uses non-loaded tileset.. check it 
    vLayers[layer]->map = map;

    // 5.Validate the map
    int iR = ValidateMapBase(layer);
    ILogSystem.Msg(iLogN," [MSTE] Info: mapdata for layer '%s' with ID '%d' loaded:\n",vLayers[layer]->szName,layer);
    vLayers[layer]->map->PrintInfo(-1,iLogN);
    if(iR == 1) 
        ILogSystem.Msg(iLogN,"         . Layer is ready\n");	
    else
    {
        // Check if this layer was disabled
        if((bLS == 1) && (vLayers[layer]->bTilesetReady == 0))
            ILogSystem.Msg(iLogN,"         . Layer was disabled as it is not ready\n");	
    }

    // 6.Return
    return 1;
}

//! Get current mapdata on the layer
/*! \param layer Selected layer to try to get its mapdata (from 0 to number of allocated layers)
\return NULL Failed
\return Any other value, a pointer to the cMSTE_MapBase or a derivated one*/
cMSTE_MapBase *cMSTE::GetLayerMapData(int layer)
{
    // We only need an initialized engine
    if(!bTileEngine) return NULL;
    if(layer < 0 || layer >= (int)vLayers.size()) return NULL;
    return vLayers[layer]->map;
}

//! Set an event function(OnUpdateEnd) to be called when the engine finishes to update the layer
/*! Use NULL if you want to remove a previous OnUpdateEnd event function.<br>
\param layer Number of layer to try to set your OnUpdateEnd event function (from 0 to number of allocated layers). 
\param myfunc pointer to your custom event function.
\return 0 Failed
\return 1 Succeed */
int cMSTE::SetLayerOnUpdateEnd(int layer, TE_OnLayerEnd myfunc)
{
    // We only need an initialized engine
    if(!bTileEngine) return 0;
    if(layer < 0 || layer >= (int)vLayers.size()) return 0;
    vLayers[layer]->OnUpdateEnd = myfunc;
    return 1;
}

//! Get current OnUpdateEnd event function if any
/*!\param layer Selected layer to get current OnUpdateEnd event function (from 0 to number of allocated layers). 
\return NULL There is not any event function set.
\return Any other value, a pointer to current event function. */
cMSTE::TE_OnLayerEnd *cMSTE::GetLayerOnUpdateEnd(int layer)
{
    // We only need an initialized engine
    if(!bTileEngine) return NULL;
    if(layer < 0 || layer >= (int)vLayers.size()) return NULL;
    return &vLayers[layer]->OnUpdateEnd;
}

//! Set an event function(OnRenderEnd) to be called when the engine finishes to render the layer
/*! Use NULL if you want to remove a previous OnRenderEnd event function.<br>
\param layer Number of layer to try to set your OnRenderEnd event function (from 0 to number of allocated layers). 
\param myfunc pointer to your custom event function.
\return 0 Failed
\return 1 Succeed */
int cMSTE::SetLayerOnRenderEnd(int layer, TE_OnLayerEnd myfunc)
{
    // We only need an initialized engine
    if(!bTileEngine) return 0;
    if(layer < 0 || layer >= (int)vLayers.size()) return 0;
    vLayers[layer]->OnRenderEnd = myfunc;
    return 1;
}

//! Get current OnRenderEnd event function if any
/*!\param layer Selected layer to get current OnRenderEnd event function (from 0 to number of allocated layers). 
\return NULL There is not any event function set.
\return Any other value, a pointer to current event function. */
cMSTE::TE_OnLayerEnd *cMSTE::GetLayerOnRenderEnd(int layer)
{
    // We only need an initialized engine
    if(!bTileEngine) return NULL;
    if(layer < 0 || layer >= (int)vLayers.size()) return NULL;
    return &vLayers[layer]->OnRenderEnd;
}

//! Add a custom mapdata
/*! There are three working ways: <ul>
<li>selected layer has not a mapdata assigned: we will call SetLayerMapData()</li>
<li>selected layer has a mapdata: it will be migrated to the new custom mapdata.<br>
If you want to use the new custom mapdata, you have first to remove the previous mapdata (SetLayerMapdata(layer,NULL))</li>
<li>selected layer has a custom mapdata: we do nothing as only support one custom mapadata per layer</li></ul>
Note that the added custom mapdata will use the resource control flag which currently is set on this layer.<br>
\param layer Selected layer to try to add the custom mapdata  (from 0 to number of allocated layers).
\param map Pointer to the custom mapdata to be added.
\return 0 Failed
\return 1 Succeed*/
int cMSTE::AddCustomMapData(int layer, cMSTE_MapBase *map)
{
    // 1.Initial checks
    if(!bTileEngine) return 0;
    if(layer < 0 || layer >= (int)vLayers.size()) return 0;
    if(map == NULL) return 0;
    // 1.Msg
    ILogSystem.Msg(iLogN,"\n [MSTE] Info: trying to add a custom mapdata type '%s' to layer '%s' with ID '%d'...\n",map->GetMapType(),vLayers[layer]->szName,layer);
    // 1.1.If there is not a map, we use SetLayerMapData()
    if(vLayers[layer]->map == NULL) return SetLayerMapData(layer,map);
    // 1.2.Do not accept a base as the type
    if(!strcmp("base",map->GetMapType()))
    {
        ILogSystem.Msg(iLogN," [MSTE] Warning: wrong mapdata type signature found.\n\n");
        return 0;
    }
    // 1.3.If we already are a custom mapdata, abort the new assigment
    if(strcmp("base",vLayers[layer]->map->GetMapType()))
    {
        ILogSystem.Msg(iLogN," [MSTE] Warning: a custom mapdata is already set.\n\n");
        return 0;
    }

    // 2.Now, we will try to clone the base map to our custom map
    if(vLayers[layer]->map->CloneMapBase(map) == 0)
    {
        ILogSystem.Msg(iLogN," [MSTE] Warning: mapdata and custom mapdata dont have compatible attributes.\n\n");
        return 0;
    }

    // 3.Depending on auto resource control flag, we remove previous map or not
    if(vLayers[layer]->bResControl >= TE_RESCONTROL_AUTO) delete vLayers[layer]->map;

    // 4.Finally, we set our new custom map and return
    vLayers[layer]->map = map;
    ILogSystem.Msg(iLogN," [MSTE] Info: custom mapdata successfully added.\n\n");

    return 1;
}

//! Delete the custom mapdata. The base mapdata is left without any modification
/*! \param layer Selected layer to try to delete the custom mapdata  (from 0 to number of allocated layers).
\return 0 Failed
\return 1 Succeed*/
int cMSTE::DeleteCustomMapData(int layer)
{
    // 1.Initial checks
    if(!bTileEngine) return 0;
    if(layer < 0 || layer >= (int)vLayers.size()) return 0;	
    // 1.1.Are we a custom map?
    if(!strcmp("base",vLayers[layer]->map->GetMapType())) return 0;

    // 2.Create a clone mapdata
    cMSTE_MapBase *BaseMap;
    BaseMap = new cMSTE_MapBase(vLayers[layer]->map->GetMapHeight(),vLayers[layer]->map->GetMapWidth());
    BaseMap->SetCellWidth(vLayers[layer]->map->GetCellWidth());
    BaseMap->SetCellHeight(vLayers[layer]->map->GetCellWidth());
    vLayers[layer]->map->CloneMapBase(BaseMap);

    // 3.Depending on auto resource control flag, we remove previous map or not
    if(vLayers[layer]->bResControl >= TE_RESCONTROL_AUTO) delete vLayers[layer]->map;
    vLayers[layer]->map = BaseMap;
    return 1;
}

//! Get the layer name
/*! \param layer Selected layer to get its name (from 0 to number of allocated layers).
\return NULL Failed
\return Any other value, the string with its name */
char *cMSTE::GetLayerName(int layer)
{
    // 1.Initial checks
    if(!bTileEngine) return NULL;
    if(layer < 0 || layer >= (int)vLayers.size()) return NULL;

    // 2.Return layer name
    return vLayers[layer]->szName;
}

//! Set a new layer name.
/*! \param layer Selected layer to set its name (from 0 to number of allocated layers).
\param name Pointer to the new name. Maximum of 32 characters.
\return 0 Failed
\return 1 Succeed*/
int	cMSTE::SetLayerName(int layer, char *name)
{
    // 1.Initial checks
    if(!bTileEngine) return 0;
    if(layer < 0 || layer >= (int)vLayers.size()) return 0;
    if(name == NULL) return 0;

    // 2.Assign the new layer name
    szCopy(vLayers[layer]->szName,name,sizeof(vLayers[layer]->szName));
    return 1;
}


//-------------------------------------------------------------------------------
// Move and update methods
//-------------------------------------------------------------------------------

//! Move the layer adding of subtracting the given X/Y values
/*! Changing the position of a layer will modify the position for the rest of linked ones according
to each parallax ratio value and holding the "global" position.<br>
Remember that a layer with a parallax ratio of 0.0 is independent from the others.<br>
Requires a valid base mapdata.<br>
\param layer Selected layer to be moved (from 0 to number of allocated layers).
\param x Float X value, positive or negative. 0 to do nothing.
\param y Float Y value, positive or negative. 0 to do nothing.
\return -1 Failed
\return 0 Succeed but no change on the position done
\return 1 Succeed and the position changed*/
int cMSTE::MoveLayer(int layer, float x, float y)
{
    float fX,fY;

    // 0.Initial checks
    if(!bTileEngine) return -1;
    if(layer < 0 || layer >= (int)vLayers.size()) return -1;

    // 1.Assume no movement at all
    fX = fY = TE_KEEP_VALUE;

    // 2.Try to move on X axis.
    if(x != 0) 
    {
        fX = vLayers[layer]->fPosX + x;
    }

    // 3.Try to move on Y axis.
    if(y != 0) 
    {
        fY = vLayers[layer]->fPosY + y;
    }

    // 4.Return
    return SetLayerPosition(layer,fX,fY);
}

//! Update layer internals. 
/*! Only layers with TE_LAYERFLAG_UPDATE or TE_LAYERFLAG_UPDATECELLS flags will be processed.<br>
It calculates the autoscroll if applies and some pre-calculations for the rendering of the layers.<br> 
With TE_LAYERFLAG_UPDATECELLS it will call CellUpdate() method for each visible cell.<br>
This method should be called as part of the logic code.<br>
Optionally, you can select the update layer order calling this method with individual layers.<br> 
At the end of each layer update, it will call to OnUpdateEnd event function if exists.
\param layer By default it is -1 for updating all layers from 0(background) to the number of allocated layers(foreground). 
Any other value for updating the given layer.
\return 0 Failed
\return 1 Succeed*/
int cMSTE::Update(int layer)
{
    st_layers *lP = NULL;
    float fDeltaTime = 0;
    Uint32 iTmp = 0;

    // 1.Initial checks
    if(bTileEngine != 2) return 0;
    if(layer >= (int)vLayers.size()) return 0;

    // 2.Loop through all layers with TE_LAYERFLAG_UPDATE or TE_LAYERFLAG_UPDATECELLS flags
    for(int cl = 0; cl < (int)vLayers.size(); cl++)
    {
        // 2.1.Check for requested layer update and for TE_LAYERFLAG_UPDATE or TE_LAYERFLAG_UPDATECELLS flags
        if((layer != -1) && (layer != cl)) continue;		
        lP = vLayers[cl]; // Pointer to the layer for easier access
        if(((lP->iStatus & TE_LAYERFLAG_UPDATE) == 0) && ((lP->iStatus & TE_LAYERFLAG_UPDATECELLS) == 0)) continue;

        // 2.2.Calculate time vars for each layer
        iTmp = SDL_GetTicks();
        fDeltaTime = (float)(iTmp - lP->iLastUpdateTime) * 0.001f;
        lP->iLastUpdateTime = iTmp;

        // 2.3.Supports TE_LAYERFLAG_AUTOSCROLL and TE_LAYERFLAG_PAUSE flags
        // Note that if one layer reaches its bounds, the rest will go on, so it can break the "global" synchronized position.
        // To avoid it, the layers with autoscroll should long the same or has a loop on the needed axis.
        if((lP->iStatus & TE_LAYERFLAG_AUTOSCROLL) && !(lP->iStatus & TE_LAYERFLAG_PAUSE))
        {
            if(lP->fSpeedX != 0.0f) TryMoveX(cl,lP->fPosX + fDeltaTime * lP->fSpeedX);
            if(lP->fSpeedY != 0.0f) TryMoveY(cl,lP->fPosY + fDeltaTime * lP->fSpeedY);
        }

        // 2.4.SmoothScroll support
        if(lP->iDisableNextSmooth == 1)
        {
            lP->fLogicX_pre = lP->fPosX;
            lP->fLogicY_pre = lP->fPosY;
            lP->iDisableNextSmooth = 0;
        }
        else
        {
            lP->fLogicX_pre = lP->fLogicX;
            lP->fLogicY_pre = lP->fLogicY;
        }
        lP->fLogicX = lP->fPosX;
        lP->fLogicY = lP->fPosY;

        // 2.5.Pre-calcs
        // These calculations are done only at cells level
        int tx = lP->map->GetCellWidth();
        int ty = lP->map->GetCellHeight();
        int iCXBegin = (int)((lP->fLogicX - (float)lP->iOffsetX) / (float)tx);
        int iCXEnd = iCXBegin + (rViewport.w + lP->iOffsetX) / tx;
        int iCYBegin = (int)(lP->fLogicY / (float)ty);
        int iCYEnd = iCYBegin + (rViewport.h + lP->iOffsetY) / ty;

        // 2.6.Updating cells inside the Viewport including when the tile is longer than the cell size and part of is visible
        // Only with TE_LAYERFLAG_UPDATECELLS flag
        int iX = -1, iY = -1;
        for(int y = iCYBegin; (y < iCYEnd && (lP->iStatus & TE_LAYERFLAG_UPDATECELLS)); y++)
        {
            // y is on "raw" mode, we have to check for map boundaries and loop property
            iY = ValidateMapY(y,lP);
            if(iY == -1) continue;
            for(int x = iCXBegin; x < iCXEnd; x++)
            {
                // x is on "raw" mode, we have to check for map boundaries and loop property
                iX = ValidateMapX(x,lP);
                if(iX == -1) continue;
                //lP->map->CellUpdate(iX,iY,this,(void*)(rViewport.x - lP->iOffsetX + (x - iCXBegin ) * tx),(void*)(rViewport.y + (y - iCYBegin) * ty));						                
                lP->map->CellUpdate(iX,iY,this,NULL,NULL);						                
            }            
        }

        // 2.7.Call to OnUpdateEnd event function if any
        iCXEnd *= (rViewport.x - lP->iOffsetX - iCXBegin * tx);
        iCYEnd *= (rViewport.y - iCYBegin * ty);
        if(lP->OnUpdateEnd != NULL) lP->OnUpdateEnd(
            iCXBegin, iCYBegin,	    			                     // Cell X and Y
            rViewport.x - lP->iOffsetX - iCXBegin * tx, iCXEnd, tx,  // Screen pixel X: start,end and inc
            rViewport.y - iCYBegin * ty, iCYEnd, ty);                // Screen pixel Y: start,end and inc
    }

    // 3.Return
    return 1;
}

//! Render of visible layer cells
/*! Only layers with TE_LAYERFLAG_RENDER flag will be processed.<br>
It takes the positions calculated by Update() in order to render the cells.<br>
This method should be called as part of the graphic update code.<br>
You can select the rendering order calling this function with a selected layer.<br> 
At the end of each layer render, it will call to OnRenderEnd event function if exists.<br>
If the layer has TE_LAYERFLAG_SMOOTHSCROLL enabled, this method will interpolate values to perform a smooth scrolling movement.
\param layer By default it is -1 which means all layers will be processed. Any other value from 0(background) to number of allocated layers(foreground)
will only process the given layer.
\return 0 Failed
\return 1 Succeed*/
int cMSTE::Render(int layer)
{
    st_layers *lP = NULL;
    SDL_Rect rTmp;

    // 1.Initial checks
    if(bTileEngine != 2) return 0;
    if(layer >= (int)vLayers.size()) return 0;

    // 2.Save current clipping and set the new one
    SDL_GetClipRect(sRenderTarget,&rTmp);
    SDL_SetClipRect(sRenderTarget,&rViewport);

    // 3.Loop through all layers with TE_LAYERFLAG_RENDER flag
    for(int cl = 0; cl < (int)vLayers.size(); cl++)
    {
        // 3.1.Check for requested layer render and for TE_LAYERFLAG_RENDER flag
        if((layer != -1) && (layer != cl)) continue;		
        lP = vLayers[cl]; // Pointer for easier access
        if((lP->iStatus & TE_LAYERFLAG_RENDER) == 0) continue;
        // 3.2.SmoothScroll support
        if(lP->iStatus & TE_LAYERFLAG_SMOOTHSCROLL)
        {
            // Our interpolation value...
            float fLogicTime = CRM32Pro.GetLogicTime();
            float fTimeStep = fLogicTime - floor(fLogicTime);
            lP->fGfxX = lP->fLogicX_pre * (1 - fTimeStep) + lP->fLogicX * fTimeStep;
            lP->fGfxY = lP->fLogicY_pre * (1 - fTimeStep) + lP->fLogicY * fTimeStep;
        }
        else
        {
            lP->fGfxX = lP->fLogicX;
            lP->fGfxY = lP->fLogicY;
        }
        // 3.2.Pre-calcs
        // These calculations are done at pixel level
        int tx = lP->map->GetCellWidth();
        int ty = lP->map->GetCellHeight();
        int iAdjX = (int)(lP->fGfxX) % tx;
        if(iAdjX < 0) iAdjX += tx;
        int iAdjY = (int)(lP->fGfxY) % ty;
        if(iAdjY < 0) iAdjY += ty;
        // Source positions
        int iSrcX = (int)(lP->fGfxX) - lP->iOffsetX - iAdjX;
        int iSrcW = iSrcX + rViewport.w + lP->iOffsetX + iAdjX;
        int iSrcY = (int)(lP->fGfxY) - iAdjY;
        int iSrcH = iSrcY + rViewport.h + lP->iOffsetY + iAdjY;
        // Destination positions
        int iDstX = rViewport.x - iAdjX - lP->iOffsetX;
        int iDstY = rViewport.y - iAdjY;

        // 3.3.Rendering cells inside the Viewport including when the tile is longer than the cell size and part of is visible
        int iCX = -1, iCY = -1;
        int iPY = iDstY;
        for(int y = iSrcY; y < iSrcH; y += ty)
        {
            // y is on "raw" mode, we have to check for map boundaries and loop property
            iCY = ValidateMapY(y / ty,lP);
            if(iCY == -1) continue;
            int iPX = iDstX;
            for(int x = iSrcX; x < iSrcW; x += tx)
            {
                // x is on "raw" mode, we have to check for map boundaries and loop property
                iCX = ValidateMapX(x / tx,lP);
                if(iCX == -1) { iPX += tx; continue; }
                lP->map->CellRender(iCX,iCY,iPX,iPY,this,lP->iAlpha); 					                
                iPX += tx;
            }            
            iPY += ty;
        }

        // 3.5.Call to OnRenderEnd event function if any
        if(lP->OnRenderEnd != NULL) lP->OnRenderEnd(
            ValidateMapX(iSrcX / tx,lP), ValidateMapY(iSrcY / ty,lP), // X and Y start cells
            iSrcX, iSrcW, tx,   // X start,end and inc
            iSrcY, iSrcW, ty);  // Y start,end and inc

        // 3.6.ShowGrid
        if(lP->iStatus & TE_LAYERFLAG_SHOWGRID)
        {
            int x = iDstX;
            int y = iDstY;
            // Horizontal lines
            for(; y < (iDstY + rViewport.h + lP->iOffsetY + iAdjY); y = y + ty)
            {
                IPrimitives->HLineColor(sRenderTarget,x,rViewport.x + rViewport.w,y,lP->iGridColor);
            }
            // Vertical lines			
            y = iDstY;
            for(; x < (iDstX + rViewport.w + lP->iOffsetX + iAdjX); x = x + tx)
            {
                IPrimitives->VLineColor(sRenderTarget,x,y,rViewport.y + rViewport.h,lP->iGridColor);
            }
        }
    }

    // Restore previous clipping
    SDL_SetClipRect(sRenderTarget,&rTmp);
    return 1;
}

// Internal method for converting a raw X map position to a valid map position. Return -1 if it is out of the map
int cMSTE::ValidateMapX(int iX, st_layers *lP)
{
    if(iX >= lP->map->GetMapWidth()) 
    {
        if(lP->iStatus & TE_LAYERFLAG_LOOPX) return iX % lP->map->GetMapWidth();
        else return -1;
    }
    else if(iX < 0)					
    {
        if(lP->iStatus & TE_LAYERFLAG_LOOPX) return (lP->map->GetMapWidth() - 1) - ((-iX - 1) % lP->map->GetMapWidth());
        else return -1;
    }
    return iX;
}

// Internal method for converting a raw Y map position to a valid map position. Return -1 if it is out of the map
int cMSTE::ValidateMapY(int iY, st_layers *lP)
{
    if(iY >= lP->map->GetMapHeight()) 
    {
        if(lP->iStatus & TE_LAYERFLAG_LOOPY) return iY % lP->map->GetMapHeight();
        else return -1;
    }
    else if(iY < 0)					
    {
        if(lP->iStatus & TE_LAYERFLAG_LOOPY) return (lP->map->GetMapHeight() - 1) - ((-iY - 1) % lP->map->GetMapHeight());
        else return -1;
    }
    return iY;
}

//! Convert screen coordinates in to layer cell values
/*! \param layer Selected layer (from 0 to number of allocated layers).
\param sx Screen X coordinate (mouse position, etc.)
\param sy Screen Y coordinate (mouse position, etc.)
\param cx Point to integer where the Cell X position will be returned if this method call succeed
\param cy Point to integer where the Cell Y position will be returned if this method call succeed
\return 0 Failed
\return 1 Succeed */
int cMSTE::Screen2LayerCell(int layer, int sx, int sy, int *cx, int *cy)
{
    st_layers *lP = NULL;

    // 1.Use Screen2LayerAbsolute()
    if(Screen2LayerAbsolute(layer,sx,sy,cx,cy))
    {
        lP = vLayers[layer];
        *cx /=  lP->map->GetCellWidth();
        *cy /=  lP->map->GetCellHeight();
        return 1;
    }

    // 2.Return
    return 0;
}

//! Convert screen coordinates in to layer absolute values
/*! \param layer Selected layer (from 0 to number of allocated layers).
\param sx Screen X coordinate (mouse position, etc.)
\param sy Screen Y coordinate (mouse position, etc.)
\param px Point to integer where the absolute X position will be returned if this method call succeed
\param py Point to integer where the absolute Y position will be returned if this method call succeed
\return 0 Failed
\return 1 Succeed */
int cMSTE::Screen2LayerAbsolute(int layer, int sx, int sy, int *px, int *py)
{
    st_layers *lP = NULL;

    // 1.Initial checks
    if(!bTileEngine) return 0;
    if(layer < 0 || layer >= (int)vLayers.size()) return 0;
    lP = vLayers[layer]; 
    if((lP->iStatus & TE_LAYERFLAG_RENDER) == 0) return 0;

    // 2.Check we are inside the viewport
    if(sx < rViewport.x) return 0;
    if(sx > rViewport.x + rViewport.w) return 0;
    if(sy < rViewport.y) return 0;
    if(sy > rViewport.y + rViewport.h) return 0;

    // 3.Get pixel coordinates
    *px = ((sx - rViewport.x) + (int)lP->fPosX);
    *py = ((sy - rViewport.y) + (int)lP->fPosY);

    // 4.Return
    return 1;
}

//-------------------------------------------------------------------------------
//  Tileset handling
//-------------------------------------------------------------------------------

//! Load a tileset
/*! bResControl is set to TE_RESCONTROL_USER: the user must remove this resource after usage<br>
szSource is set to TE_TILESET_DEFAULT_SOURCE: if the level is saved, this tileset will be saved in the same place as the level<br>
After loading the tileset, it'll call ValidateMapBase().
\param tileset Pointer to a valid CRM32Pro_CTile.
\param iID Global ID for this tileset. By default -1 to try to find a valid gID. A value of 0 is not allowed (minimum iD = 1)
\return -1 Failed
\return >=0 Succeed and the vTileset index where it was stored*/
int cMSTE::LoadTileset(CRM32Pro_CTile *tileset, int iID)
{
    unsigned int iTmpID = 0;
    int iX = 0, iY = 0, iNumT = 0;

    // 1.Initial checks
    if(!bTileEngine) return -1;
    if(tileset == NULL) return -1;
    if(iID == 0) return -1;
    if(iID <= -1) iID = -1;

    // 2.Get number of tiles and the offsets
    tileset->GetOffset(&iX,&iY);
    iNumT = (tileset->GetSizeX() / iX) * (tileset->GetSizeY() / iY);

    // 3.If we have an iID check if it is valid
    if(iID != -1)
    {
        // Check if the gID is a valid one
        int iC1 = 0, iC2 = 0;
        iC1 = CheckTile(iID);
        iC2 = CheckTile(iID + iNumT);
        if((iC1 != -1) || (iC2 != -1))
        {
            ILogSystem.Msg(iLogN," [MSTE] Warning: tileset '%s' with global ID '%d' was not loaded (collision with gID '%d')\n",
                tileset->GetName(),iID,(iC1!=-1)?vTilesets[iC1]->iGID:vTilesets[iC2]->iGID);
            return -1; 
        }
        // It is a valid id
        iTmpID = iID;
    }
    // 3.1.Otherwise, look for one
    else
    {
        // Get the highest current gID
        iTmpID = 1;
        for(int i = 0; i < (int)vTilesets.size(); i++)
        {
            if(iTmpID < vTilesets[i]->iGID + vTilesets[i]->iNumTiles) iTmpID = vTilesets[i]->iGID + vTilesets[i]->iNumTiles;
        }		
        if(iTmpID != 1) iTmpID++; // We leave one empty slot between different tilesets
    }

    // 4.Load the new tileset. We dont sort them out for efficiency reasons: most used tiles must be on the low tilesets
    vTilesets.push_back(new st_tilesets(iTmpID,TE_TILESET_DEFAULT_SOURCE,tileset,TE_RESCONTROL_USER));

    // 5.Run ValidateMapBase() on layers with map and with .bTilesetReady = 0
    int iNL = 0;
    for(int i = 0; i < (int)vLayers.size(); i++)
    {
        if((vLayers[i]->map != NULL) && (vLayers[i]->bTilesetReady == 0)) // peta aki
        {
            if(ValidateMapBase(i)) iNL++;
        }
    }

    // 6.Print info
    ILogSystem.Msg(iLogN," [MSTE] Info: tileset '%s' with global ID '%d' loaded:\n",tileset->GetName(),iTmpID);
    ILogSystem.Msg(iLogN,"         . %d tiles of %dx%d\n",iNumT,iX,iY);
    if(iNL != 0) ILogSystem.Msg(iLogN,"         . %d %s now ready\n",iNL,(iNL > 1)?"layers are":"layer is");	
    return iTmpID;
}

//! Load a tileset
/*! bResControl is set to TE_RESCONTROL_AUTO: the tileset is automatically removed after usage<br>
szSource is set to the DPF that contains the tileset<br>
After loading the tileset, it'll call ValidateMapBase().
\param sDPFfile Pointer to DPF filename that contains the tileset.
\param sTileName Pointer to the Tilename to be loaded
\param iID Global ID for this tileset. By default -1 to try to find a valid gID. A value of 0 is not allowed (minimum iD = 1)
\return -1 Failed
\return >=0 Succeed and the vTileset index where it was stored*/
int cMSTE::LoadTileset(char *sDPFfile, char *sTileName, int iID)
{
    CRM32Pro_CTile *tLoad = NULL;
    int iRet = -1;

    // Try to load the tileset
    tLoad = new CRM32Pro_CTile;
    if(!tLoad->Load(sDPFfile,sTileName))
    {
        delete tLoad;
        return iRet;
    }
    // Pass it to our loader method
    iRet = LoadTileset(tLoad,iID);
    if( iRet == -1)
    {
        delete tLoad;
        return iRet;
    }

    // Change the bResControl and szSource
    int iIndex = CheckTile(iRet);
    vTilesets[iIndex]->bResControl = TE_RESCONTROL_AUTO;
    SetTilesetSource(iRet,sDPFfile);

    // Loaded!
    return iRet;
}

//! Load a tileset
/*! bResControl is set to TE_RESCONTROL_AUTO: the tileset is automatically removed after usage<br>
szSource is set to the external image<br>
After loading the tileset, it'll call ValidateMapBase().
\param sIMGfile Pointer to PNG or BMP filename
\param sTileName Pointer to the Tilename to be loaded
\param iTileWidth Integer with the tile width
\param iTileHeight Integer with the tile height
\param iID Global ID for this tileset. By default -1 to try to find a valid gID. A value of 0 is not allowed (minimum iD = 1)
\return -1 Failed
\return >=0 Succeed and the vTileset index where it was stored*/
int cMSTE::LoadTileset(char *sIMGfile, char *sTileName, int iTileWidth, int iTileHeight, int iID)
{
    CRM32Pro_CTile *tLoad = NULL;
    int iRet = -1;

    // Try to load the tileset
    tLoad = new CRM32Pro_CTile;
    if(!tLoad->Create(sIMGfile,sTileName))
    {
        delete tLoad;
        return iRet;
    }

    // Try to create the tileset
    if(!tLoad->SetTileSet(1,iTileWidth,iTileHeight))
    {
        delete tLoad;
        return iRet;
    }

    // Pass it to our loader method
    iRet = LoadTileset(tLoad,iID);
    if( iRet == -1)
    {
        delete tLoad;
        return iRet;
    }

    // Change the bResControl and szSource
    int iIndex = CheckTile(iRet);
    vTilesets[iIndex]->bResControl = TE_RESCONTROL_AUTO;
    SetTilesetSource(iRet,sIMGfile);

    // Loaded!
    return iRet;
}

//! Unload a given tileset
/*! If it success unloading the tileset, it'll call ValidateMapBase().
\param iTN Global ID or any tile of the tileset to be removed
\return 0 Failed
\return 1 Succeed*/
int cMSTE::UnloadTileset(int iTN)
{
    st_tilesets *Ts;
    int iX = 0, iY = 0;

    // 1.We only need an initialized engine
    if(!bTileEngine) return 0;

    // 2.Look for the tileset
    int resul = CheckTile(iTN);
    if(resul != -1)
    {
        int iNL = 0;
        vTilesets[resul]->tTileset->GetOffset(&iX,&iY);
        Ts = vTilesets[resul]; // Save pointer for removing tilesets after displaying info		
        vTilesets.erase(vTilesets.begin() + resul); 
        // Run ValidateMapBase() on layers with map and with .bTilesetReady = 1
        for(int i = 0; i < (int)vLayers.size(); i++)
        {
            if((vLayers[i]->map != NULL) && (vLayers[i]->bTilesetReady == 1))
            {
                ValidateMapBase(i);
                // Check if the layer is disabled due to the tileset used is now being unloaded..
                if(vLayers[i]->bTilesetReady == 0) iNL++;
            }
        }
        ILogSystem.Msg(iLogN," [MSTE] Info: tileset '%s' with global ID '%d' unloaded:\n",Ts->tTileset->GetName(),Ts->iGID);
        ILogSystem.Msg(iLogN,"         . %d tiles of %dx%d\n",Ts->iNumTiles,iX,iY);
        if(iNL != 0) ILogSystem.Msg(iLogN,"         . %d %s disabled\n",iNL,(iNL > 1)?"layers were":"layer was");
        delete Ts;

        // Return
        return 1;
    }
    // Return
    else return 0;
}

//! Get a pointer to the tileset
/*! It can be used for modifying any tileset attributes
\param iTN Global ID or any tile of the tileset
\return NULL Failed
\return Pointer to CRM32Pro_CTile*/
CRM32Pro_CTile *cMSTE::GetTileset(int iTN)
{
    // Checks
    if(!bTileEngine) return NULL;

    // Try to get the pointer
    int resul = CheckTile(iTN);
    if(resul != -1)
    {
        return vTilesets[resul]->tTileset;
    }

    // Failed!
    return NULL;
}

//! Get tileset source
/*! The tileset source is used when saving the level
\param iTN Global ID or any tile of the tileset
\return NULL Failed
\return Pointer to char with: TE_TILESET_DEFAULT_SOURCE, a full path and filename of a DPF, PNG or BMP*/
char *cMSTE::GetTilesetSource(int iTN)
{
    // Checks
    if(!bTileEngine) return NULL;

    // Try to get the pointer
    int resul = CheckTile(iTN);
    if(resul != -1)
    {
        return vTilesets[resul]->szSource;
    }
    return NULL;
}

//! Set tileset source
/*! The tileset source is used when saving the level
\param iTN Global ID or any tile of the tileset
\param sSource New source: TE_TILESET_DEFAULT_SOURCE, a full path and filename of a DPF, PNG or BMP
\return 0 Failed
\return 1 Succeed*/
int cMSTE::SetTilesetSource(int iTN, char *sSource)
{
    // Checks
    if(!bTileEngine) return 0;
    if(sSource == NULL) return 0;
    if(!CheckFileExtension(sSource,".png") && !CheckFileExtension(sSource,".bmp") && 
        !CheckFileExtension(sSource,".dpf") && strcmp(sSource,TE_TILESET_DEFAULT_SOURCE)) return 0;

    // Try to get the pointer
    int resul = CheckTile(iTN);
    if(resul != -1)
    {
        if(vTilesets[resul]->szSource != NULL) delete []vTilesets[resul]->szSource;
        vTilesets[resul]->szSource = new char[strlen(sSource) + 1];
        strcpy(vTilesets[resul]->szSource,sSource);
        return 1;
    }
    return 0;
}