/*----------------------------------------------
   MegaStorm Tile-based Engine - Roberto Prieto
   Copyright (C) 2007-2012 MegaStorm Systems
  ----------------------------------------------
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
    
    Roberto Prieto
    contact@megastormsystems.com
    http://www.megastormsystems.com

    Used software licenses, including the one for this engine, are located
    at /licenses directory. Please, include them on your distributions.

 --------------------------------------------------------------------------
             MSTE Main engine class
    
    Changelog: read MSTE.h to see all changes

 -------------------------------------------------------------------------- */
#include <CRM32Pro.h>
#include "MSTE.h"
#include "MSTE_MapData.h"
#include "MSTE_Tools.h"

//-------------------------------------------------------------------------------
// ImportImage() and ExportImage() methods - IO Filters for images
//-------------------------------------------------------------------------------

//! Import an image file(BMP or PNG) into a base mapdata.
/*! It will create a new base mapdata and tileset.
<br>The selected layer must be empty (no mapdata present).
<br>The level name will be used for generating the layer and tileset names.
\param layer Selected layer ID (from 0 to number of allocated layers). 
\param fileimage Pointer to a string with path and filename of a BMP or PNG.
\param tileSizeX Tile width size to divide the given image.
\param tileSizeY Tile height size to divide the given image.
\param accuracy 
\param colorkey Transparent color for the given image. Useless if the surface uses alpha per pixel.
\return -1 Failed: bad parameters
\return -2 Failed: error loading image. 24bits color depth images are not supported.
\return -3 Failed: tileSizeX does not divide the given image
\return -4 Failed: tileSizeY does not divide the given image
\return -5 Failed: the tileset creation failed
\return -6 Failed: can not load the specified file image
\return 1 Succeed */ 
int cMSTE::ImportImage(int layer, char *fileimage, int tileSizeX, int tileSizeY, int accuracy, int colorkey)
{
    SDL_Surface *surfTmp = NULL;
    int ret;

    // Only one check...
    if(fileimage == NULL) return -1;
    
    // Is it a BMP?
    if(IImage->isBMP(fileimage))  surfTmp = IImage->LoadBMP(fileimage);
    // Is it a PNG?
    else if(IImage->isPNG(fileimage)) surfTmp = IImage->LoadPNG(fileimage);

    // If any error...
    if(surfTmp == NULL) 
    {
        ILogSystem.Msg(iLogN,"\n [MSTE] Error: can not load '%s' file image in to layer ID '%d'\n", fileimage,layer);
        return -6;
    }

    // Calling ImportImage()
    ret = ImportImage(layer,surfTmp,tileSizeX,tileSizeY,accuracy, colorkey);
    SDL_FreeSurface(surfTmp);

    // Returning
    return ret;
}

//! Import a surface into a base mapdata.
/*! It will create a new base mapdata and a new tileset.
<br>The selected layer must be empty (no mapdata present).
<br>The level name will be used for generating the layer and tileset names.
\param layer Selected layer ID (from 0 to number of allocated layers). 
\param surfRaw Pointer to the SDL_Surface to be imported
\param tileSizeX Tile width size to divide the given image.
\param tileSizeY Tile height size to divide the given image.
\param accuracy 
\param colorkey Transparent color for the given image. Useless if the surface uses alpha per pixel.
\return -1 Failed: bad parameters
\return -2 Failed: error loading image. 24bits color depth images are not supported.
\return -3 Failed: tileSizeX does not divide the given image
\return -4 Failed: tileSizeY does not divide the given image
\return -5 Failed: the tileset creation failed
\return 1 Succeed */ 
#define TILESET_OUTPUT_WIDTH 1024
int cMSTE::ImportImage(int layer, SDL_Surface *surfRaw, int tileSizeX, int tileSizeY, int accuracy, int colorkey)
{
    SDL_Surface *surfWork, *surfTmp;
    int iNumTL, iSizeX, iSizeY, iMaxDeviation;
    int i, j, x, y, iTmp, autoSizeX, autoSizeY;	
    SDL_Rect rDest,rOrig;
    char nameTS[32], sTmp[5];
    char bPerPixel = 0;
    CRM32Pro_CTile *tTile;
    cMSTE_MapBase  *MapWork;
    _sTileProcess  **iTileArray;
    Uint32 saved_flags;
    Uint8  saved_alpha;

    // 1.Initial checks
    if(!bTileEngine) return -1;
    if(layer < 0 || layer >= (int)vLayers.size()) return -1;
    if(vLayers[layer]->map != NULL) return -1;
    if((tileSizeX <= 0) || (tileSizeY <= 0)) return -1;
    if(accuracy > 100) accuracy = 100;
    if(accuracy < 1) accuracy =   1;

    // 2.What kind of surface?
    if(surfRaw == NULL) return -2; 
    // 2.1.OpenGL textures can not be used, so we detect it and create a surface for working with it    
    if(CRM32Pro.IsTexture(surfRaw)) surfTmp = SDL_DisplayFormat(surfRaw);
    else surfTmp = surfRaw;
    // 2.2.Detecting alpha per-pixel
    if((surfTmp->flags & SDL_SRCALPHA) && (surfTmp->format->alpha == 255)) { colorkey = -1; bPerPixel = 1; }
    if(surfTmp->format->BitsPerPixel == 24) { if(surfTmp != surfRaw) SDL_FreeSurface(surfTmp); return -2; }

    // 3.Does tile sizes divide the image?
    if((surfTmp->w % tileSizeX) > 0) { if(surfTmp != surfRaw) SDL_FreeSurface(surfTmp); return -3; }
    if((surfTmp->h % tileSizeY) > 0) { if(surfTmp != surfRaw) SDL_FreeSurface(surfTmp); return -4; }
    ILogSystem.Msg(iLogN,"\n [MSTE] Info: trying to import a %dx%d surface using %dx%d tiles in to layer ID '%d'.\n",
        surfTmp->w,surfTmp->h,tileSizeX,tileSizeY,layer);

    // 4.Preparing data for the first process: looking for unique tiles
    iNumTL = 0;
    iSizeX = surfTmp->w / tileSizeX;
    iSizeY = surfTmp->h / tileSizeY;
    iMaxDeviation = (int)((float)(tileSizeX * tileSizeY) * (float)((100.0 - (float)accuracy) / 100.0));
    MapWork = new cMSTE_MapBase(iSizeY,iSizeX);
    MapWork->SetCellWidth(tileSizeX);
    MapWork->SetCellHeight(tileSizeY);
    iTileArray = MapWork->Create2DArray<_sTileProcess>(iSizeY,iSizeX);
    for(y = 0; y < iSizeY; y++)		
        for(x = 0; x < iSizeX; x++)
        {
            iTileArray[y][x].status = -1;
            iTileArray[y][x].posx = x * tileSizeX;
            iTileArray[y][x].posy = y * tileSizeY;
        }
        ILogSystem.Msg(iLogN," [MSTE] Info: using an accuracy of %%%d on tile matching (max pixels deviation %d).\n",
            accuracy, iMaxDeviation);

        // 5.Looking for unique tiles
        for(y = 0; y < iSizeY; y++)
        {
            for(x = 0; x < iSizeX; x++)
            {
                // Is it a non-processed tile?
                if(iTileArray[y][x].status == -1)
                {
                    // Now, we will look for all matching tiles...
                    for(j = 0; j < iSizeY; j++)
                        for(i = 0; i < iSizeX; i++)
                        {
                            if(iTileArray[j][i].status == -1)
                            {		
                                // Compare both tiles using iMaxDeviation. It detects transparent tiles.
                                iTmp = CompareTiles(&iTileArray[y][x],&iTileArray[j][i],surfTmp,iMaxDeviation,tileSizeX,tileSizeY,colorkey);
                                switch(iTmp)
                                {
                                case 0: // Diff tile									
                                    break;
                                case 1: // Same tile
                                    if(iTileArray[y][x].status == -1) iTileArray[y][x].status = iNumTL + 1;
                                    iTileArray[j][i].status = iTileArray[y][x].status;
                                    break;
                                case 2: // Transparent tile
                                    iTileArray[j][i].status = 0;
                                    break;
                                }								
                            }
                        }
                        // Inc number of tiles. Discard transparent tiles
                        if(iTileArray[y][x].status != 0) iNumTL++;
                }
            }
        }

        // 6.Important error checking 
        if((iNumTL <= 0) || (iNumTL > (iSizeX * iSizeY))) 
        {
            ILogSystem.Msg(iLogN," [MSTE] Error: failed to convert the surface to a tileset(%d).\n",iNumTL);
            MapWork->Delete2DArray(iTileArray);
            delete MapWork;
            if(surfTmp != surfRaw) SDL_FreeSurface(surfTmp);
            return -5;
        }
        ILogSystem.Msg(iLogN," [MSTE] Info: found %d unique tiles from a total of %d(%%%d compression).\n",
            iNumTL,iSizeX*iSizeY,100 - (iNumTL*100)/(iSizeX*iSizeY));

        // 7.Tileset creation: creating the surface and the Tile
        autoSizeY = ((iNumTL / (TILESET_OUTPUT_WIDTH / tileSizeY)) + 1) * tileSizeY;
        autoSizeX = iNumTL * tileSizeX;
        if(autoSizeX > TILESET_OUTPUT_WIDTH) autoSizeX = TILESET_OUTPUT_WIDTH;
        if(bPerPixel)
            surfWork = SDL_CreateRGBSurface(SDL_SWSURFACE,autoSizeX,autoSizeY, 32,
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
            0x000000FF,
            0x0000FF00,
            0x00FF0000,
            0xFF000000
#else
            0xFF000000,
            0x00FF0000,
            0x0000FF00,
            0x000000FF
#endif
            );
        else surfWork = SDL_CreateRGBSurface(SDL_SWSURFACE,autoSizeX,autoSizeY, 32, 0, 0, 0, 0);

        // Store alpha and to allow a full copy of the surface(including alpha information)
        if(bPerPixel)
        {
            saved_flags = surfTmp->flags&(SDL_SRCALPHA|SDL_RLEACCELOK);
            saved_alpha = surfTmp->format->alpha;
            SDL_SetAlpha(surfTmp, 0, 0);
        }
        else SDL_FillRect(surfWork,NULL,0);
        iTmp = 0;
        for(j = 0; j < (autoSizeY / tileSizeY) ;j++) 
            for(i = 0; i < (autoSizeX / tileSizeX); i++) 
            {
                // Loop looking for each unique Tile
                for(y = 0; y < iSizeY; y++)
                    for(x = 0; x < iSizeX; x++)
                        if((iTileArray[y][x].status - 1) == iTmp) 
                        {
                            // Blit to our work surface
                            rOrig.x = iTileArray[y][x].posx; rOrig.w = tileSizeX;
                            rOrig.y = iTileArray[y][x].posy; rOrig.h = tileSizeY;				
                            rDest.x = i * tileSizeX; rDest.w = tileSizeX;
                            rDest.y = j * tileSizeY; rDest.h = tileSizeY;				
                            SDL_BlitSurface(surfTmp,&rOrig,surfWork,&rDest);
                        }				
                        // End condition
                        iTmp++;
                        if(iTmp >= iNumTL) break;
            }
             ILogSystem.Msg(iLogN," [MSTE] Info: here!\n");
            // Restore alpha information
            if(bPerPixel) SDL_SetAlpha(surfWork, saved_flags, saved_alpha);

            // Creating our tile
            strcpy(nameTS,"TS-ImportedImage-Layer");
            _itoa(layer, sTmp, 10);
            strcat(nameTS,sTmp);
            tTile = new CRM32Pro_CTile;
            tTile->Create(surfWork,nameTS);
            if(colorkey != -1) tTile->SetColorKey(colorkey);
            SDL_FreeSurface(surfWork); 
            tTile->SetTileSet(1,tileSizeX,tileSizeY);
            iTmp = LoadTileset(tTile); 
            vTilesets[CheckTile(iTmp)]->bResControl = TE_RESCONTROL_INTERNAL;

            // 8.Map creation
            for(y = 0; y < iSizeY; y++)
            {
                for(x = 0; x < iSizeX; x++)
                {
                    if(iTileArray[y][x].status == 0) MapWork->iTilemap[y][x] = 0;
                    else MapWork->iTilemap[y][x] = iTileArray[y][x].status + iTmp - 1;
                }
            }
            MapWork->Delete2DArray(iTileArray);
            if(surfTmp != surfRaw) SDL_FreeSurface(surfTmp);

            // 9.1.Map assigment
            if(SetLayerMapData(layer,MapWork))
            {
                strcpy(nameTS,"ImportedImage-Layer");
                strcat(nameTS,sTmp);
                SetLayerName(layer,nameTS);
                vLayers[layer]->bResControl = TE_RESCONTROL_INTERNAL;
                vLayers[layer]->iAlpha = 255;
                ILogSystem.Msg(iLogN," [MSTE] Info: surface successfully imported.\n\n");
                return 1;
            }	
            // 9.2.Error!
            ILogSystem.Msg(iLogN," [MSTE] Error: it was not possible to import the surface.\n\n");
            delete MapWork;
            delete tTile;	
            return 0;
}


//! Export a base mapdata to a surface.
/*! In case any tileset used by the base mapdata has alpha per-pixel enabled, the output will be a surface with alpha per-pixel 
(colorkey tilesets are automatically converted).
\param layer Layer to export, from 0 to number of allocated layers or -1 to export all layers. 
It requires a layer with TE_LAYERFLAG_RENDER flag and with alpha layer different to 0 (transparent)
\param pSurf Surface pointer reference (must be NULL as the surface will be created by this method)
\return -1 Failed: TileEngine is not initialized
\return -2 Failed: wrong layer
\return -3 Failed: pSurf must be empty with a NULL value
\return 1 Succeed */ 
int cMSTE::ExportImage(int layer, SDL_Surface *&pSurf)
{
    int iDstWidth, iDstHeight;
    int iCellW, iCellH;
    int i,j, iLS,iLE,iL, iNumExported = 0;
    int bAlphaPerPixel, iCKey;
    SDL_Surface *sTmp = NULL;
    CRM32Pro_CTile *tWork = NULL;
    int iLastCached = -1;

    // 1.Initial checks
    if(!bTileEngine) return -1;
    if(layer >= (int)vLayers.size()) return -2;
    if(pSurf != NULL) return -3;
    
    // 2.Main loop
    iLS = layer; iLE = layer + 1;
    if(layer < 0) { iLS = 0; iLE = (int)vLayers.size(); } 
    for(iL = iLS; iL < iLE; ++iL)
    {
        if(vLayers[iL]->bTilesetReady == 0) continue;
        if(vLayers[iL]->map == NULL) continue;
        if((vLayers[iL]->iStatus & TE_LAYERFLAG_RENDER) == 0) continue;
        if(vLayers[iL]->iAlpha == 0) continue;
 
        // 2.1.Prepare vars
        bAlphaPerPixel = 0; iCKey = -1;
        iCellW = vLayers[iL]->map->GetCellWidth();
        iCellH = vLayers[iL]->map->GetCellHeight();
        iDstWidth = iCellW * vLayers[iL]->map->GetMapWidth();
        iDstHeight = iCellH * vLayers[iL]->map->GetMapHeight();

        // 2.2.Initial check of all involved tileset looking for alpha per-pixel tileset
        for(j = 0; (j < vLayers[iL]->map->GetMapHeight() && bAlphaPerPixel == 0); j++)
        {
            for(i = 0; i < vLayers[iL]->map->GetMapWidth(); i++)
            {
                int iT = vLayers[iL]->map->iTilemap[j][i];
                if(iT > 0)
                {
                    iT = CheckTile(iT);
                    if(vTilesets[iT]->tTileset->GetAlpha() == 256) 
                    {
                        bAlphaPerPixel = 1;
                        break;
                    }
                    else if(vTilesets[iT]->tTileset->GetColorKey() != -1)
                    {
                        iCKey = vTilesets[iT]->tTileset->GetColorKey(); // Use the latest colorkey found in the tilesets of this mapdata
                    }
                }
            }
        }

        // 2.3.Creating the destination surface
        if(bAlphaPerPixel)
        {
            pSurf = SDL_CreateRGBSurface(SDL_SWSURFACE,iDstWidth,iDstHeight, 32,
        #if SDL_BYTEORDER == SDL_LIL_ENDIAN
                0x000000FF,
                0x0000FF00,
                0x00FF0000,
                0xFF000000
        #else
                0xFF000000,
                0x00FF0000,
                0x0000FF00,
                0x000000FF
        #endif
            );
        }
        else 
        {
            pSurf = SDL_CreateRGBSurface(SDL_SWSURFACE,iDstWidth,iDstHeight, 32, 0, 0, 0, 0);
            if(iCKey == -1) iCKey = 0x00000000;
            SDL_FillRect(pSurf,NULL,iCKey);
        }

        // 2.3.Loop through the map blitting each tile
        for(j = 0; j < vLayers[iL]->map->GetMapHeight(); j++)
        {
            for(i = 0; i < vLayers[iL]->map->GetMapWidth(); i++)
            {
                unsigned int iT = vLayers[iL]->map->iTilemap[j][i];
                if(iT > 0)
                {
                    int saved_flags, saved_alpha;
                    int iTS = CheckTile(iT);
                    if(CRM32Pro.IsTexture(vTilesets[iTS]->tTileset->GetSurface()))
                    {
                        // With an OpenGL texture, we create a cloned tileset but with a plain surface
                        // We perform this conversion once or if the tileset changes while processing this layer
                        // In case of several tiles of different tileset, this will be performing several conversion
                        // as we are only caching the last one used...
                        if(iLastCached != iTS)
                        {
                            if(tWork != NULL) delete tWork;
                            tWork = new CRM32Pro_CTile();
                            SDL_Surface *sTmp = SDL_DisplayFormat(vTilesets[iTS]->tTileset->GetSurface());
                            CRM32Pro.SurfaceAutomaticConversion(0);
                            tWork->Create(sTmp,"TileTmp");
                            CRM32Pro.SurfaceAutomaticConversion(1);
                            SDL_FreeSurface(sTmp);
                            int iW,iH;
                            vTilesets[iTS]->tTileset->GetOffset(&iW,&iH);
                            tWork->SetTileSet(1,iW,iH);
                            iLastCached = iTS;
                        }
                    }
                    else tWork = vTilesets[iTS]->tTileset;

                    // Alpha per-pixel pre-tasks
                    if(tWork->GetAlpha() == 256) 
                    {   
                        saved_flags = tWork->GetSurface()->flags&(SDL_SRCALPHA|SDL_RLEACCELOK);
                        saved_alpha = tWork->GetSurface()->format->alpha;
                        SDL_SetAlpha(tWork->GetSurface(), 0, 0);
                    }
                    // Blit
                    tWork->SetPosition(i * iCellW, j * iCellH);
                    tWork->Draw(pSurf,iT - vTilesets[iTS]->iGID + 1);
                    // Alpha per-pixel final task
                    if(tWork->GetAlpha() == 256) SDL_SetAlpha(tWork->GetSurface(), saved_flags, saved_alpha);
                }
            }
        }
        if(iLastCached != -1) if(vTilesets[iLastCached]->tTileset != tWork) delete tWork;
        tWork = NULL; iLastCached = -1; 

        // 2.4.Blit over any possible previous exported layer
        if(iNumExported != 0)
        {
            if(!bAlphaPerPixel) SDL_SetColorKey(pSurf, SDL_SRCCOLORKEY, iCKey);
            SDL_BlitSurface(pSurf,NULL,sTmp,NULL);
            SDL_FreeSurface(pSurf);
        }
        else sTmp = pSurf;
        iNumExported++;
    }

    // Return
    pSurf = sTmp;
    if(iNumExported) 
    {
        ILogSystem.Msg(iLogN,"\n [MSTE] Info: exported '%d' layers to the given surface\n",iNumExported);
        return 1;
    }
    else return 0;
}

//! Export a base mapdata to an external image (BMP/PNG).
/*! In case any tileset used by the base mapdata has alpha per-pixel enabled, the output will be a surface with alpha per-pixel 
(colorkey tilesets are automatically converted).
\param layer Layer to export, from 0 to number of allocated layers or -1 to export all layers. It requires the layer has TE_LAYERFLAG_RENDER flag. 
\param filename Selected file name, the extension will set a BMP or PNG. Note that only PNG supports alpha blending information.
\return -1 Failed: bad parameters
\return -2 Failed: needed tilesets are not loaded
\return -3 Failed: layer has not a mapdata
\return -4 Failed: surface is not empty
\return -5 Failed: filename has not a .bmp or .png extension
\return 1 Succeed */ 
int cMSTE::ExportImage(int layer, char *filename)
{
    SDL_Surface *tmp = NULL;
    int ret;

    if(filename == NULL) return -1;
    int iS = SDL_GetTicks();
    ret = ExportImage(layer,tmp);
    ILogSystem.Msg(LOG_NORMAL,"Export time: %dms\n",SDL_GetTicks() - iS);

    // Save to file
    if(ret == 1)
    {
        if(CheckFileExtension(filename,".bmp")) IImage->SaveBMP(filename,tmp);
        else if(CheckFileExtension(filename,".png")) IImage->SavePNG(filename,tmp);
        else ret = -5;
    }
    SDL_FreeSurface(tmp);
    if(ret == 1) ILogSystem.Msg(iLogN,"                         - surface saved to '%s' image file\n",filename);
    return ret;
}

