/*----------------------------------------------
   MegaStorm Tile-based Engine - Roberto Prieto
   Copyright (C) 2007-2012 MegaStorm Systems
  ----------------------------------------------
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
    
    Roberto Prieto
    contact@megastormsystems.com
    http://www.megastormsystems.com

    Used software licenses, including the one for this engine, are located
    at /licenses directory. Please, include them on your distributions.

 --------------------------------------------------------------------------
             MSTE Main engine class
    
    Changelog: read MSTE.h to see all changes

 -------------------------------------------------------------------------- */
#include <CRM32Pro.h>
#include "MSTE.h"
#include "MSTE_MapData.h"
#include "MSTE_Tools.h"

//-------------------------------------------------------------------------------
// SaveLevel() and LoadLevel() methods - IO Filters for native format
//-------------------------------------------------------------------------------

//! Load a native format level
/*! It loads the level on current TileEngine instance.<br>
If the new layer or tileset IDs are already in use, they are not loaded, however, the viewport, creation/modification dates and the current level name are overwritten.<br>
It will only load the first valid base mapdata and custom mapdata (if any) found on the file.
\param sLevelFilename Filename of the level to be loaded (extension .msm). It supports optional paths.
\param sSrcDPF Optional parameter. String with the DPF filename that contains sLevelFilename. It supports optional paths.
\param sLevelname Optional parameter. Selected level name to load. By default it is set to NULL to load the first occurrence.
\return -1 Failed: engine was not initialized.
\return -2 Failed: sLevelFilename is not valid.
\return -3 Failed: sLevelFilename could not be loaded.
\return -4 Failed: a 'msse_level' node was not found.
\return -5 Failed: selected level name was not found.
\return -6 Failed: a 'name' node was not found.
\return 0 Warning: some tilesets and/or layers were not loaded. Check the log to further information.
\return 1 Succeed (all tilesets and layers found on the level loaded) */ 
int cMSTE::LoadLevel(char *sLevelFilename, char *sSrcDPF, char *sLevelname)
{
    int		idXML, retvalue;
    char	*szSrcPath = NULL, *szMSMName = NULL, *szSrcFile = NULL, *szAux = NULL;
    int		iSrcDPF = 0;
    int		iNumTS = 0, iNumTSL = 0; // Number of Tileset and tileset loaded
    int		iNumLY = 0, iNumLYL = 0; // Number of layers and layers loaded	
    int     iTmp = 0, iID = 0;
    char    *szTmp = NULL, *szTmp2 = NULL;
    double   dTmp = 0.0;

    SDL_Rect rectTmp;

    // 1.Initial checks
    if(!bTileEngine) return -1;
    if(sLevelFilename == NULL) return -2;

    // 2.Load the XML that contains our .msm
    if(sSrcDPF == NULL) idXML = CRM32Pro.XMLOpen(sLevelFilename);
    else 
    {
        idXML = CRM32Pro.XMLOpen(sSrcDPF,sLevelFilename);
    }
    if(idXML == 0) return -3;

    // 3.Point to the first msse_level node
    if(CRM32Pro.XMLNodePointTo(idXML,1,"msse_level") == 0) return -4;

    // 4.If we have sLevelname, look for it and if it is not there.. return with error
    if(sLevelname != NULL)
    {
        int bFlag = 1;
        do
        {
            retvalue = CRM32Pro.XMLAttributeGet(idXML,"name",&szTmp);
            if(!strcmp(sLevelname,szTmp)) bFlag = 2;
            if(bFlag == 1) bFlag = CRM32Pro.XMLNodeNext(idXML);
        } while(bFlag == 1); 
        if(bFlag == 0) { CRM32Pro.XMLClose(idXML); return -5; }
    }
    else if(!CRM32Pro.XMLAttributeGet(idXML,"name",&szTmp)) return -6;
    szCopy(szName,szTmp,sizeof(szName));

    // 5.Load the "header" attributes
    // 5.1.Memory allocations
    szSrcFile = new char[_MAX_FNAME];
    szSrcPath = new char[_MAX_PATH];
    szMSMName = new char[_MAX_FNAME];
    szAux = new char[_MAX_PATH + _MAX_FNAME];

    // 5.2.Building the right path+name for the MSM/DPF/Tilesets
    // szSrcFile = .msm or .dpf
    // szSrcPath = the path (if any) to to the .msm/.dpf
    // szMSMName = always the .msm filename	
    // sSrcDPF = the full path+name of the DPF, it set iSrcDPF to 1
    if(sSrcDPF == NULL)
    {
        _splitpath(sLevelFilename, NULL, szSrcPath, szSrcFile, NULL);
        strcat(szSrcFile,".msm");
        strcpy(szMSMName,szSrcFile);	
    }
    else
    {
        _splitpath(sLevelFilename, NULL, NULL, szMSMName, NULL);
        strcat(szMSMName,".msm");
        _splitpath(sSrcDPF,NULL,szSrcPath,szSrcFile,NULL); 
        strcat(szSrcFile,".dpf");
        iSrcDPF = 1; 
    }	
    ILogSystem.Msg(iLogN,"\n [MSTE] Info: trying to load '%s' level from '%s'%s\n",szName,szMSMName,(iSrcDPF)?" stored in a DPF...":"...");	
    // 5.3.Version check
    // ToDO
    // 5.4.Dates
    if(CRM32Pro.XMLAttributeGet(idXML,"creation_date",&szTmp) == 1) szCopy(szCreationDate,szTmp,sizeof(szCreationDate));
    if(CRM32Pro.XMLAttributeGet(idXML,"modification_date",&szTmp) == 1) szCopy(szModificationDate,szTmp,sizeof(szModificationDate));
    // 5.5.Viewport
    rectTmp.x = 0; rectTmp.y = 0; rectTmp.w = 0; rectTmp.h = 0;
    if(CRM32Pro.XMLAttributeGet(idXML,"viewport_x",&iTmp) == 1) rectTmp.x = iTmp;
    if(CRM32Pro.XMLAttributeGet(idXML,"viewport_width",&iTmp) == 1) rectTmp.w = iTmp;
    if(CRM32Pro.XMLAttributeGet(idXML,"viewport_y",&iTmp) == 1) rectTmp.y = iTmp;
    if(CRM32Pro.XMLAttributeGet(idXML,"viewport_height",&iTmp) == 1) rectTmp.h = iTmp;
    SetViewport(&rectTmp);

    // 6.Loop through the whole msse_level nodes looking for tilesets
    retvalue = CRM32Pro.XMLNodeChild(idXML);
    do
    {
        // Is it a tileset node?
        if(!strcmp(CRM32Pro.XMLNodeGetName(idXML),"tileset"))
        {
            iNumTS++;
            // tileset attributes
            retvalue = CRM32Pro.XMLAttributeGet(idXML,"gid",&iTmp);
            retvalue = CRM32Pro.XMLAttributeGet(idXML,"name",&szTmp);
            retvalue = CRM32Pro.XMLAttributeGet(idXML,"source",&szTmp2); 
            // Prepare the path+file to be loaded
            strcpy(szAux,szSrcPath);
            strcat(szAux,szTmp2);
            // load tileset with additional attributes (if applies, aka: external images)
            if(CheckFileExtension(szTmp2,".png") || CheckFileExtension(szTmp2,".bmp"))
            {
                int iW,iH,iCk,iA;
                retvalue = CRM32Pro.XMLAttributeGet(idXML,"tilewidth",&iW);
                retvalue = CRM32Pro.XMLAttributeGet(idXML,"tileheight",&iH);
                retvalue = CRM32Pro.XMLAttributeGet(idXML,"colorkey",&iCk);
                retvalue = CRM32Pro.XMLAttributeGet(idXML,"alphavalue",&iA);				
                if(LoadTileset(szAux,szTmp,iW,iH,iTmp) != -1)
                {
                    retvalue = CheckTile(iTmp);
                    vTilesets[retvalue]->tTileset->SetColorKey(iCk);
                    vTilesets[retvalue]->tTileset->SetAlpha(iA);
                    vTilesets[retvalue]->bResControl = TE_RESCONTROL_INTERNAL;
                    iNumTSL++;
                }
            }
            // load from a dpf
            else if(LoadTileset(szAux,szTmp,iTmp) != -1) 
            {
                retvalue = CheckTile(iTmp);
                vTilesets[retvalue]->bResControl = TE_RESCONTROL_INTERNAL;
                iNumTSL++;
            }
        }		
    } while(CRM32Pro.XMLNodeNext(idXML)); // next node

    // 7.Loop through the whole the msse_level nodes looking for layers
    retvalue = CRM32Pro.XMLNodeParent(idXML); // Point to msse_level
    retvalue = CRM32Pro.XMLNodeChild(idXML);  // And now, go to the first subnode
    do
    {
        // Is it a layer node?
        szTmp = CRM32Pro.XMLNodeGetName(idXML);
        if(!strcmp(CRM32Pro.XMLNodeGetName(idXML),"layer"))
        {
            // 7.1.Get and check the id
            int iStatus = 0;
            iNumLY++;
            retvalue = CRM32Pro.XMLAttributeGet(idXML,"id",&iID);
            if(retvalue != 1) continue;
            if((iID >= (int)vLayers.size()) || (iID < 0)) continue;
            if(vLayers[iID]->map != NULL) continue;

            // 7.2.Loading main attributes of this layer (all of them are required)
            if(CRM32Pro.XMLAttributeGet(idXML,"name",&szTmp) != 1) continue;
            szCopy(vLayers[iID]->szName,szTmp,sizeof(vLayers[iID]->szName));
            if(CRM32Pro.XMLAttributeGet(idXML,"status",&iStatus) != 1) continue; // We set the status when mapdata and tileset are loaded
            if(CRM32Pro.XMLAttributeGet(idXML,"parallax_ratio",&dTmp) != 1) continue;
            vLayers[iID]->fParallaxRatio = (float)dTmp;
            if(CRM32Pro.XMLAttributeGet(idXML,"position_x",&dTmp) != 1) continue;
            vLayers[iID]->fPosX = (float)dTmp;
            if(CRM32Pro.XMLAttributeGet(idXML,"position_y",&dTmp) != 1) continue;
            vLayers[iID]->fPosY = (float)dTmp;
            if(CRM32Pro.XMLAttributeGet(idXML,"speed_x",&dTmp) != 1) continue;
            vLayers[iID]->fSpeedX = (float)dTmp;
            if(CRM32Pro.XMLAttributeGet(idXML,"speed_y",&dTmp) != 1) continue;
            vLayers[iID]->fSpeedY = (float)dTmp;
            if(CRM32Pro.XMLAttributeGet(idXML,"gridcolor",&iTmp) != 1) iTmp = 0; // The gridcolor is not required, if it is not found, it is set to 0 (black color)
            vLayers[iID]->iGridColor = iTmp;
            if(CRM32Pro.XMLAttributeGet(idXML,"alphavalue",&iTmp) != 1) iTmp = 255; // The alpha value is not required, if it is not found, it is set to 255 (full opaque)
            vLayers[iID]->iAlpha = iTmp;

            // 7.3.Looking for mapdata nodes
            if(CRM32Pro.XMLNodeChild(idXML) != 1)
            {
                // We dont have the mapdata
                ILogSystem.Msg(iLogN," [MSTE] Warning: 'mapdata' subnode was not found while parsing layer ID '%d'\n",iID);
                continue;
            }

            // 7.4.Loop through all layers subnodes looking for 'mapdata'. We will load the first base mapdata found and the first custom mapdata (if any)
            int bMapBase = 0, bMapCustom = 0, bValidatedBase = 0;
            cMSTE_MapBase *myMap = NULL;
            do
            {
                // 7.4.1.Get a node and check for 'mapdata'
                szTmp = CRM32Pro.XMLNodeGetName(idXML);
                if(strcmp("mapdata",szTmp)) continue;

                // 7.4.2.Get the maptype
                if(!CRM32Pro.XMLAttributeGet(idXML,"type",&szTmp)) continue;

                // 7.4.3.After loading a base mapdata, avoid to load a possible second one. The same for custom mapdata types...
                if(!strcmp("base",szTmp) && bMapBase != 0) continue;				
                else if(strcmp("base",szTmp) && bMapCustom != 0) continue;

                // 7.4.4.Try to create a map of that type and call to the load method
                myMap = CreateMapDataInstance(szTmp,0,0);
                if(myMap == NULL) 
                {
                    ILogSystem.Msg(iLogN," [MSTE] Warning: handler for parsing layer ID '%d' mapdata type '%s' was not found.\n",iID,szTmp);
                    continue;
                }
                retvalue = myMap->LoadMapdataNode(idXML,0);	

                // 7.4.5.If everything was fine, mapdata loaded! 
                if(retvalue == 1)
                {
                    // Resource control set to internal
                    vLayers[iID]->bResControl = TE_RESCONTROL_INTERNAL;

                    // With 'base' mapdata type, try to validate it. It checks if a custom mapdata is already loaded...
                    if(!strcmp("base",szTmp)) 
                    {
                        if(bMapCustom)
                        {
                            cMSTE_MapBase *myMapTmp = vLayers[iID]->map;
                            vLayers[iID]->map = myMap;
                            bValidatedBase = ValidateMapBase(iID);	
                            if(AddCustomMapData(iID,myMapTmp)) bMapBase = 1;					
                            else 
                            { 
                                delete myMap; 
                                myMap = NULL; 
                                vLayers[iID]->map = myMapTmp;
                            }							
                        }
                        else
                        {
                            vLayers[iID]->map = myMap;
                            bValidatedBase = ValidateMapBase(iID);						
                            bMapBase = 1;
                        }
                    }
                    // Custom mapdata types. It checks if a base mapdata is already loaded.
                    else 
                    {
                        // If there is a base mapdata loaded...add the custom
                        if(bMapBase) 
                        {
                            if(AddCustomMapData(iID,myMap)) bMapCustom = 1;					
                            else { delete myMap; myMap = NULL; }
                        }
                        // It is just a custom mapdata
                        else 
                        {
                            vLayers[iID]->map = myMap;
                            bMapCustom = 1;
                        }
                    }
                }
                // 7.4.6.Error found while parsing
                else
                {
                    ILogSystem.Msg(iLogN," [MSTE] Warning: it was not possible to parse mapdata type '%s' found in layer ID '%d'\n",szTmp,iID);
                    vLayers[iID]->iStatus = TE_LAYERFLAG_DISABLE;
                    delete myMap;
                    myMap = NULL;
                }
            } while(CRM32Pro.XMLNodeNext(idXML) && ((bMapBase == 0) || (bMapCustom == 0)));

            // 7.5.Info about last layer
            if(bMapBase || bMapCustom)
            {
                iNumLYL++;
                ILogSystem.Msg(iLogN," [MSTE] Info: mapdata for layer '%s' with ID '%d' loaded:\n",vLayers[iID]->szName,iID);
                vLayers[iID]->map->PrintInfo(-1,iLogN);
                if(bValidatedBase == 1) 
                {
                    vLayers[iID]->iStatus = iStatus;
                    ILogSystem.Msg(iLogN,"         . Layer is ready\n");	
                }
                else 
                {
                    ILogSystem.Msg(iLogN,"         . Layer was disabled as it is not ready\n");	
                }
            }

            // 7.6.Point to parent node (layer)
            CRM32Pro.XMLNodeParent(idXML);
        }		
    } while(CRM32Pro.XMLNodeNext(idXML)); // next node

    // 8.End of process	
    if((iNumLY != iNumLYL) || (iNumTS != iNumTSL))
    {		
        ILogSystem.Msg(iLogN," [MSTE] Warning: some resources could not be loaded: tileset %d/%d and layers %d/%d\n\n",iNumTSL,iNumTS,iNumLYL,iNumLY);
        iTmp = 0;
    }
    else 
    {
        ILogSystem.Msg(iLogN," [MSTE] Info: '%s' level successfully loaded.\n\n",szName);
        iTmp = 1;
    }
    CRM32Pro.XMLClose(idXML);
    delete []szAux;
    delete []szSrcFile;
    delete []szSrcPath;
    delete []szMSMName;
    return iTmp;
}

//! Save current level on native format.
/*! It saves the whole level including mapdata and tilesets. If the level or tileset output files exist, they are overwritten but external images.
\param sLevelFilename Filename to save the level. The .msm extension will be automatically added in case it was missing. It supports optional paths.
\param sDPF Optional parameter. String with the DPF name to use as container for the level. It supports optional paths.
\param sLevelname Optional parameter. String with the new level name overriding the existing level name.
\return -1 Failed: engine was not initialized.
\return -2 Failed: sLevelFilename is not valid.
\return -3 Failed: internal error.
\return 1 Succeed */ 
int cMSTE::SaveLevel(char *sLevelFilename,char *sDPF, char *sLevelname)
{
    int	  retvalue, idXML, i, iDstDPF = 0, iTsDst = 0, iErrorFlag = 0;	
    char *szDstFile,*szDstPath, *szError, *szMSMName, *szTileset, *szAux;

    // 1.Initial checks
    if(!bTileEngine) return -1;
    if(sLevelFilename == NULL) return -2;
    // 1.1.Memory allocation
    szDstFile = new char[_MAX_FNAME];
    szDstPath = new char[_MAX_PATH];
    szError = new char[_MAX_FNAME];
    szAux = new char[_MAX_PATH + _MAX_FNAME];
    szMSMName = new char[_MAX_FNAME];
    szTileset = new char[_MAX_PATH + _MAX_FNAME];

    // 2.Create the MSM: set the "header" attributes. 	
    idXML = CRM32Pro.XMLCreate("msse_level");	
    CRM32Pro.XMLCommentSet(idXML,"MegaStorm TileEngine - CRM32Pro SDK");
    retvalue = CRM32Pro.XMLAttributeSet(idXML,"version",MSTE_VERSION);
    retvalue = CRM32Pro.XMLAttributeSet(idXML,"name",(sLevelname == NULL)?szName:sLevelname);	
    retvalue = CRM32Pro.XMLAttributeSet(idXML,"creation_date",szCreationDate);
    szCopy(szModificationDate,GetDateStamp(),sizeof(szModificationDate));
    retvalue = CRM32Pro.XMLAttributeSet(idXML,"modification_date",szModificationDate);	
    retvalue = CRM32Pro.XMLAttributeSet(idXML,"viewport_x",rViewport.x);
    retvalue = CRM32Pro.XMLAttributeSet(idXML,"viewport_width",rViewport.w);
    retvalue = CRM32Pro.XMLAttributeSet(idXML,"viewport_y",rViewport.y);
    retvalue = CRM32Pro.XMLAttributeSet(idXML,"viewport_height",rViewport.h);

    // 3.Building the right path+name for the MSM/DPF/Tilesets
    // szDstFile = .msm or .dpf
    // szDstPath = the path (if any) to szDstFile
    // szMSMName = always the .msm filename
    // szAux = path + name of the tileset (to be used in the source field)
    // szTileset = szDstPath + szAux (to save the tilesets)
    // szError = used for displaying errors
    if(sDPF == NULL)
    {
        _splitpath(sLevelFilename, NULL, szDstPath, szDstFile, NULL);
        strcat(szDstFile,".msm");
        strcpy(szMSMName,szDstFile);		
    }
    else
    {
        _splitpath(sLevelFilename, NULL, NULL, szMSMName, NULL);
        strcat(szMSMName,".msm");
        _splitpath(sDPF,NULL,szDstPath,szDstFile,NULL); 
        strcat(szDstFile,".dpf");
        iDstDPF = 1; // Pack the level into the DPF
    }	

    // 4.Loop through all layers creating a new node for the ones that have a mapdata
    for(i = 0; i < (int)vLayers.size(); i++)
    {
        // 4.1.Checking the existence of a mapdata		
        if(vLayers[i]->map == NULL) continue;

        // 4.2.Main attributes of this layer
        retvalue = CRM32Pro.XMLNodeCreate(idXML,"layer"); // Create a new node and point to it.
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"id",i);
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"name",vLayers[i]->szName);
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"status",(int)vLayers[i]->iStatus);
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"parallax_ratio",vLayers[i]->fParallaxRatio);
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"position_x",vLayers[i]->fPosX);
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"position_y",vLayers[i]->fPosY);
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"speed_x",vLayers[i]->fSpeedX);
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"speed_y",vLayers[i]->fSpeedY);
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"gridcolor",(int)vLayers[i]->iGridColor);
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"alphavalue",vLayers[i]->iAlpha);

        // 4.3.Call the mapdata save method
        vLayers[i]->map->SaveMapdataNode(idXML,0);		

        // 4.4.Point to father node(msse_level)		
        CRM32Pro.XMLNodeParent(idXML);
    }

    // 5.Loop through all tilesets creating a new node for each one
    for(i = 0; i < (int)vTilesets.size(); i++) // Assume an element in the vector always has a CRM32Pro_CTile assigned
    {	
        // 5.1.Prepare the source attribute
        // = DPF file, pack into the DPF
        // = BMP or PNG file, write to external file image
        // = TE_TILESET_DEFAULT_SOURCE, use the same storage as the level(external images or DPF)
        // szAux contain the source field with path+filename and iTsDst flag: 0->Error, 1->DPF, 2->BMP, 3->PNG		
        strcpy(szAux,vTilesets[i]->szSource);
        // = DPF
        if(CheckFileExtension(szAux,".dpf")) iTsDst = 1;
        // = BMP
        else if(CheckFileExtension(szAux,".bmp")) iTsDst = 2;
        // = PNG
        else if(CheckFileExtension(szAux,".png")) iTsDst = 3;
        // = TE_TILESET_DEFAULT_SOURCE
        else if(!strcmp(szAux,TE_TILESET_DEFAULT_SOURCE))
        {	
            // Use the level DPF file
            if(iDstDPF == 1) 
            {
                strcpy(szAux,szDstFile); 
                iTsDst = 1;
            }
            // ... or save to an external PNG
            else 
            {
                strcpy(szAux,vTilesets[i]->tTileset->GetName());
                strcat(szAux,".png"); 
                iTsDst = 3;
            }
        }
        else { return -3; } // We never should reach this point!

        // 5.2.Save main tileset attributes
        retvalue = CRM32Pro.XMLNodeCreate(idXML,"tileset"); // Create a new node and point to it.
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"gid",vTilesets[i]->iGID);
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"name",vTilesets[i]->tTileset->GetName());
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"source",szAux);

        // 5.3.Save additional tileset attributes if applies (external images)
        if(iTsDst != 1)
        {
            int iW = 0, iH = 0;
            vTilesets[i]->tTileset->GetOffset(&iW,&iH);
            retvalue = CRM32Pro.XMLAttributeSet(idXML,"tilewidth",iW);
            retvalue = CRM32Pro.XMLAttributeSet(idXML,"tileheight",iH);
            retvalue = CRM32Pro.XMLAttributeSet(idXML,"colorkey",vTilesets[i]->tTileset->GetColorKey());
            retvalue = CRM32Pro.XMLAttributeSet(idXML,"alphavalue",vTilesets[i]->tTileset->GetAlpha());
        }

        // 5.4.1.Prepare the paths..
        strcpy(szTileset,szDstPath);
        strcat(szTileset,szAux);

        // 5.4.2.Save the tileset to a DPF
        if(iTsDst == 1)
        {
            // To a DPF
            if(vTilesets[i]->tTileset->Save(szTileset) == -1)
            {
                // Try to remove a possible exiting tile and try again
                if(RemoveTile(szTileset,vTilesets[i]->tTileset->GetName()))
                {
                    if(vTilesets[i]->tTileset->Save(szTileset) == -1)
                    {
                        // Error!
                        iErrorFlag++;
                        sprintf(szError,"error_msg_%d",iErrorFlag);
                        retvalue = CRM32Pro.XMLNodeCreate(idXML,szError); // Create a new node and point to it.
                        sprintf(szError,"Tileset '%s' with ID '%d' could not be saved into the DPF '%s'",vTilesets[i]->tTileset->GetName(),vTilesets[i]->iGID,szTileset);
                        CRM32Pro.XMLCommentSet(idXML,szError);
                        ILogSystem.Msg(iLogN," [MSTE] Warning: can not save '%s' tileset with ID '%d' into the DPF '%s'\n",
                            vTilesets[i]->tTileset->GetName(),vTilesets[i]->iGID,szAux);
                        CRM32Pro.XMLNodeParent(idXML); // go back
                    }
                }
            }
        }

        // 5.4.3.Or save it to an external file
        else
        {
            // Handling possible colorkey
            int ck = vTilesets[i]->tTileset->GetColorKey();
            vTilesets[i]->tTileset->SetColorKey(-1); // Disable any possible colorkey

            // Save the image but do not overwrite any already existing file
            if(!fexists(szTileset)) 
            {
                if(iTsDst == 2) IImage->SaveBMP(szTileset,vTilesets[i]->tTileset->GetSurface());
                else if(iTsDst == 3) IImage->SavePNG(szTileset,vTilesets[i]->tTileset->GetSurface());
            }

            // Enable the possible colorkey
            vTilesets[i]->tTileset->SetColorKey(ck);
        }
        // 5.5.Next tileset		
        CRM32Pro.XMLNodeParent(idXML); // Point to father node(msse_level)
    }

    // 6.Save the MSM to a DPF or an external file	
    if(iDstDPF == 1) 
    {		
        strcpy(szAux,szDstPath);
        strcat(szAux,szDstFile);
        CRM32Pro.XMLSave(idXML, szAux, szMSMName);
    }	
    else
    {
        // Saving XML on an external file		
        strcpy(szAux,szDstPath);
        strcat(szAux,szDstFile);
        CRM32Pro.XMLSave(idXML,szAux);
    }

    // 7.Close and info message
    CRM32Pro.XMLClose(idXML);
    if(iErrorFlag == 0)
        ILogSystem.Msg(iLogN," [MSTE] Info: '%s' level successfully saved to '%s'.\n",(sLevelname == NULL)?szName:sLevelname,szAux);
    else 
        ILogSystem.Msg(iLogN," [MSTE] Info: '%s' level saved to '%s' but some issues were detected.\n",(sLevelname == NULL)?szName:sLevelname,szAux);

    // 8.Remove resources and return
    delete []szDstFile;
    delete []szDstPath;
    delete []szError;
    delete []szAux;
    delete []szMSMName;
    delete []szTileset;
    return 1;
}
