/*----------------------------------------------
   MegaStorm Tile-based Engine - Roberto Prieto
   Copyright (C) 2007-2012 MegaStorm Systems
  ----------------------------------------------
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
    
    Roberto Prieto
    contact@megastormsystems.com
    http://www.megastormsystems.com

    Used software licenses, including the one for this engine, are located
    at /licenses directory. Please, include them on your distributions.

 --------------------------------------------------------------------------
             MSTE Main engine class
    
    Changelog: read MSTE.h to see all changes

 -------------------------------------------------------------------------- */
#include "MSTE.h"
#include "MSTE_MapData.h"
#include "MSTE_Tools.h"

//-------------------------------------------------------------------------------
// ImportTMX() and ExportTMX() methods - IO Filters for Tiled Qt (TMX)
//-------------------------------------------------------------------------------

//! Export current level to TMX 1.0 file format (Tiled Qt)
/*! We add special MSM properties.<br>It saves the whole level including mapdata and tilesets. If the level or tileset output files exist, they are overwritten but external images.
\param sTMXFilename Filename to save the level. The .tmx extension will be automatically added in case it was missing. It supports optional paths.
\param sLevelname Optional parameter. String with the new level name overriding the existing level name.
\return -1 Failed: engine was not initialized.
\return -2 Failed: sTMXFilename is not valid.
\return -3 Failed: the level has not any valid layer.
\return -4 Failed: wrong tileset source value.
\return 1 Succeed */ 
int cMSTE::ExportTMX(char *sTMXFilename, char *sLevelname)
{
    int	  retvalue, idXML, i, iErrorFlag = 0;	
    char *szDstFile,*szDstPath, *szError, *szTileset, *szAux;
    st_layers *lP = NULL;

    // 1.1.Initial checks
    if(!bTileEngine) return -1;
    if(sTMXFilename == NULL) return -2;	
    // 1.2.Look for one valid layer
    for(i = 0; i < (int)vLayers.size(); i++)
    {		
        if(vLayers[i]->map != NULL) { lP = vLayers[i]; break; }
    }
    if(lP == NULL) return -3;
    // 1.3.Memory allocation
    szDstFile = new char[_MAX_FNAME];
    szDstPath = new char[_MAX_PATH];
    szError = new char[_MAX_FNAME];
    szAux = new char[_MAX_PATH + _MAX_FNAME];
    szTileset = new char[_MAX_PATH + _MAX_FNAME];

    // 2.Create the TMX v1.0: set the "header" attributes. 	
    idXML = CRM32Pro.XMLCreate("map");	
    CRM32Pro.XMLCommentSet(idXML,"MegaStorm TileEngine - CRM32Pro SDK");
    retvalue = CRM32Pro.XMLAttributeSet(idXML,"version","1.0");
    retvalue = CRM32Pro.XMLAttributeSet(idXML,"orientation","orthogonal");
    retvalue = CRM32Pro.XMLAttributeSet(idXML,"width",lP->map->GetMapWidth());
    retvalue = CRM32Pro.XMLAttributeSet(idXML,"height",lP->map->GetMapHeight());
    retvalue = CRM32Pro.XMLAttributeSet(idXML,"tilewidth",lP->map->GetCellWidth());
    retvalue = CRM32Pro.XMLAttributeSet(idXML,"tileheight",lP->map->GetCellHeight());
    // 2.1.Save MSM custom header vars as properties in the TMX
    retvalue = CRM32Pro.XMLNodeCreate(idXML,"properties");	// Start "properties" block
    retvalue = WriteTMXProperty(idXML,"version",MSTE_VERSION);
    retvalue = WriteTMXProperty(idXML,"name",(sLevelname == NULL)?szName:sLevelname);	
    retvalue = WriteTMXProperty(idXML,"creation_date",szCreationDate);
    szCopy(szModificationDate,GetDateStamp(),sizeof(szModificationDate));
    retvalue = WriteTMXProperty(idXML,"modification_date",szModificationDate);	
    retvalue = WriteTMXProperty(idXML,"viewport_x",rViewport.x);
    retvalue = WriteTMXProperty(idXML,"viewport_width",rViewport.w);
    retvalue = WriteTMXProperty(idXML,"viewport_y",rViewport.y);
    retvalue = WriteTMXProperty(idXML,"viewport_height",rViewport.h);
    CRM32Pro.XMLNodeParent(idXML); // End "properties" block

    // 3.Building the right path+name for the TMX
    // szDstFile = .tmx
    // szDstPath = the path (if any) to szDstFile
    // szAux = path + name of the tileset (to be used in the source field)
    // szTileset = szDstPath + szAux (to save the tilesets)
    // szError = used for displaying errors
    _splitpath(sTMXFilename, NULL, szDstPath, szDstFile, NULL);
    if(!CheckFileExtension(szDstPath,".tmx")) strcat(szDstFile,".tmx");

    // 4.Loop through all tilesets creating a new node for each one
    for(i = 0; i < (int)vTilesets.size(); i++) // Assume an element in the vector always has a CRM32Pro_CTile assigned
    {	
        // 5.1.Prepare the source attribute
        // szAux contain the source field with path+filename+extension and iTsDst flag: 0->Error, 1->BMP, 2->PNG		
        int iW = 0, iH = 0, iTsDst = 0;
        strcpy(szAux,vTilesets[i]->szSource);
        if(CheckFileExtension(szAux,".bmp")) iTsDst = 1;
        else if(CheckFileExtension(szAux,".png")) iTsDst = 2;
        else if(!strcmp(szAux,TE_TILESET_DEFAULT_SOURCE))
        {	
            strcpy(szAux,vTilesets[i]->tTileset->GetName());
            strcat(szAux,".png"); 
            iTsDst = 2;
        }
        else return -4; // We never should reach this point!

        // 5.2.Save main tileset attributes
        retvalue = CRM32Pro.XMLNodeCreate(idXML,"tileset"); // Create a new node and point to it.
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"firstgid",vTilesets[i]->iGID);
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"name",vTilesets[i]->tTileset->GetName());
        vTilesets[i]->tTileset->GetOffset(&iW,&iH);
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"tilewidth",iW);
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"tileheight",iH);
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"spacing",0);
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"margin",0);
        // 5.2.1.Save MSM custom tileset vars as properties in the TMX
        retvalue = CRM32Pro.XMLNodeCreate(idXML,"properties");	// Start "properties" block
        retvalue = WriteTMXProperty(idXML,"alphavalue",vTilesets[i]->tTileset->GetAlpha());
        CRM32Pro.XMLNodeParent(idXML); // End "properties" block
        // 5.2.2.Now save the "image" TMX subnode
        retvalue = CRM32Pro.XMLNodeCreate(idXML,"image");	// Start "image" subnode
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"source",szAux);
        retvalue = vTilesets[i]->tTileset->GetColorKey(); 
        if(retvalue != - 1) 
        {
            // trans in Hex format (00 00 00)
            char sTmp[8];
            _ltoa(retvalue,sTmp,16);
            CRM32Pro.XMLAttributeSet(idXML,"trans",sTmp);
        }
        iW = vTilesets[i]->tTileset->GetSurface()->w;
        iH = vTilesets[i]->tTileset->GetSurface()->h;
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"width",iW);
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"height",iH);		
        retvalue = CRM32Pro.XMLNodeParent(idXML);	// End "image" subnode

        // 5.3.Prepare the paths and save it to external PNG file..
        strcpy(szTileset,szDstPath);
        strcat(szTileset,szAux);
        // Handling possible colorkey
        int ck = vTilesets[i]->tTileset->GetColorKey();
        vTilesets[i]->tTileset->SetColorKey(-1); // Disable any possible colorkey
        // Save the image but do not overwrite any already existing file
        if(!fexists(szTileset)) 
        {
            if(iTsDst == 1) IImage->SaveBMP(szTileset,vTilesets[i]->tTileset->GetSurface());
            else IImage->SavePNG(szTileset,vTilesets[i]->tTileset->GetSurface());
        }
        // Enable the possible colorkey
        vTilesets[i]->tTileset->SetColorKey(ck);

        // 5.4.Next tileset		
        CRM32Pro.XMLNodeParent(idXML); // Point to father node(msse_level)
    }

    // 5.Loop through all layers creating a new node for the ones that have a mapdata
    for(i = 0; i < (int)vLayers.size(); i++)
    {
        // 4.1.Checking the existence of a mapdata and the same map sizes
        if(vLayers[i]->map == NULL) continue;
        if((vLayers[i]->map->GetMapHeight() != lP->map->GetMapHeight()) || (vLayers[i]->map->GetMapWidth() != lP->map->GetMapWidth()))
        {
            iErrorFlag++;
            continue;
        }

        // 4.2.Main attributes of this layer
        retvalue = CRM32Pro.XMLNodeCreate(idXML,"layer"); // Create a new node and point to it.
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"name",vLayers[i]->szName);
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"width",lP->map->GetMapWidth());
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"height",lP->map->GetMapHeight());		
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"opacity",(double)(vLayers[i]->iAlpha / 255.0));
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"visible",(vLayers[i]->iStatus & TE_LAYERFLAG_RENDER)?1:0);
        retvalue = CRM32Pro.XMLNodeCreate(idXML,"properties");	// Start "properties" block
        retvalue = WriteTMXProperty(idXML,"id",i);
        retvalue = WriteTMXProperty(idXML,"status",(int)vLayers[i]->iStatus);
        retvalue = WriteTMXProperty(idXML,"parallax_ratio",vLayers[i]->fParallaxRatio);
        retvalue = WriteTMXProperty(idXML,"position_x",vLayers[i]->fPosX);
        retvalue = WriteTMXProperty(idXML,"position_y",vLayers[i]->fPosY);
        retvalue = WriteTMXProperty(idXML,"speed_x",vLayers[i]->fSpeedX);
        retvalue = WriteTMXProperty(idXML,"speed_y",vLayers[i]->fSpeedY);
        retvalue = WriteTMXProperty(idXML,"gridcolor",(int)vLayers[i]->iGridColor);
        CRM32Pro.XMLNodeParent(idXML); // End "properties" block

        // 4.3.Call the mapdata save method
        vLayers[i]->map->SaveMapdataNode(idXML,1);		

        // 4.4.Point to father node(msse_level)		
        CRM32Pro.XMLNodeParent(idXML);
    }

    // 6.Save the TMX file		
    strcpy(szAux,szDstPath);
    strcat(szAux,szDstFile);
    CRM32Pro.XMLSave(idXML,szAux);

    // 7.Close and info message
    CRM32Pro.XMLClose(idXML);
    if(iErrorFlag == 0)
        ILogSystem.Msg(iLogN," [MSTE] Info: '%s' level successfully exported to '%s'.\n",(sLevelname == NULL)?szName:sLevelname,szAux);
    else 
        ILogSystem.Msg(iLogN," [MSTE] Warning: '%s' level exported to '%s' but some issues were detected.\n",(sLevelname == NULL)?szName:sLevelname,szAux);

    // 8.Remove resources and return
    delete []szDstFile;
    delete []szDstPath;
    delete []szError;
    delete []szAux;
    delete []szTileset;
    return 1;
}

//! Import a TMX 1.0 file (Tiled Qt) on current TileEngine instance
/*! We can import "raw" TMX files or with special MSM properties.<br>
If the new layer or tileset IDs are already in use, they are not loaded, however, the viewport, creation/modification dates and the current level name are overwritten.<br>
\param sTMXFilename Filename to load the level. The .tmx extension will be automatically added in case it was missing. It supports optional paths.
\return -1 Failed: engine was not initialized.
\return -2 Failed: sTMXFilename is not valid.
\return -3 Failed: sTMXFilename could not be loaded.
\return -4 Failed: a 'map' node or main 'map' attributes were not found or were wrong.
\return 0 Warning: some tilesets and/or layers were not loaded. Check the log to further information.
\return 1 Succeed (all tilesets and layers found on the level loaded) */ 
int cMSTE::ImportTMX(char *sTMXFilename)
{
    int		idXML, retvalue, iTmp = 0;
    char	*szSrcPath = NULL, *szSrcFile = NULL, *szAux = NULL;
    int		iNumTS = 0, iNumTSL = 0; // Number of Tileset and tileset loaded
    int		iNumLY = 0, iNumLYL = 0; // Number of layers and layers loaded	
    char    *szTmp = NULL, *szTmp2 = NULL;
    int		iMX = 0, iMY = 0, iTX = 0, iTY = 0; 
    SDL_Rect rectTmp;

    // 1.Initial checks
    if(!bTileEngine) return -1;
    if(sTMXFilename == NULL) return -2;

    // 2.Load the XML that contains the .tmx
    idXML = CRM32Pro.XMLOpen(sTMXFilename);
    if(idXML == 0) return -3;

    // 3.Point to the first map node and load the main attributes
    if(CRM32Pro.XMLNodePointTo(idXML,1,"map") == 0) return -4;
    if(CRM32Pro.XMLAttributeGet(idXML,"version",&szTmp) == 0) return -4;
    if(strcmp(szTmp,"1.0")) return -4;
    if(CRM32Pro.XMLAttributeGet(idXML,"orientation",&szTmp) == 0) return -4;
    if(strcmp(szTmp,"orthogonal")) return -4;
    if(CRM32Pro.XMLAttributeGet(idXML,"width",&iMX) == 0) return -4; // Used for the data width
    if(CRM32Pro.XMLAttributeGet(idXML,"height",&iMY) == 0) return -4; // Used for the data height
    if(CRM32Pro.XMLAttributeGet(idXML,"tilewidth",&iTX) == 0) return -4; // Used when the tileset or spcial layer properties dont contain the width
    if(CRM32Pro.XMLAttributeGet(idXML,"tileheight",&iTY) == 0) return -4; // Used when the tileset or spcial layer properties dont contain the height

    // 4.1.Memory allocations and viewport init
    szSrcFile = new char[_MAX_FNAME];
    szSrcPath = new char[_MAX_PATH];
    szAux = new char[_MAX_PATH + _MAX_FNAME];
    rectTmp.x = 0; rectTmp.y = 0; rectTmp.w = 0; rectTmp.h = 0;
    // 4.2.Building the right path+name for the TMX
    // szSrcFile = .tmx
    // szSrcPath = the path (if any) to szSrcFile
    _splitpath(sTMXFilename, NULL, szSrcPath, szSrcFile, NULL);
    if(!CheckFileExtension(szSrcPath,".tmx")) strcat(szSrcFile,".tmx");	
    ILogSystem.Msg(iLogN,"\n [MSTE] Info: trying to import '%s' TMX file\n",sTMXFilename);	

    // 5.Loop through all map subnodes looking for special msse_level properties
    szTmp2 = szSrcFile; // Default level name
    retvalue = CRM32Pro.XMLNodeChild(idXML);
    do
    {
        // Is it a properties node?
        if(!strcmp(CRM32Pro.XMLNodeGetName(idXML),"properties"))
        {
            // Try to load our special msse_level properties
            if(ReadTMXProperty(idXML,"name",&szTmp)) szTmp2 = szTmp;
            if(ReadTMXProperty(idXML,"creation_date",&szTmp)) szCopy(szCreationDate,szTmp,sizeof(szCreationDate));
            if(ReadTMXProperty(idXML,"modification_date",&szTmp)) szCopy(szModificationDate,szTmp,sizeof(szModificationDate));
            // ToDO: Version check
            if(ReadTMXProperty(idXML,"viewport_x",&iTmp) == 1) rectTmp.x = iTmp;
            if(ReadTMXProperty(idXML,"viewport_width",&iTmp) == 1) rectTmp.w = iTmp;
            if(ReadTMXProperty(idXML,"viewport_y",&iTmp) == 1) rectTmp.y = iTmp;
            if(ReadTMXProperty(idXML,"viewport_height",&iTmp) == 1) rectTmp.h = iTmp;
        }		
    } while(CRM32Pro.XMLNodeNext(idXML)); // next map subnode
    SetViewport(&rectTmp);
    szCopy(szName,szTmp2,sizeof(szName));


    // 6.Loop through all map subnodes nodes looking for tilesets
    retvalue = CRM32Pro.XMLNodeParent(idXML); // Point to map node
    retvalue = CRM32Pro.XMLNodeChild(idXML);  // And now, go to the first subnode
    do
    {
        // Is it a tileset node?
        if(!strcmp(CRM32Pro.XMLNodeGetName(idXML),"tileset"))
        {
            // Special tileset vars
            int iW = 0, iH = 0, iCk = -1, iA = 255, iSpacing = 0, iMargin = 0;
            iNumTS++;

            // Tileset attributes
            if(CRM32Pro.XMLAttributeGet(idXML,"firstgid",&iTmp) != 1) continue;
            if(CRM32Pro.XMLAttributeGet(idXML,"name",&szTmp) != 1) continue;
            if(CRM32Pro.XMLAttributeGet(idXML,"tilewidth",&iW) != 1) iW = iTX;
            if(CRM32Pro.XMLAttributeGet(idXML,"tileheight",&iH) != 1) iH = iTY;
            retvalue = CRM32Pro.XMLAttributeGet(idXML,"spacing",&iSpacing); 
            retvalue = CRM32Pro.XMLAttributeGet(idXML,"margin",&iMargin);
            // Look for properties or image subnodes
            int iTileLoaded = 0; // Just load the first tile image
            retvalue = CRM32Pro.XMLNodeChild(idXML);
            do
            {
                // Is it a properties node?
                if(!strcmp(CRM32Pro.XMLNodeGetName(idXML),"properties"))
                {
                    // Try to load our special tileset properties
                    ReadTMXProperty(idXML,"alphavalue",&iA);
                }		
                // Is it a image node?
                if((!strcmp(CRM32Pro.XMLNodeGetName(idXML),"image")) && (!iTileLoaded))
                {
                    // trans in Hex format (00 00 00)
                    if(CRM32Pro.XMLAttributeGet(idXML,"trans",&szTmp2))	iCk = strtol(szTmp2,NULL,16); 
                    if(CRM32Pro.XMLAttributeGet(idXML,"source",&szTmp2))
                    {
                        strcpy(szAux,szSrcPath);
                        strcat(szAux,szTmp2);
                        // Only support PNG and BMP images
                        if(CheckFileExtension(szTmp2,".png") || CheckFileExtension(szTmp2,".bmp"))
                        {				
                            // Check for spacing & margins
                            if((iSpacing != 0) || (iMargin != 0))
                            {
                                CRM32Pro_CTile *myTile;
                                // Convert to spacing=0 and margins=0
                                myTile = ConvertToTileset(szAux,szTmp,iW,iH,iSpacing,iMargin);
                                if(LoadTileset(myTile,iTmp) != -1) iTileLoaded = 2; // Set the source name change on
                            }
                            // Load from the external image
                            else if(LoadTileset(szAux,szTmp,iW,iH,iTmp) != -1) iTileLoaded = 1; 
                            // Tileset loaded. Just load the first tile image
                            if(iTileLoaded)
                            {
                                int iID = CheckTile(iTmp);
                                vTilesets[iID]->tTileset->SetColorKey(iCk);
                                vTilesets[iID]->tTileset->SetAlpha(iA);
                                vTilesets[iID]->bResControl = TE_RESCONTROL_INTERNAL;
                                // Allow a possible future ExportTMX() to write to a different file image than the source one when it was converted
                                if(iTileLoaded == 2) 
                                {
                                    strcpy(szAux,vTilesets[iID]->tTileset->GetName());
                                    strcat(szAux,"-converted.png");
                                    SetTilesetSource(iTmp,szAux);
                                }
                                iNumTSL++;
                            }
                        }
                    }
                }	
            } while(CRM32Pro.XMLNodeNext(idXML)); // next tileset subnode
            // Go back if we entered on the properties or image subnode
            if(retvalue) CRM32Pro.XMLNodeParent(idXML);
        }		
    } while(CRM32Pro.XMLNodeNext(idXML)); // next map subnode

    // 7.Loop through all map subnodes nodes looking for looking for layers
    retvalue = CRM32Pro.XMLNodeParent(idXML); // Point to map node
    retvalue = CRM32Pro.XMLNodeChild(idXML);  // And now, go to the first subnode
    do
    {
        // Is it a layer node?
        szTmp = CRM32Pro.XMLNodeGetName(idXML);
        if(!strcmp(CRM32Pro.XMLNodeGetName(idXML),"layer"))
        {
            // Special layer vars (default values)
            char szBase[] = "base";
            int iID = -1, iGridColor = 0, iStatus = TE_LAYERFLAG_UPDATE, bValidatedBase = 0;
            double dPosX = 0, dPosY = 0, dSpeedX = 0, dSpeedY = 0, dParallax = 1, dOpacity = 1;
            cMSTE_MapBase *myMap = NULL;
            int bMap = 0; // 0: no map loaded, 1: map loaded
            szTmp2 = szBase;
            iNumLY++;

            // 7.1.Get the name or abort this layer.
            // Width and height if present, are discarded as indicated by Tiled Qt doc.
            // Read possible visible and opacity fields
            if(CRM32Pro.XMLAttributeGet(idXML,"name",&szTmp) != 1) continue;
            CRM32Pro.XMLAttributeGet(idXML,"opacity",&dOpacity);
            retvalue = 1; // By default in case visible tag is not on TMX
            CRM32Pro.XMLAttributeGet(idXML,"visible",&retvalue);
            if(retvalue >= 1) iStatus = iStatus | TE_LAYERFLAG_RENDER;

            // 7.2.Look for properties subnodes
            retvalue = CRM32Pro.XMLNodeChild(idXML);
            do
            {
                // Is it a properties node?
                if(!strcmp(CRM32Pro.XMLNodeGetName(idXML),"properties"))
                {
                    // Try to load our special layer properties
                    ReadTMXProperty(idXML,"id",&iID);
                    ReadTMXProperty(idXML,"gridcolor",&iGridColor);
                    ReadTMXProperty(idXML,"status",&iStatus);
                    ReadTMXProperty(idXML,"position_x",&dPosX);
                    ReadTMXProperty(idXML,"position_y",&dPosY);
                    ReadTMXProperty(idXML,"speed_x",&dSpeedX);
                    ReadTMXProperty(idXML,"speed_y",&dSpeedY);
                    ReadTMXProperty(idXML,"parallax_ratio",&dParallax);
                    ReadTMXProperty(idXML,"custom-type",&szTmp2);
                }
            } while(CRM32Pro.XMLNodeNext(idXML)); // next layer subnode
            // Go back if we entered on the properties or data subnode
            if(retvalue) CRM32Pro.XMLNodeParent(idXML);

            // 7.3.Find a valid ID
            if(iID != -1) // loaded from the special layer properties
            {
                if(iID >= (int)vLayers.size()) continue;
                if(vLayers[iID]->map != NULL) continue;				
            }
            else // find the first ID available
            {
                for(iID = 0; iID < (int)vLayers.size(); iID++) if(vLayers[iID]->map == NULL) break;
                if(iID == (int)vLayers.size()) continue;
            }		
            // From here, iID is our valid layer ID!. szTmp has the layer name and szTmp2 point to the possible custom map (if NULL, then is a base map)

            // 7.4.Look for data. We will load the last valid one following Tiled Qt philosophy.
            retvalue = CRM32Pro.XMLNodeChild(idXML);
            do
            {	
                // Is it a data node?
                if(!strcmp(CRM32Pro.XMLNodeGetName(idXML),"data"))
                {
                    // Create a map
                    bMap = 0;
                    if(myMap != NULL) { delete myMap; myMap = NULL; }
                    myMap = CreateMapDataInstance(szTmp2,iMY,iMX); // Use iMX and iMY for the TMX
                    if(myMap == NULL) 
                    {
                        ILogSystem.Msg(iLogN," [MSTE] Warning: handler for parsing '%s' layer with data type '%s' was not found.\n",szTmp,szTmp2);						
                        continue;
                    }
                    myMap->SetCellWidth(iTX); myMap->SetCellHeight(iTY); // Set cell sizes

                    if(myMap->LoadMapdataNode(idXML,1))
                    {
                        vLayers[iID]->map = myMap;
                        bValidatedBase = ValidateMapBase(iID);						
                        bMap = 1;
                    }
                    // Error found while parsing
                    else
                    {
                        ILogSystem.Msg(iLogN," [MSTE] Warning: it was not possible to parse '%s' layer with data type '%s'\n",szTmp,szTmp2);
                        delete myMap;
                        myMap = NULL;
                    }
                }	
            } while(CRM32Pro.XMLNodeNext(idXML)); // next layer subnode
            // Go back if we entered on the properties or data subnode
            if(retvalue) CRM32Pro.XMLNodeParent(idXML);

            // 7.5.Set loaded layer parameters
            if(bMap)
            {
                iNumLYL++;
                szCopy(vLayers[iID]->szName,szTmp,sizeof(vLayers[iID]->szName));
                vLayers[iID]->iGridColor = iGridColor;
                vLayers[iID]->iAlpha = (int)(dOpacity * 255.0);
                // Alpha value of 128 is faster so we check if loaded and converted alpha is +-5 to 128 and modify it
                if(abs(vLayers[iID]->iAlpha - 128) <= 5) vLayers[iID]->iAlpha = 128;
                vLayers[iID]->fPosX = (float)dPosX;
                vLayers[iID]->fPosY = (float)dPosY;
                vLayers[iID]->fSpeedX = (float)dSpeedX;
                vLayers[iID]->fSpeedY = (float)dSpeedY;
                vLayers[iID]->fParallaxRatio = (float)dParallax;
                vLayers[iID]->bResControl = TE_RESCONTROL_INTERNAL;
                ILogSystem.Msg(iLogN," [MSTE] Info: mapdata for layer '%s' with ID '%d' loaded:\n",vLayers[iID]->szName,iID);
                vLayers[iID]->map->PrintInfo(-1,iLogN);
                if(bValidatedBase == 1) 
                {
                    vLayers[iID]->iStatus = iStatus;
                    ILogSystem.Msg(iLogN,"         . Layer is ready\n");	
                }
                else 
                {
                    ILogSystem.Msg(iLogN,"         . Layer is not ready\n");						 
                    // Remove update and render if present
                    if(iStatus & TE_LAYERFLAG_RENDER) iStatus = iStatus ^ TE_LAYERFLAG_RENDER;
                    if(iStatus & TE_LAYERFLAG_UPDATE) iStatus = iStatus ^ TE_LAYERFLAG_UPDATE;
                    vLayers[iID]->iStatus = iStatus;
                }
            }
        }		
    } while(CRM32Pro.XMLNodeNext(idXML)); // next node

    // 8.End of process	
    if((iNumLY != iNumLYL) || (iNumTS != iNumTSL))
    {		
        ILogSystem.Msg(iLogN," [MSTE] Warning: some resources could not be imported: tileset %d/%d and layers %d/%d\n\n",iNumTSL,iNumTS,iNumLYL,iNumLY);
        iTmp = 0;
    }
    else 
    {
        ILogSystem.Msg(iLogN," [MSTE] Info: '%s' level successfully imported.\n\n",szName);
        iTmp = 1;
    }
    CRM32Pro.XMLClose(idXML);
    delete []szAux;
    delete []szSrcFile;
    delete []szSrcPath;
    return iTmp;
}

