/*----------------------------------------------
   MegaStorm Tile-based Engine - Roberto Prieto
   Copyright (C) 2007-2012 MegaStorm Systems
  ----------------------------------------------
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
    
    Roberto Prieto
    contact@megastormsystems.com
    http://www.megastormsystems.com

    Used software licenses, including the one for this engine, are located
    at /licenses directory. Please, include them on your distributions.

 --------------------------------------------------------------------------
             Kraftwerk MapData class
    
    Changelog: read MSTE_MapData.h to see all changes
   
 -------------------------------------------------------------------------- */
#include <stdio.h>
#include <memory.h>
#include "MSTE_MapData.h"
#include "MSTE_Tools.h"

//-------------------------------------------------------------------------------
// MSTE MapData base class - General stuff methods
//-------------------------------------------------------------------------------

//! MapData constructor
cMSTE_MapData::cMSTE_MapData()
{
    iCellWidth = 0;
    iCellHeight = 0;
    iMapWidth = 0;
    iMapHeight = 0;
    strcpy(sMapType,"empty");
}

//! MapData destructor
cMSTE_MapData::~cMSTE_MapData() 
{ 	
    iCellWidth = 0;
    iCellHeight = 0;
    iMapWidth = 0;
    iMapHeight = 0;
    strcpy(sMapType,"");
}

//-------------------------------------------------------------------------------
// MSTE MapData base class - Properties methods
//-------------------------------------------------------------------------------

//! Get map type
/*! \return Pointer to the string with the map type name */
char *cMSTE_MapData::GetMapType() { return sMapType; }

//! Get cell width
/*!\return integer with the cell width (in pixels)*/
int cMSTE_MapData::GetCellWidth() { return iCellWidth; }

//! Set cell width
/*! \param size integer with cell width (in pixels)*/
void cMSTE_MapData::SetCellWidth(int size) { iCellWidth = size; }

//! Get cell height
/*! \return integer with the cell height (in pixels)*/
int cMSTE_MapData::GetCellHeight() { return iCellHeight; }

//! Set tile height
/*! \param size integer with cell height (in pixels)*/
void cMSTE_MapData::SetCellHeight(int size) { iCellHeight = size; }

//! Get map width
/*! \return width of this map(in number of tiles)*/
int cMSTE_MapData::GetMapWidth() { return iMapWidth; }

//! Get map height
/*! \return height of this map(in number of tiles)*/
int cMSTE_MapData::GetMapHeight() { return iMapHeight; }


//-------------------------------------------------------------------------------
// MSTE MapBase derivated class - General stuff methods
//-------------------------------------------------------------------------------

//! MapBase default constructor
cMSTE_MapBase::cMSTE_MapBase()
{
    iTilemap = NULL; 
    strcpy(sMapType,"base");
}

//! MapBase mandatory constructor
/*! It will create a 2D array for the cells allocation. Each element on this 2D array is a cell.<br>
When you inherit from cMSTE_MapBase to create a custom map you HAVE TO create a constructor to overload<br>
this one but do not forget to call this mandatory constructor from your own constructor.
\param height Selected height of your map. 0 avoid the iTilemap creation.
\param width Selected width of your map. 0 avoid the iTilemap creation. */
cMSTE_MapBase::cMSTE_MapBase(int height,int width)
{
    if(width > 0 && height > 0)	
    {
        iTilemap = Create2DArray<unsigned int>(height,width);
        iMapWidth = width;
        iMapHeight = height;
        SetMapBaseValue(-1,-1,0);
    }
    else iTilemap = NULL;
    strcpy(sMapType,"base");
}

//! MapBase destructor
cMSTE_MapBase::~cMSTE_MapBase() 
{ 
    if(iTilemap != NULL) Delete2DArray(iTilemap); 
    iTilemap = NULL; 	
}

//! Create an instance of this class. 
/*! When you inherit from cMSTE_MapBase to create a custom map you HAVE TO create a method to overload<br>
this one creating an instance of your own derivated class. This is mandatory.
\param y Value to be passed to the mandatory constructor
\param x Value to be passed to the mandatory constructor
\return A cMSTE_MapBase pointer to the created instance */
cMSTE_MapBase *cMSTE_MapBase::createInstance(int y, int x)
{
    return new cMSTE_MapBase(y,x);
}

//-------------------------------------------------------------------------------
// MSTE MapBase derivated class - MapBase custom methods
//-------------------------------------------------------------------------------

//! Create the 2D array for the tiles allocation(mapbase).
/*! Each element on this 2D array is a cell.
\param height Selected height of your map
\param width Selected width of your map 
\param value Default value to be assigned on each cell.
\return 0 Failed
\return 1 Succeed*/
int cMSTE_MapBase::CreateMapBase(int height, int width, short int value)
{
    if(iTilemap != NULL) return 0;
    iTilemap = Create2DArray<unsigned int>(height,width);
    iMapWidth = width;
    iMapHeight = height;
    SetMapBaseValue(-1,-1,value);
    return 1;
}

//! Delete the 2D array(mapbase)
/*! \return 0 Failed
\return 1 Succeed */
int cMSTE_MapBase::DeleteMapBase()
{
    if(iTilemap != NULL) 
    {
        Delete2DArray(iTilemap);
        iTilemap = NULL;
        iMapWidth = 0;
        iMapHeight = 0;
        return 1;
    }
    return 0;
}

//! Set arbitrary values on 2D array(mapbase)
/*! It is mainly used creating manual maps
\param y Selected Y position. -1 to affect the whole axis
\param x Selected X position. -1 to affect the whole axis
\param value The value to be set on the given position
\return 0 Failed
\return 1 Succeed */
int cMSTE_MapBase::SetMapBaseValue(int y, int x, short int value)
{
    int yy,xx;

    // Previous check
    if(iTilemap == NULL) return 0;
    if(iMapWidth == 0) return 0;
    if(iMapHeight == 0) return 0;
    if(y < -1) y = -1;
    if(y > iMapHeight) y = iMapHeight;
    if(x < -1) x = -1;
    if(x > iMapWidth) x = iMapWidth;
    if(value < 0) value = 0;

    // Loop filling the gaps
    for(yy = 0; yy < iMapHeight; yy++)
    {
        if((yy == y) || (y == -1))
        {
            for(xx = 0; xx < iMapWidth; xx++)
            {
                if((xx == x) || (x == -1)) iTilemap[yy][xx] = value;
            }
        }
    }
    return 1;
}

//! Copy the current values of the 2D array iTilemap to the given MapBase
/*! The destination must have identical attributes (iMapWidth and iMapHeight).
If the destination has not an iTilemap we will create one identical to the source
\param dest Pointer to destination cMSTE_MapBase or derivated object
\return 0 Failed
\return 1 Succeed */
int cMSTE_MapBase::CloneMapBase(cMSTE_MapBase *dest)
{
    // Previous check
    if(iTilemap == NULL) return 0;
    if(iMapWidth == 0) return 0;
    if(iMapHeight == 0) return 0;
    if(dest == NULL) return 0;	

    // If there is not a destination iTilemap, we create it
    if(dest->iTilemap == NULL) 
    {
        dest->iTilemap = dest->Create2DArray<unsigned int>(iMapHeight,iMapWidth);
        dest->iMapHeight = iMapHeight;
        dest->iMapWidth = iMapWidth;
    }

    // Map size check
    if(iMapWidth != dest->iMapWidth) return 0;
    if(iMapHeight != dest->iMapHeight) return 0;

    // Copy data
    for(int i = 0; i < iMapHeight; i++)
        for(int j = 0; j < iMapWidth; j++)
            dest->iTilemap[i][j] = iTilemap[i][j];

    // Copy rest of attributes
    dest->iCellWidth = iCellWidth;
    dest->iCellHeight = iCellHeight;

    // return
    return 1;
}


//-------------------------------------------------------------------------------
// Generic Map Access methods
//-------------------------------------------------------------------------------

//! Map validation
/*! It validates the iTilemap 2D array(mapbase):<ul>
<li>If there is any <0, it will print info with the error</li>
<li>Check the mapbase width and height</li>
<li>Check the mapbase cell width and height</li></ul>
When you use custom mapdata (inherit from cMSTE_MapBase), you can overload this method to check your custom mapdata but you <b>must</b> call first this method.
\return 0 Failed
\return 1 Succeed*/
int cMSTE_MapBase::CheckMap()
{
    int iRet = 1;

    // Only for "base" maps. If a derivated custom mapdata has not implemented its custom CheckMap(),
    // we will reach this method and we can not check a custom mapdata so, just check it and return 
    if(strcmp(sMapType,"base")) return 1;
  
    // Previous check
    if(iTilemap == NULL) return 0;
    if(iMapWidth == 0) return 0;
    if(iMapHeight == 0) return 0;
    if(iCellWidth == 0) return 0;
    if(iCellHeight == 0) return 0;

    // Loop through iTilemap looking for <0 values
    for(int i = 0; i < iMapHeight; i++)
        for(int j = 0; j < iMapWidth; j++)
        {
            if(iTilemap[i][j] < 0) 
            { 
                ILogSystem.Msg(LOG_NORMAL," [MSTE MapBase] Error: map[%d][%d] has a wrong value of %d.\n",i,j,iTilemap[i][j]); 				
                iRet = 0; 
            }			
        }	
        // Return
        return iRet;
}

//! Print using ILogSystem some details of this mapdata
/*! When you use custom mapdata (inherit from cMSTE_MapBase), you can overload this method to show your custom information but you <b>should</b> call first this method.
\param bTilesetReady -1 do not display anything, 0 it is not ready and 1 it is ready
\param iLogLevel MSTE verbosity level
\return 0 Failed
\return 1 Succeed*/
int cMSTE_MapBase::PrintInfo(char bTilesetReady, int iLogLevel)
{
    if(iTilemap == NULL) return 0;
    ILogSystem.Msg(iLogLevel,"         . Map type 'base' - Map size %dx%d - Cell size %dx%d",iMapWidth,iMapHeight,iCellWidth,iCellHeight);
    if(bTilesetReady >= 0) ILogSystem.Msg(iLogLevel," - Tilesets ready: %s\n", (bTilesetReady==0)?"no":"yes");
    else ILogSystem.Msg(iLogLevel,"\n");
    return 1;
}

//! Save the mapdata base type to a XML node of a MSM or TMX file
/*! When you use custom mapdata (inherit from cMSTE_MapBase), you can overload this method to save your custom mapdata but you <b>must</b> call first this method. 
\param idXML id of an opened XML file with the pointer inside a layer subnode
\param iOutput integer with 0 for MSM or 1 for TMX files.
\return 0 Failed
\return 1 Succeed*/
int cMSTE_MapBase::SaveMapdataNode(int idXML, int iOutput)
{
    char	*szTmp = NULL;
    int		retvalue = 0;

    // 1.Select our type
    switch(iOutput)
    {
        // TMX output
    case 1:
        // A1.Create TMX node
        retvalue = CRM32Pro.XMLNodeCreate(idXML,"data"); // Create a new node and point to it.
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"encoding","csv");		
        break;

        // MSM output
    default:
    case 0:
        // B1.We should receive the idXML pointing inside a layer node
        retvalue = CRM32Pro.XMLNodeCreate(idXML,"mapdata"); // Create a new node and point to it.
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"type","base");
        // B2.Our attributes
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"width",iMapWidth);
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"height",iMapHeight);
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"cellwidth",iCellWidth);
        retvalue = CRM32Pro.XMLAttributeSet(idXML,"cellheight",iCellHeight);		
        break;
    }	

    // 2.Save our data (iTilemap[][] in CSV format)
    retvalue = A2DtoCSV(iTilemap,iMapWidth,iMapHeight,szTmp);
    if(retvalue)
    {
        CRM32Pro.XMLTextSet(idXML,szTmp);
        delete []szTmp; szTmp = NULL;
    }	

    // 3.Point back and return
    CRM32Pro.XMLNodeParent(idXML); // To our layer
    return retvalue;
}

//! Load the mapdata base type from a node inside a MSM or TMX file
/*! When you use custom mapdata (inherit from cMSTE_MapBase), you can overload this method to load your custom mapdata and you <b>must not</b> call this method.  
\param idXML id of an opened XML file with the pointer inside a mapdata/data subnode
\param iInput integer with 0 for MSM or 1 for TMX files.
\return 0 Failed
\return 1 Succeed*/
int cMSTE_MapBase::LoadMapdataNode(int idXML, int iInput)
{
    int iW = 0, iH = 0, iCW = 0, iCH = 0;
    char *szTmp = NULL, *szData = NULL;

    // 1.Select our type
    switch(iInput)
    {
        // TMX input
    case 1:
        // A1.Load TMX node
        if(CRM32Pro.XMLAttributeGet(idXML,"encoding",&szTmp) != 1) return 0;
        szData = CRM32Pro.XMLTextGet(idXML);
        if(szData == NULL) return 0;
        // A2.CSV support
        if(!strcmp(szTmp,"csv")) 
        {
            if(CSVtoA2D(iTilemap,iMapWidth,iMapHeight,szData)) return 1;
        }
        // A3.Base64 support. No compression supported
        else if(!strcmp(szTmp,"base64"))
        {	
            std::string szUnencoded;
            // Base64 decode
            szUnencoded = decodeBase64(szData);
            // Support compressed maps
            if(CRM32Pro.XMLAttributeGet(idXML,"compression",&szTmp) == 1) 
            {						
                szData = decompress(szUnencoded,iMapWidth * iMapHeight * 4);
                if(szData == NULL)
                {
                    ILogSystem.Msg(LOG_NORMAL," [MSTE] Warning: base64 input data is compressed using '%s' which is not a supported format\n",szTmp);
                    return 0;
                }
                if(StringtoA2D(iTilemap,iMapWidth,iMapHeight,szData))
                {
                    free(szData);
                    return 1;
                }
                free(szData);
            }
            // Uncompressed maps
            else 
            { 
                szData = (char*)szUnencoded.c_str(); 
                if(StringtoA2D(iTilemap,iMapWidth,iMapHeight,szData)) return 1; 
            }
            return 0;
        }	
        break;

        // MSM input
    default:
    case 0:
        // B1.Get attributes
        CRM32Pro.XMLAttributeGet(idXML,"width",&iW);
        CRM32Pro.XMLAttributeGet(idXML,"height",&iH);
        CRM32Pro.XMLAttributeGet(idXML,"cellwidth",&iCW);
        CRM32Pro.XMLAttributeGet(idXML,"cellheight",&iCH);
        // B2.Some validations
        if(iW <=0 || iH <= 0) return 0;
        if(iCW <=0 || iCH <= 0) return 0;
        // B3.Create the iTilemap and load the base data
        szTmp = CRM32Pro.XMLTextGet(idXML);
        if(szTmp == NULL) return 0;
        iCellWidth = iCW;
        iCellHeight = iCH;
        CreateMapBase(iH,iW);
        if(CSVtoA2D(iTilemap,iW,iH,szTmp)) return 1;
        break;
    }

    // 2.End
    return 0;
}

//! Convert a 2D integers array to a CSV formatted string (SerializeOut)
/*! \param Array2D 2D array to be parsed
\param w width of the 2D array
\param h height of the 2D array
\param szOutput Pointer to the output string with the integers in CSV format.
\return 0 Failed
\return 1 Succeed*/
int cMSTE_MapBase::A2DtoCSV(unsigned int **Array2D, int w, int h, char *&szOutput)
{
    int x,y;
    char sBuffer[10]; // Max characters is 10 for a 32bits integer

    // Allocating a buffer to save the map values							  
    szOutput = new char[h * w * sizeof(sBuffer)];

    // Loop through our map saving values to szOutput
    strcpy(szOutput,"");
    for(y = 0; y < h; y++)
    {
        for(x = 0; x < w; x++)
        {				
            sprintf(sBuffer,"%u",Array2D[y][x]);			
            strcat(szOutput,sBuffer);
            if(x != w - 1 || y != h - 1) strcat(szOutput,",");
        }
        // We add a blank space
        strcat(szOutput," ");
    }
    return 1;
}

//! Convert a CSV formatted string to a 2D integers array (SerializeIn)
/*! \param Array2D destination 2D array of the parsed integers
\param w width of the 2D array
\param h height of the 2D array
\param szInput Pointer to the input string with the integers in CSV format.
\return 0 Failed
\return 1 Succeed*/
int cMSTE_MapBase::CSVtoA2D(unsigned int **Array2D, int w, int h, char *&szInput)
{
    char  seps[] = ", ";
    char *token;
    unsigned int   iTmp = 0, x = 0, y = 0;

    // Start the integers parsing process
    token = strtok(szInput, seps);
    while((token != NULL) && (y != h))
    {      
        iTmp = strtoul(token,NULL,0);
        Array2D[y][x] = iTmp;
        x++;
        if(x == w) 
        { 
            x = 0; 
            y++; 
        }
        token = strtok( NULL, seps );
    }	
    // With (x = 0) and (y = h) we now at least that enough values were loaded in to our iTilemap
    if((x == 0) && (y == h)) 
    {
        // Finally, we check the map and return
        return CheckMap();
    }
    ILogSystem.Msg(LOG_NORMAL," [MSTE] Warning: 2D integer array is bigger than CSV input data\n");
    return 0;
}

//! Convert a string to a 2D integers array (SerializeIn)
/*! \param Array2D destination 2D array of the parsed integers
\param w width of the 2D array
\param h height of the 2D array
\param szInput Pointer to the input string
\return 0 Failed
\return 1 Succeed*/
int cMSTE_MapBase::StringtoA2D(unsigned int **Array2D, int w, int h, char *&szInput)
{
    int x = 0, y = 0, i;

    // Fill up our array
    const unsigned char *data = reinterpret_cast<const unsigned char*>(szInput);
    for(i = 0; i < (w * h * 4) - 3; i += 4)
    {
        const unsigned int gid = data[i] | data[i + 1] << 8 | data[i + 2] << 16 | data[i + 3] << 24;
        Array2D[y][x] = gid;
        x++;
        if(x == w)
        {
            x = 0;
            y++;
        }
    }

    // With (x = 0) and (y = h) we now at least that enough values were loaded in to our array
    if((x == 0) && (y == h)) 
    {
        // Finally, we check the map and return
        return CheckMap();
    }
    return 0;
}

//! Update a cell
/*! When you inherit from cMSTE_MapBase to create a custom mapdata you could overload this method<br>
to execute your own updating code. This method does not perform any action in the base mapdata level
\param x_cell Cell X coordinate to be updated
\param y_cell Cell y coordinate to be updated
\param TE Pointer to the TileEngine instance
\param v1 Pointer to an user defined data value or structure
\param v2 Pointer to an user defined data value or structure
\return 0 Failed
\return 1 Succeed*/
int cMSTE_MapBase::CellUpdate(int x_cell, int y_cell, cMSTE *TE, void *v1, void *v2)
{	
    // Sample code for testing Update() method
    /*SDL_Rect myR;
    myR.x = (int)v1; myR.w = 64; myR.y = (int)v2; myR.h = 64;
    SDL_FillRect(CRM32Pro.screen,&myR,0x00ff0000);
    char sAux1[8], sAux2[8];
    sprintf(sAux1,"(%d,",x_cell);
    sprintf(sAux2,"%d)",y_cell);
    char sTmp[16];
    strcpy(sTmp,sAux1); strcat(sTmp,sAux2);
    CRM32Pro.ConGetFont()->PutString(CRM32Pro.screen,myR.x,myR.y,sTmp);*/
    // ---------------------------------------

    return 1;
}

//! Render a cell
/*! When you inherit from cMSTE_MapBase to create a custom mapdata you could overload this method
to execute your own rendering code. Additionally, you must call this method in order to render the base mapdata<br>
We support rendering tiles with different size to the underlying cells: larger tiles will extend at the top and right.
\param x_cell Cell X coordinate to be rendered
\param y_cell Cell y coordinate to be rendered
\param x_dst Destination surface X coordinate to start to render in case you wish to
\param y_dst Destination surface Y coordinate to start to render in case you wish to
\param TE Pointer to the TileEngine instance
\param iAlpha Alpha per-surface value for this layer. With 255, it will use the alpha per-surface of the used tilesets.
\return 0 Failed
\return 1 Succeed*/
int cMSTE_MapBase::CellRender(int x_cell, int y_cell, int x_dst, int y_dst, cMSTE *TE, int iAlpha)
{
    int preAlpha = -1;

    // Get the tile to render
    unsigned int iCv = iTilemap[y_cell][x_cell];
    int iT = TE->CheckTile(iCv);

    // Only draw the tile if iT != 0
    if(iT >= 0)
    {
        // Support rendering tiles with different sizes and keep the original alpha of the tile (if applies)
        int iOffY = 0; // Calcs for Y axis
        TE->vTilesets[iT]->tTileset->GetOffset(NULL,&iOffY);
        iOffY = iCellHeight - iOffY;
        // Save source tileset alpha and set the new one if applies
        if(iAlpha != 255)
        {
            preAlpha = TE->vTilesets[iT]->tTileset->GetAlpha();
            // Avoid with alpha per-pixel
            if(preAlpha != 256) TE->vTilesets[iT]->tTileset->SetAlpha(iAlpha);
        }
        // Set tileposition and draw it!
        TE->vTilesets[iT]->tTileset->SetPosition(x_dst, y_dst + iOffY);
        TE->vTilesets[iT]->tTileset->Draw(TE->GetRenderTarget(),iCv - TE->vTilesets[iT]->iGID + 1);
        // Restore source tileset alpha
        if((iAlpha != 255) && (preAlpha != -1)) TE->vTilesets[iT]->tTileset->SetAlpha(preAlpha);
    }

    // Sample code for testing Render() method
    /*char sAux1[8], sAux2[8];
    sprintf(sAux1,"(%d,",x_cell);
    sprintf(sAux2,"%d)",y_cell);
    char sTmp[16];
    strcpy(sTmp,sAux1); strcat(sTmp,sAux2);
    CRM32Pro.ConGetFont()->PutString(CRM32Pro.screen,x_dst,y_dst,sTmp);*/
    // ---------------------------------------
    return 1;
}




