/*----------------------------------------------
   MegaStorm Tile-based Engine - Roberto Prieto
   Copyright (C) 2007-2012 MegaStorm Systems
  ----------------------------------------------
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
    
    Roberto Prieto
    contact@megastormsystems.com
    http://www.megastormsystems.com

    Used software licenses, including the one for this engine, are located
    at /licenses directory. Please, include them on your distributions.

 --------------------------------------------------------------------------
             MSTE MapData class
    
    Changelog: 
 v2.3 July/2012		- CellRender() support layer opacity
 v2.2 June/2012		- CellUpdate() and CellRender() methods
 v2.1 May/2012		- New map format
 v2.0 July/2007		- Final hierarchy: MapData is the base class. 
                    - Full working custom map support.
                    - MapBase ready to be used and to be derivated on your custom map
 v1.1 June/2007		- Parse a CSV string to allocate a map
 v1.0 April/2007	- Basic and needed stuff for a minimum map definition.
                    - Dynamic 2D array for the tilemap
                    - Public template for dynamic 2D arrays
   
 -------------------------------------------------------------------------- */
#ifndef _MSTE_MapData_H_
#define _MSTE_MapData_H_
#include <CRM32Pro.h>
#include "MSTE.h"

/** @defgroup MSTE_MAP MSTE MapData v2.3 - MapData class hierachy
\image html logotipo_mste.png
* <br>These are the MapData classes.
* - Created on 19 July 2007
* - Last modified on 6 July 2012*/
// -----------------------------------------------------------------------------
//! @{  

//-------------------------------------------------------------------------------
// MSTE MapData abstract base class.
//-------------------------------------------------------------------------------
//! MapData abstract base class
/*! MapData class definition.*/ 
class cMSTE_MapData
{
public:
    // Methods for dynamic 2D arrays management
    template < typename T >  T **Create2DArray( int height, int width);
    template < typename T >  void Delete2DArray(T** Array);

    // General stuff methods
    cMSTE_MapData();
    virtual ~cMSTE_MapData();

    // Pure virtual Map Access methods (public)
    virtual int CheckMap() = 0; 
    virtual int PrintInfo(char,int) = 0;
    virtual int CellUpdate(int, int, cMSTE *TE, void *, void *) = 0;
    virtual int CellRender(int, int, int, int, cMSTE *TE, int) = 0;
    virtual int SaveMapdataNode(int,int) = 0;
    virtual int LoadMapdataNode(int,int) = 0;

    // Properties methods
    char *GetMapType();
    int   GetCellWidth();
    void  SetCellWidth(int size);
    int   GetCellHeight();
    void  SetCellHeight(int size);
    int   GetMapWidth();
    int   GetMapHeight();

protected:
    // Internal properties
    int  iCellWidth;   // Width in pixels of each cell
    int  iCellHeight;  // Height in pixels of each cell
    int  iMapWidth;    // Width in cells of the map
    int  iMapHeight;   // Height in cells of the map
    char sMapType[16]; // Map type
};


//-------------------------------------------------------------------------------
// MSTE MapBase derivated class. Inherit from it your own custom map.
//-------------------------------------------------------------------------------
//! MapBase derivated class.
/*! This is a valid mapdata to be used with MSTE.
<br>Inherit from it your own custom MapData class definition.*/ 
class cMSTE_MapBase : public cMSTE_MapData
{
public:

    // General stuff methods
    cMSTE_MapBase();
    cMSTE_MapBase(int height,int width);
    ~cMSTE_MapBase();	 
    static cMSTE_MapBase *createInstance(int height,int width); // Create an instance of this class.

    // Definition of Map Access methods for this kind of map (public)
    int CheckMap(); 	
    int PrintInfo(char,int);
    int CellUpdate(int, int, cMSTE *TE, void *, void *);
    int CellRender(int, int, int, int, cMSTE *TE,int);
    int SaveMapdataNode(int i,int iOutput);
    int LoadMapdataNode(int i,int iInput);	

    // MapBase methods
    int CreateMapBase(int height, int width, short int value = -1);
    int DeleteMapBase();	
    int SetMapBaseValue(int y, int x, short int value);
    int CloneMapBase(cMSTE_MapBase *);

    // Dynamic 2D array for the tilemap
    unsigned int **iTilemap;	

protected:
    // Protected methods
    int A2DtoCSV(unsigned int **Array2D, int w, int h, char *&szOutput); 
    int CSVtoA2D(unsigned int **Array2D, int w, int h, char *&szInput);
    int StringtoA2D(unsigned int **Array2D, int w, int h, char *&szInput);
};

//-------------------------------------------------------------------------------
// Templates for dynamic 2D arrays
//-------------------------------------------------------------------------------
//! Create a dynamic 2D array of a given data type.
/*! \param height Selected height of your 2D array
\param width Selected width of your 2D array
\return A Pointer to the dynamic 2D array*/
template < typename T > T** cMSTE_MapData::Create2DArray(int height,int width)
{
    T **ppi;
    T *pool;
    T *curPtr;

    // 1.Allocate memory for array of elements of column
    ppi = new T*[height];

    // 2.Allocate memory for array of elements of each row
    pool = new T [width * height];

    // 3.Now point the pointers in the right place
    curPtr = pool;
    for( int i = 0; i < height; i++)
    {
        *(ppi + i) = curPtr;
        curPtr += width;
    }
    // 4.Return
    return ppi;
}

//! Delete a dynamic 2D array
template < typename T > void cMSTE_MapData::Delete2DArray(T** Array)
{
    delete [] *Array;
    delete [] Array;
}
//-------------------------------------------------------------------------------
//! @}
#endif
//-------------------------------------------------------------------------------
// End of MSTE MapData definition
//-------------------------------------------------------------------------------
